//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/MorphPlugin/Morpher.cs 23    5/24/06 10:18p Eaddy $
//
// Description:
//
//    Morpher is an API wrapper for running the TypeMorphPhase and
//    WeavingPhase phases.
//
// Remarks:
//
//    The TypeMorphPhase and WeavingPhase phases can be executed on
//    an assembly using pereader.exe:
//
//       pereader.exe /plugin:MorphPlugin.dll /r:...
//
//    In this case, the MorphPlugin reads configuration options from
//    the pereader command-line.  Morpher just provides a programmatic
//    interface equivalent to the pereader command-line interface.
//
//    In the command-line interface, aspects are specified in one shot
//    using the /r option.  If the base program happens to reference
//    an aspect but doesn't also specify it via /r, the aspect won't be
//    woven.
//
//    Morpher, however, can handle aspects being loaded dynamically,
//    in response to an assembly reference being resolved or a user's
//    interactive request, for example.  The base program is rewoven
//    each time a reference/aspect is added.
//    
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/MorphPlugin/Morpher.cs $
 * 
 * 23    5/24/06 10:18p Eaddy
 * o Verbosity of 0 now surpresses all messages
 * 
 * 22    5/24/06 9:48p Eaddy
 * o Now prints diagnostic messages in color
 * o Verbosity is now static
 * 
 * 21    5/11/06 10:49p Eaddy
 * o Fixed exception
 * 
 * 20    5/11/06 9:47p Eaddy
 * o Tried suppressing assertion messages (doesn't work)
 * 
 * 19    5/11/06 11:07a Eaddy
 * o Can now run mdbg from a different directory
 * 
 * 18    5/05/06 6:18p Eaddy
 * o Did some renaming to improve clarity
 * o Can now list current advice and enable/disable (although
 * enable/disable doesn't actually work yet)
 * 
 * 17    5/05/06 12:06p Eaddy
 * o Changed message
 * 
 * 16    5/05/06 11:48a Eaddy
 * o No longer appends an extra counter if the filename already has one
 * 
 * 15    5/04/06 8:01p Eaddy
 * o Now cleans up tmp files if nothing changed
 * 
 * 14    5/03/06 12:21a Eaddy
 * o Now reports when weaving fails
 * 
 * 13    5/02/06 5:12p Eaddy
 * o Now client is assumed to specify heuristics
 * 
 * 12    5/02/06 5:11p Eaddy
 * 
 * 11    5/02/06 5:05p Eaddy
 * o Optimizing dynamic weaving
 * 
 * 10    4/20/06 4:42p Eaddy
 * o Changed comment
 * 
 * 9     4/19/06 1:16a Eaddy
 * 
 * 8     4/19/06 12:15a Eaddy
 * 
 * 7     4/19/06 12:02a Eaddy
 * o Made OutputPath setter public so that clients can clear it
 * o SetInputModule no longer triggers an automatic weave
 * 
 * 6     4/11/06 12:41p Eaddy
 * o Fixed bug where the same assembly was loaded manually using the
 * "aspect" mdbg
 * command and automatically because it was linked by the client
 * application, which
 * triggered weaving to occur twice
 * 
 * 5     4/04/06 9:27p Eaddy
 * o Change tmp path names
 * 
 * 4     3/16/06 1:56a Eaddy
 * o Added OutputPath
 * 
 * 3     3/15/06 2:36p Eaddy
 * o Provides way to set breakpoint callback
 * o Creates unique tmp file name
 * 
 * 2     3/15/06 12:53p Eaddy
 * o Now MorphPlugin adds Summary Phase
 * 
 * 1     3/15/06 12:30p Eaddy
 * o Initial checkin
 */

using System;
using System.IO;
using System.Text;
using System.Collections;
using System.Collections.Generic;

using Dbg = System.Diagnostics.Debug;

using Phx.Morph.Editors;

namespace Phx.Morph
{
   public class Morpher
   {
      string inputImagePath = null;
      string outputImagePath = null;
      string tmpDir = Path.Combine(Environment.CurrentDirectory, "tmp");

      static int verbosity = 0; // 0=errors, 1=errors+warnings, 2=all
      int _tmpCount = 0;

      public enum WeaveResult { FAILED, SUCCESS, NO_CHANGE };

      public Morpher()
      {
         // Setup X86/MSIL Target and arch

         Phx.Targets.Archs.Arch x86Arch =
            Phx.Targets.Archs.X86.Arch.New();
         Phx.Targets.Runtimes.Runtime win32x86Runtime =
            Phx.Targets.Runtimes.VCCRT.Win32.X86.Runtime.New(x86Arch);
         Phx.GlobalData.RegisterTargetArch(x86Arch);
         Phx.GlobalData.RegisterTargetRuntime(win32x86Runtime);

         Phx.Targets.Archs.Arch msilArch =
            Phx.Targets.Archs.MSIL.Arch.New();
         Phx.Targets.Runtimes.Runtime win32MSILRuntime =
            Phx.Targets.Runtimes.VCCRT.Win32.MSIL.Runtime.New(msilArch);
         Phx.GlobalData.RegisterTargetArch(msilArch);
         Phx.GlobalData.RegisterTargetRuntime(win32MSILRuntime);

         Phx.Init.BeginInit();
         Phx.GlobalData.StaticInit();

         // Load the Phx.Morph plugin

         string morphPlugin = null;
         string morphPlugin1 = null;
         string morphPlugin2 = null;

         morphPlugin1 = Path.Combine(Environment.CurrentDirectory, "MorphPlugin.dll");
         if (!System.IO.File.Exists(morphPlugin1))
         {
            // Find out the executing assembly information
            System.Reflection.Assembly executingAssembly = System.Reflection.Assembly.GetExecutingAssembly();

            // And the executable folder
            string exeFolder = System.IO.Path.GetDirectoryName(executingAssembly.Location);

            morphPlugin2 = Path.Combine(exeFolder, "MorphPlugin.dll");
            if (System.IO.File.Exists(morphPlugin2))
            {
               morphPlugin = morphPlugin2;
            }
         }
         else
         {
            morphPlugin = morphPlugin1;
         }

         if (!System.IO.File.Exists(morphPlugin))
         {
             Morpher.WriteError("Morph plugin not found, weaving is disabled:");
             Morpher.WriteError("    " + morphPlugin1 + " (NOT FOUND)");
             Morpher.WriteError("    " + morphPlugin2 + " (NOT FOUND)");
         }
         else
         {
            Phx.Ctrls.PlugInCtrlValue pluginCtrl = Phx.Ctrls.PlugInCtrlValue.New();
            pluginCtrl.ValueString = morphPlugin;
            Phx.GlobalData.RegisterPlugIn(pluginCtrl);
         }

         string[] extraFlags;
         if (Verbosity > 0)
            extraFlags = new string[] { "/assertlimit=0" };
         else
            extraFlags = new string[] { };

         // Causes MorphPlugin.RegisterObjects() to get called
         Phx.Init.EndInit("PHX|*|_PHX_", extraFlags );
      }

      public static int Verbosity
      {
         [System.Diagnostics.DebuggerStepThrough]
         get { return verbosity; }

         [System.Diagnostics.DebuggerStepThrough]
         set { verbosity = value; }
      }

      public string InputModule
      {
         [System.Diagnostics.DebuggerStepThrough]
         get { return inputImagePath; }

         [System.Diagnostics.DebuggerStepThrough]
         set { inputImagePath = value; }
      }

      public string OutputPath
      {
         [System.Diagnostics.DebuggerStepThrough]
         get { return outputImagePath; }

         [System.Diagnostics.DebuggerStepThrough]
         set { outputImagePath = value; }
      }

      public static IEnumerable Advices
      {
         [System.Diagnostics.DebuggerStepThrough]
         get { return WeavingPhase.Advices; }
      }

      public static Phx.Aop.BreakpointWeaverCallback BreakpointWeaverCallback
      {
         [System.Diagnostics.DebuggerStepThrough]
         set { BreakpointWeaver.BreakpointWeaverCallback = value; }
      }

      public void AddAspectCandidate(string aspectCandidatePath)
      {
         TypeMorphPhase.AddAspectCandidate(aspectCandidatePath);
      }

      public WeaveResult Weave()
      {
         if (inputImagePath == null)
            return WeaveResult.FAILED;

         Phx.Ctrls.Parser.ParseArgString(null, "/v:" + verbosity);

         Phx.PEModuleUnit moduleUnit = Phx.PEModuleUnit.Open(inputImagePath);
         Phx.Phases.PhaseList phaseList = moduleUnit.SymsReadingPhase.Config.PhaseList;
         BuildPhaseLists(phaseList);

         //phaseList.Dump();

         if (BreakpointWeaver.BreakpointWeaverCallback == null)
         {
            outputImagePath = createOutputFilePath();

            moduleUnit.OutputImagePath = outputImagePath;
            moduleUnit.OutputPdbPath = Path.Combine(tmpDir,
               Path.GetFileNameWithoutExtension(outputImagePath) + ".pdb");

            Dbg.Assert(!File.Exists(moduleUnit.OutputImagePath));
            Dbg.Assert(!File.Exists(moduleUnit.OutputPdbPath));
         }
         else
         {
            outputImagePath = null;
         }

         WeaveResult result = WeaveResult.FAILED;

         try
         {
            phaseList.DoPhase(moduleUnit);
            moduleUnit.Close();
            result = TypeMorphPhase.SomethingChanged ? WeaveResult.SUCCESS : WeaveResult.NO_CHANGE;
         }
         catch (Exception ex)
         {
            Morpher.WriteError("A {0} exception occured while morphing '{1}': {2}",
               ex.GetType(), inputImagePath, ex.Message);

            result = WeaveResult.FAILED;
         }

         if (result != WeaveResult.SUCCESS)
         {
            // Cleanup any temporary files that won't be used
            if (File.Exists(moduleUnit.OutputImagePath))
            {
               File.Delete(moduleUnit.OutputImagePath);
            }

            if (File.Exists(moduleUnit.OutputPdbPath))
            {
               File.Delete(moduleUnit.OutputPdbPath);
            }

            outputImagePath = null;
         }

         return result;
      }

      void BuildPhaseLists(Phx.Phases.PhaseList phaseList)
      {
         Phx.Phases.PhaseList funcUnitPhases = Phx.PE.UnitListPhaseList.New(phaseList.Config,
            Phx.PE.UnitListWalkOrder.PrePass);

         Phx.Phases.Phase raiseIRPhase = Phx.PE.RaiseIRPhase.New(phaseList.Config,
            Phx.FuncUnit.LowLevelIRBeforeLayoutFuncUnitState);

         funcUnitPhases.AppendPhase(raiseIRPhase);

         phaseList.AppendPhase(funcUnitPhases);

         Phx.GlobalData.BuildPlugInPhases(phaseList.Config);
      }

		/// <summary>
		/// Creates a temporary filename for the module in the tmp directory.
		/// </summary>
		/// <returns>Path to temporary file.</returns>
      private string createOutputFilePath()
      {
         string inputImageFileName = Path.GetFileNameWithoutExtension(inputImagePath);
         string inputImageFileExt = Path.GetExtension(inputImagePath);
         string outputImagePath;

         // If the input file name already has a counter then use it
         // instead of appending another counter

         string ext = Path.GetExtension(inputImageFileName);
         int num = 0;
         if (  ext != null && 
               ext.Length > 1 && 
               int.TryParse(ext.Substring(1), out num) 
               && num >= 0)
         {
            inputImageFileName = Path.GetFileNameWithoutExtension(inputImageFileName);
            _tmpCount = num + 1;
         }

         if (!Directory.Exists(tmpDir))
            Directory.CreateDirectory(tmpDir);

         // Keep incrementing file name until we find one that doesn't exist
         do
         {
            outputImagePath = Path.Combine(tmpDir,
               inputImageFileName + "." + _tmpCount.ToString() + inputImageFileExt);
            ++_tmpCount;

         } while (File.Exists(outputImagePath));

         return outputImagePath;
      }

      static public void WriteOutput(string format, params object[] args)
      {
         if (verbosity >= 3)
         {
            ConsoleColor fc = Console.ForegroundColor;
            Console.ForegroundColor = ConsoleColor.White;
            System.Console.WriteLine("Phx.Morph: " + format, args);
            Console.ForegroundColor = fc;
         }
      }

      static public void WriteWarning(string format, params object[] args)
      {
         if (verbosity >= 2)
         {
            ConsoleColor fc = Console.ForegroundColor;
            Console.ForegroundColor = ConsoleColor.Yellow;
            System.Console.WriteLine("Phx.Morph: Warning: " + format, args);
            Console.ForegroundColor = fc;
         }
      }

      static public void WriteError(string format, params object[] args)
      {
         if (verbosity >= 1)
         {
            ConsoleColor fc = Console.ForegroundColor;
            Console.ForegroundColor = ConsoleColor.Red;
            System.Console.WriteLine("Phx.Morph: " + format, args);
            Console.ForegroundColor = fc;
         }
      }
   }
}
