//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Aop/Attributes/JoinpointAttributes.cs 6     5/11/06 3:42p Eaddy $
//
// Description:
//
//    Extension attributes for accessing joinpoint information
//
// Remarks:
//
//    Joinpoint extension attributes can be specified on method
//    parameters.  The parameters will be inlined at morph time
//    with the appropriate values.  The only purpose of the
//    attributes is to mark the method parameters for morph-time
//    inlining.
//
//    Some attributes are not yet implemented (see below).
//
// Example:
//
// [Advice(AdviceType.around, "call(...Add)")]
// static public void MyTrace([AdviceType] AdviceType adviceType,
//                            [Kind] PointcutKind kind,
//                            [SignatureAttribute.Name] string name,
//                            [SourceLocationAttribute.WithinSignature] string withinSignature,
//                            [SourceLocationAttribute.FilePath] string filePath,
//                            [SourceLocationAttribute.Line] uint line,
//                            [This] Node This)
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Aop/Attributes/JoinpointAttributes.cs $
 * 
 * 6     5/11/06 3:42p Eaddy
 * o Target now supported for execution pointcuts
 * 
 * 5     5/03/06 12:44p Rr2318
 * Added AnnotationAttribute
 * 
 * 4     3/03/06 2:46p Eaddy
 * o Added FileName
 * 
 * 3     2/24/06 8:17p Eaddy
 * o Initial checkin
 */

using System;

namespace Phx.Aop.Attributes
{
   /// <summary>
   /// Equivalent to thisJoinPoint.Args[0]
   /// </summary>
   /// <remarks>
   /// Only works for execution pointcuts
   /// </remarks>

   [AttributeUsage(AttributeTargets.Parameter)]
   public class ArgAttribute : Attribute
   {
      int index;

      public ArgAttribute(int index)
      {
         this.index = index;
      }

      public int Index
      {
         get { return this.index; }
      }
   }

   [AttributeUsage(AttributeTargets.Parameter)]
   public class AnnotationAttribute : Attribute
   {
      int index;

      public AnnotationAttribute(int index)
      {
         this.index = index;
      }

      public int Index
      {
         get { return this.index; }
      }
   }

   [AttributeUsage(AttributeTargets.Parameter)]
   public class AnnotationsAttribute : Attribute
   {
      int size;
      System.Attribute[] attrList;

      public AnnotationsAttribute(int s, System.Attribute[] attr)
      {
         this.attrList = attr;
         this.size = s;
      }

      public System.Attribute[] AttributeList
      {
         get { return this.attrList; }
      }

      public int Size
      {
         get { return this.size; }
      }
   }

   /// <summary>
   /// Equivalent to thisJoinPoint.Kind
   /// </summary>

   [AttributeUsage(AttributeTargets.Parameter)]
   public class KindAttribute : Attribute
   {
   }

   /// <summary>
   /// Equivalent to AspectJ's Signature.ToString()
   /// </summary>

   [AttributeUsage(AttributeTargets.Parameter)]
   public class SignatureAttribute : Attribute
   {
      /// <summary>
      /// Equivalent to AspectJ's Signature.DeclaringType 
      /// </summary>

      [AttributeUsage(AttributeTargets.Parameter)]
      public class DeclaringTypeAttribute : Attribute
      {
         /// <summary>
         /// ISSUE-NYI-t-marcea-2005/08/02
         /// </summary>

         // private so no one can create (since its NYI)

         private DeclaringTypeAttribute()
         { }
      }

      /// <summary>
      /// Equivalent to AspectJ's Signature.DeclaringTypeName 
      /// </summary>

      [AttributeUsage(AttributeTargets.Parameter)]
      public class DeclaringTypeNameAttribute : Attribute
      {
      }

      /// <summary>
      /// Equivalent to AspectJ's Signature.Name
      /// </summary>

      [AttributeUsage(AttributeTargets.Parameter)]
      public class NameAttribute : Attribute
      {
      }
   }

   /// <summary>
   /// Equivalent to AspectJ's SourceLocation.ToString()
   /// </summary>

   [AttributeUsage(AttributeTargets.Parameter)]
   public class SourceLocationAttribute : Attribute
   {
      [AttributeUsage(AttributeTargets.Parameter)]
      public class FilePathAttribute : Attribute
      {
      }

      [AttributeUsage(AttributeTargets.Parameter)]
      public class FileNameAttribute : Attribute
      {
      }

      [AttributeUsage(AttributeTargets.Parameter)]
      public class LineAttribute : Attribute
      {
      }

      [AttributeUsage(AttributeTargets.Parameter)]
      public class WithinTypeNameAttribute : Attribute
      {
      }

      [AttributeUsage(AttributeTargets.Parameter)]
      public class WithinNameAttribute : Attribute
      {
      }

      [AttributeUsage(AttributeTargets.Parameter)]
      public class WithinSignatureAttribute : Attribute
      {
      }
   }

   /// <summary>
   /// Equivalent to AspectJ's thisJoinPoint.Target
   /// </summary>
   /// <remarks>
   /// Only works for execution pointcuts
   /// </remarks>

   [AttributeUsage(AttributeTargets.Parameter)]
   public class TargetAttribute : Attribute
   {
   }

   /// <summary>
   /// When used on a Parameter, causes the parameter to
   /// replaced with the this ptr similar to 
   /// AspectJ's thisJoinPoint.This.  When used on a Field,
   /// marks the field so references to the field are replaced
   /// with the this ptr.
   /// </summary>
   /// <remarks>
   /// When ThisAttribute is applied to a parameter, it is
   /// treated as AspectJ's thisJoinPoint.This.  When it is
   /// applied to a Field, it is serves to "mark" the field
   /// so that during morphing all references to the field
   /// are replaced with the actual this ptr.  This is needed
   /// when refering to the this ptr inside a method that's
   /// imported from another class.
   /// <br>
   /// The AOP version of ThisAttribute only works for
   /// execution pointcuts.
   /// </br>
   /// </remarks>

   [AttributeUsage(AttributeTargets.Parameter | AttributeTargets.Field)]
   public class ThisAttribute : Attribute
   {
   }

   /// <summary>
   /// There is no equivalent for AdviceType in AspectJ
   /// </summary>

   [AttributeUsage(AttributeTargets.Parameter)]
   public class AdviceTypeAttribute : Attribute
   {
   }
}
