//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Aop/Pointcuts/Expressions/WildcardMatch.cs 1     3/01/06 2:17a Eaddy $
//
// Description:
//
//    Simple wildcard matcher.
//
// Remarks:
//
//    Implements a subset of AspectJ's wildcarding.
//
//    * matches any number of characters not including a '.'
//    .. matches any number of characters including a '.'
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Aop/Pointcuts/Expressions/WildcardMatch.cs $
 * 
 * 1     3/01/06 2:17a Eaddy
 * o Now uses Phx.Aop library for pointcuts, advice and joinpoint matching
 * 
 * 2     2/24/06 8:27p Eaddy
 * o Added checkin comment
 */

using System;
using System.Text;

namespace Phx.Aop.Pointcuts
{
   public class WildcardMatch
   {
      enum WildKind { NoDot, All, None };

      const string WildNoDot = "*";
      const string WildAll = "..";
      static int WildNoDotLen = WildNoDot.Length;
      static int WildAllLen = WildAll.Length;

      string _pattern;

      StringBuilder sbPattern;
      StringBuilder sbMatchString;

      public WildcardMatch
      (
         string pattern
      )
      {
         this._pattern = pattern;
      }

      //--------------------------------------------------------------------------
      //
      // Description:
      //
      //    Determine if matchString matches the pattern.
      //
      // Remarks:
      //
      //    This is the external entry point which simply delegates to the
      //    internal worker.
      //
      //--------------------------------------------------------------------------

      public bool
      IsAMatch
      (
         string matchString
      )
      {
         InitStringBuilders(matchString);
         return CheckMatch(sbPattern, sbMatchString);
      }

      public bool
      IsAMatch
      (
         StringBuilder matchString
      )
      {
         InitStringBuilders(matchString.ToString());
         return CheckMatch(sbPattern, sbMatchString);
      }

      private void InitStringBuilders(string matchString)
      {
         if (sbPattern == null)
            sbPattern = new StringBuilder(_pattern);
         else
         {
            sbPattern.Remove(0, sbPattern.Length);
            sbPattern.Append(_pattern);
         }

         if (sbMatchString == null)
            sbMatchString = new StringBuilder(matchString);
         else
         {
            sbMatchString.Remove(0, sbMatchString.Length);
            sbMatchString.Append(matchString);
         }
      }

      private static bool
      CheckMatch
      (
         StringBuilder pattern,
         StringBuilder matchString
      )
      {
         WildKind wildKind = WildKind.None;

         int len = pattern.Length;
         for(; len > 0; --len)
         {
            wildKind = LeadingWildString(pattern);
            if (wildKind != WildKind.None)
            {
               break;
            }

            if (matchString.Length == 0 ||
               pattern[0] != matchString[0])
            {
               return false;
            }

            pattern.Remove(0, 1);
            matchString.Remove(0, 1);
         }

         if (len == 0 && matchString.Length == 0)
         {
            return true;
         }
         else if (wildKind == WildKind.None)
         {
            return false;
         }

         // Consume the wildcard.

         int lenToConsume = wildKind == WildKind.All ? WildAllLen : WildNoDotLen;

         len -= lenToConsume;
         pattern.Remove(0, lenToConsume);

         for (; ; )
         {
            if (CheckMatch(pattern, matchString))
            {
               return true;
            }
            else if (matchString.Length == 0)
            {
               return false;
            }

            char c = matchString[0];

            if (wildKind == WildKind.All || c != '.')
            {
               matchString.Remove(0, 1);
            }
            else
            {
               return false;
            }
         }
      }

      //--------------------------------------------------------------------------
      //
      // Description:
      //
      //    Determine if the pattern starts with a wildcard.
      //
      // Returns:
      //
      //    Returns the wildcard string or null if none.
      //
      //--------------------------------------------------------------------------

      private static WildKind
      LeadingWildString(StringBuilder pattern)
      {
         int len = pattern.Length;
         int i = 0;
         for (; i < len && i < WildNoDotLen; ++i)
         {
            if (pattern[i] != WildNoDot[i])
               break;
         }

         if (i == WildNoDotLen)
            return WildKind.NoDot;

         i = 0;
         for (; i < len && i < WildAllLen; ++i)
         {
            if (pattern[i] != WildAll[i])
               break;
         }

         if (i == WildAllLen)
            return WildKind.All;

         return WildKind.None;
      }

      public bool
      MatchesAnything()
      {
         int len = _pattern.Length;
         int i = 0;

         if (len == WildNoDotLen)
         {
            for (; i < len && i < WildNoDotLen; ++i)
            {
               if (_pattern[i] != WildNoDot[i])
                  break;
            }

            if (i == WildNoDotLen)
               return true;
         }

         if (len == WildAllLen)
         {
            i = 0;
            for (; i < len && i < WildAllLen; ++i)
            {
               if (_pattern[i] != WildAll[i])
                  break;
            }

            if (i == WildAllLen)
               return true;
         }

         return false;
      }

      public override string ToString()
      {
         return _pattern.ToString();
      }

#if MatchTest

   static private void
   Test
   (
      WildcardMatch matcher,
      string        matchString
   )
   {
      bool isAMatch = matcher.IsAMatch(matchString);

      Console.Write("{0} <=> {1} == ", matcher.Pattern, matchString);

      if (isAMatch)
      {
         Console.WriteLine("Matched");
      }
      else
      {
         Console.WriteLine("No match");
      }
   }

   static public void
   Main()
   {
      WildcardMatch most = new WildcardMatch("*");
      WildcardMatch none = new WildcardMatch("");
      WildcardMatch allscoped = new WildcardMatch("*.*");
      WildcardMatch all = new WildcardMatch("..");
      WildcardMatch derivedfooclick = new WildcardMatch("foo.*.Click");
      WildcardMatch allfooclick = new WildcardMatch("foo...Click");

      WildcardMatch.Test(none, "any");
      WildcardMatch.Test(most, "any");
      WildcardMatch.Test(most, "foo.bar");
      WildcardMatch.Test(all, "foo.bar");
      WildcardMatch.Test(allscoped, "foo.bar");
      WildcardMatch.Test(allscoped, "foo.bar.Click");
      WildcardMatch.Test(allfooclick, "foo.Click");
      WildcardMatch.Test(allfooclick, "foo.Click.Snuggle");
      WildcardMatch.Test(allfooclick, "foo.Derived.Click");
      WildcardMatch.Test(allfooclick, "foo.DifferentClick");
      WildcardMatch.Test(allfooclick, "notfoo.Click");
      WildcardMatch.Test(derivedfooclick, "foo.Derived.Click");
      WildcardMatch.Test(derivedfooclick, "foo.DifferentClick");
   }

#endif
   }
}
