//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Morph/Attributes/AttributeHelper.cs 15    6/02/06 2:37a Eaddy $
//
// Description:
//
//    Helper class for dealing with attributes
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Morph/Attributes/AttributeHelper.cs $
 * 
 * 15    6/02/06 2:37a Eaddy
 * o Shouldn't show warnings since we expect
 * them to happen
 * 
 * 14    5/26/06 9:13p Eaddy
 * o Fixed bug where Add attribute with named parameters
 * wasn't handled
 * 
 * 13    5/24/06 9:56p Eaddy
 * o Now tries to resolve assemblies that can't be loaded
 * o Improved instantiation of custom attributes and finding
 * the correct ctor for a given blob
 * 
 * 12    5/12/06 9:36a Eaddy
 * o Rolled back Rajesh's change because it broke everything
 * 
 * 11    5/12/06 3:09a Rr2318
 * Added checks for other internal attributes (IsPhxAttribute)
 * 
 * 10    5/11/06 10:54p Rr2318
 * Modified to check blob
 * 
 * 9     5/04/06 8:43p Eaddy
 * o Added extra check
 * 
 * 8     5/04/06 4:49p Eaddy
 * o InstantiateAttribute should return an attribute
 * 
 * 7     5/04/06 3:08p Eaddy
 * o Added handler to TypeLoadException.  This will happen if that
 * attribute contains
 * a System.Type, MethodInfo, or other metadata class that internally
 * refer to 
 * tokens.  Since the tokens won't be in the EmittedAssembly, type loading
 * will fail.
 * 
 * 6     5/04/06 11:52a Eaddy
 * o We were not handling named parameters correctly, which caused the
 * visitor
 * test to fail
 * 
 * 5     5/03/06 5:09p Eaddy
 * o Fixed bug where an attribute with no ctor was causing
 * GetCustomAttributes
 * to return a zero length array
 * 
 * 4     5/03/06 12:31a Eaddy
 * o Fixed assertion
 * 
 * 3     5/02/06 12:48p Eaddy
 * o Added parameter check
 * 
 * 2     2/24/06 8:25p Eaddy
 * o Now uses attribute handler plugins instead of using 
 * attributes directly
 */

using System;
using System.Collections.Generic;

using Dbg = System.Diagnostics.Debug;

using Phx.Morph;

namespace Phx.Aop.Attributes
{
   /// <summary>
   /// Helper class for dealing with attributes
   /// </summary>

   public class AttributeHelper
   {
      /// <summary>
      /// Used for instantiating a custom attribute using its
      /// metadata blob
      /// </summary>

      System.Reflection.Emit.TypeBuilder typeBuilder = null;

      /// <summary>
      /// Used for instantiating a custom attribute using its
      /// metadata blob
      /// </summary>

      Dictionary<System.Type, int> typeBuilderIndex =
         new Dictionary<System.Type, int>(2);

      /// <summary>
      /// Looks for extension attributes in all the reference modules
      /// </summary>
      /// <remarks>
      /// An extension attributes is an attribute derived from
      /// IExtensionAttribute
      /// </remarks>
      /// <param name="moduleUnitRefs">Module references
      /// (specified via /r on the command-line)</param>
      /// <param name="extensions">
      /// Collection of found IExtensionAttribute objects
      /// </param>

      public void
      LocateExtensions
      (
         Phx.PEModuleUnit[] moduleUnitRefs,
         ref List<Phx.Morph.Attributes.IExtensionAttribute> extensions
      )
      {
         for (int i = 0; moduleUnitRefs != null && i < moduleUnitRefs.Length; ++i)
         {
            if (moduleUnitRefs[i] != null)
            {
               // Find all extension attributes

               LocateExtensions(moduleUnitRefs[i], ref extensions);
            }
         }
      }

      //--------------------------------------------------------------------------
      //
      // Description:
      //
      //    Looks for AddField attributes in the module
      //
      //--------------------------------------------------------------------------

      /// <summary>
      /// Looks for extension attributes in the module
      /// </summary>
      /// <param name="moduleUnit">Module to scan</param>
      /// <param name="extensions">
      /// Collection of found IExtensionAttribute objects
      /// </param>
      /// <returns>Count of found extension attributes</returns>

      public int
      LocateExtensions
      (
         Phx.PEModuleUnit moduleUnit,
         ref List<Phx.Morph.Attributes.IExtensionAttribute> extensions
      )
      { 
         if (moduleUnit == null ||
            moduleUnit.SymTable == null ||
            moduleUnit.SymTable.LocalIdMap == null ||
            moduleUnit.SymTable.LocalIdMap.InternalMap == null)
         {
            Dbg.Assert(false);
            return 0;
         }

         // Iterate over all the symbols in the module

         int count = 0;

         foreach (Phx.Syms.Sym sym in moduleUnit.SymTable.LocalIdMap.InternalMap)
         {
            count += LocateExtensions(moduleUnit, sym, ref extensions);
         }

         return count;
      }

      /// <summary>
      /// Looks for extension attributes for the symbol
      /// </summary>
      /// <remarks>
      /// We have to use .NET Reflection in order to
      /// instantiate an attribute with the same
      /// parameters specified when the attribute
      /// was used
      /// </remarks>
      /// <param name="moduleUnit">Symbol's module</param>
      /// <param name="sym">Symbol to scan</param>
      /// <param name="extensions">
      /// Collection of found IExtensionAttribute objects
      /// </param>
      /// <returns>Count of found extension attributes</returns>

      public int
      LocateExtensions
      (
         Phx.PEModuleUnit moduleUnit,
         Phx.Syms.Sym sym,
         ref List<Phx.Morph.Attributes.IExtensionAttribute> extensions
      )
      {
         Dbg.Assert(sym != null && extensions != null);
         if (sym == null || extensions == null)
         {
            return 0;
         }

         // Get the list of attributes for this symbol (type, field, etc.)

         Phx.Syms.AttrSymListExtensionObject attrSymListExtObj =
            Phx.Syms.AttrSymListExtensionObject.GetExtensionObject(sym);

         if (attrSymListExtObj == null || attrSymListExtObj.AttrSymList == null)
         {
            // No attributes found, ignore

            return 0;
         }

         // Iterate over all the attributes

         foreach (Phx.Syms.Sym attrExtSym in attrSymListExtObj.AttrSymList)
         {
            if (attrExtSym == null ||
               !attrExtSym.IsAttrSym)
            {
               // Ignore PermissionSyms

               continue;
            }

            Phx.Syms.AttrSym attrSym = attrExtSym.AsAttrSym;
            if (attrSym == null ||
                attrSym.AttrType == null)
            {
               Dbg.Assert(attrSym == null || attrSym.Kind == Phx.Syms.AttrSymKind.Marshaling);
               continue;
            }

            // Convert from Phx.Types.Type to System.Type

            // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
            // for unmanaged implementation

            System.Type attrTypeCurrent =
               ReflectionHelpers.ConvertPhxTypeToClrType(attrSym.AttrType);

            if (attrTypeCurrent == null)
            {
               // If an attribute can't be converted, its not one of
               // ours.

               continue;
            }

            // See if the attribute type implements IExtensionAttribute

            // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
            // for unmanaged implementation

            // ISSUE-BUG-t-marcea-2005/08/09-Phx.Types.Type.IsAssignable()
            // doesn't always tell the truth. If the type passed in is a
            // TypeRef it will always return true.

            Phx.Morph.Attributes.IAttributeHandler attrHandler = 
               Phx.Morph.Attributes.AttributeHandlerRegistry.GetHandler(attrTypeCurrent.ToString());

            if (attrHandler == null)
            {
               continue; // Doesn't match, ignore
            }

            // We found an extension attribute so instantiate it using
            // the attribute values used.  The attributes values are
            // stored in the metadata as an attribute blob.
            //
            // In other words, we want to turn:
            //
            //    [Add("HelloNS.Hello", Names="b1,b2")]
            //    bool b;
            //
            // into:
            //
            //    AddAttribute attr = new AddAttribute("HelloNS.Hello");
            //    attr.Names = "b1,b2";
            //

            Phx.Morph.Attributes.IExtensionAttribute extensionAttr =
               attrHandler.Create(attrSym, attrTypeCurrent, this);

            if (extensionAttr == null)
            {
               Helpers.WriteError("Failed to read '{0}' attribute.",
                  attrTypeCurrent.ToString());
               continue;
            }

            // Complete the initialization by hooking up the extension
            // attribute with the symbol that its attached to

            if (extensionAttr.Init(attrSym.OwnerSym, moduleUnit.TypeTable))
            {
               extensions.Add(extensionAttr);
            }
            else
            {
               Helpers.WriteError("Failed to initialize '{0}' attribute.",
                  attrTypeCurrent.ToString());
            }
         }

         return extensions.Count;
      }

      /// <summary>
      /// Checks to see if a symbol has a specific attribute
      /// </summary>
      /// <param name="sym">Symbol to check</param>
      /// <param name="attrType">Type of the attribute</param>
      /// <returns>true if the symbol has the attribute</returns>

      public static bool
      HasAttribute
      (
         Phx.Syms.Sym sym,
         System.Type attrType
      )
      {
         Dbg.Assert(sym != null && attrType != null);
         if (sym == null || attrType == null)
         {
            return false;
         }

         // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
         // for unmanaged implementation

         // Get the list of attributes for this symbol (type, field, etc.)

         Phx.Syms.AttrSymListExtensionObject attrSymListExtObj =
            Phx.Syms.AttrSymListExtensionObject.GetExtensionObject(sym);

         if (attrSymListExtObj == null || attrSymListExtObj.AttrSymList == null)
         {
            // No attributes found

            return false;
         }

         // Iterate over all the attributes

         foreach (Phx.Syms.AttrSym attrSym in attrSymListExtObj.AttrSymList)
         {
            Dbg.Assert(attrSym != null);

            // Convert from Phx.Types.Type to System.Type

            // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
            // for unmanaged implementation

            System.Type attrTypeCurrent =
               ReflectionHelpers.ConvertPhxTypeToClrType(attrSym.AttrType);

            if (attrTypeCurrent == null)
            {
               // Shouldn't happen

               Dbg.Assert(false);
               continue;
            }

            // See if the attribute matches the one we're looking for

            if (attrType.IsAssignableFrom(attrTypeCurrent))
            {
               return true;
            }
         }

         return false;
      }
      static public bool IsPhxAttribute(System.Type t)
      {
         string type = t.FullName;
         if (type.Equals("Phx.Aop.Attributes.AddAttribute")) return true;
         else if (type.Equals("Phx.Aop.Attributes.AdviceAttribute")) return true;
         else if (type.Equals("Phx.Aop.Attributes.ExtendsAttribute")) return true;
         else if (type.Equals("Phx.Aop.Attributes.ReplaceAttribute")) return true;
         else if (type.Equals("Phx.Aop.Attributes.StatementAttribute")) return true;
         return false;
      }
      /// <summary>
      /// Instantiate a custom attribute using the parameter values
      /// stored in the metadata
      /// </summary>
      /// <remarks>
      /// ISSUE-HACK-t-marcea-2005/08/03-The parameter values are
      /// stored using a binary format in attrSym.Blob.  The only
      /// way to legally read this blob without cracking it ourselves
      /// is to call
      /// TypeBuilder.SetCustomAttribute(ConstructorInfo ctor, byte[] blob).
      /// Unfortunately, this requires that we actually build a
      /// type, which also requires building a module and assembly.
      /// Phoenix should provide a way to do this.
      /// </remarks>
      /// <param name="attrSym">
      /// Phx symbol for the attribute. Needed for the attribute blob
      /// and info about the attribute's constructors.
      /// </param>
      /// <param name="attrType">.NET type for the attribute</param>
      /// <returns>Rehydrated attribute</returns>

      public System.Attribute
      InstantiateAttribute
      (
         Phx.Syms.AttrSym attrSym,
         System.Type attrType
      )
      {
         Dbg.Assert(attrSym != null);
         Dbg.Assert(attrType != null, "Phx.Morph: Did you load the module that defines this type?");
         if (attrSym == null || attrType == null)
            return null;

         // Step 1:  Create and cache the type builder

         if (typeBuilder == null)
         {
            typeBuilder = CreateTypeBuilder();
         }

         // Step 2:  Inject the custom attribute into the type

         // Step 2a: Find the constructor that corresponds to our attribute
         //          constructor parameter blob

         System.Reflection.ConstructorInfo ctorToTry =
            ReflectionHelpers.FindMatchingConstructorForBlob(attrType, attrSym.Blob, false);
         if (ctorToTry != null)
         {
            System.Attribute attributeNew = TryAttributeConstructor(ctorToTry, attrSym.Blob);
            if (attributeNew != null)
               return attributeNew;
         }

         System.Reflection.ConstructorInfo[] ctors =
            attrType.GetConstructors();

         Dbg.Assert(ctors.Length > 0);

         Phx.Syms.FuncSym ctorSym = attrSym.FuncSym;

         // Subtract one for this ptr

         int ctorArgCount = ctorSym.FuncType.CountArgs() - 1;

         // Retrieve all the custom attributes

         foreach(System.Reflection.ConstructorInfo ctorToTry2 in ctors)
         {
            // Find the ctor whose metadata token matches the ctor associated
            // with our custom attribute 

            // ISSUE-HACK-t-marcea-2005/08/03-Currently, there is no way to
            // match the funcSym constructor associated with the attribute
            // blob to the proper Reflection constructor.  We can't compare
            // Tokens since the Reflection token is a MethodDef and the
            // funcSym token is a MethodRef.  Since we explicitly loaded the
            // module into Phoenix during initialization, it should be
            // possible to locate the MethodDef funcSym for this constructor.
            // However, we can't do it by name comparison in the same way
            // we locate TypeDefs since constructors can be overloaded and
            // all have the same name (.ctor).

            if (ctorToTry2.MetadataToken != ctorSym.OriginalToken &&
               ctorToTry2.GetParameters().Length != ctorArgCount)
            {
               continue;
            }
            else if (ctorArgCount == 0 && attrSym.Blob.Length <= 4)
            {
               // There are no arguments and no named parameters (blob length is 4)
               // so just call the default ctor directly.  This is the only case
               // where we have an exact unambiguous match because we aren't
               // relying on the blob at all.

               return ctorToTry2.Invoke(null) as Attribute;
            }
            else
            {
               System.Attribute attributeNew = 
                  TryAttributeConstructor(ctorToTry2, attrSym.Blob);

               if (attributeNew != null)
                  return attributeNew;
            }
         }

         Helpers.WriteError(
            "Failed to locate matching custom attribute ctor for {0}. " +
            "Make sure the assembly that defines the custom attribute is up-to-date.",
            attrSym.AttrType.TypeSym.AsMsilTypeSym.NameString);

         return null;
      }

      public System.Attribute
      TryAttributeConstructor
      (
         System.Reflection.ConstructorInfo ctor,
         byte[] blob
      )
      {
         Dbg.Assert(ctor != null);
         Dbg.Assert(blob != null);
         Dbg.Assert(this.typeBuilder != null);

         // Step 2b: Inject/instantiate the custom attribute into this type.
         //          Note: If we call the wrong ctor with the wrong parameters,
         //          there is no way to know.

         this.typeBuilder.SetCustomAttribute(ctor, blob);

         // Step 3:  Retrieve the custom attribute instance

         // Create the newly minted type

         System.Type fakeType = this.typeBuilder.CreateType();

         try
         {
            object[] attributes = fakeType.GetCustomAttributes(ctor.DeclaringType, false);
            Dbg.Assert(attributes != null);
            Dbg.Assert(attributes.Length != 0,
               "Did you forget to load the attribute's assembly using Reflection?");

            int nextAttrIndex;
            if (!this.typeBuilderIndex.TryGetValue(ctor.DeclaringType, out nextAttrIndex))
            {
               nextAttrIndex = 0;
            }

            Dbg.Assert(attributes.Length == nextAttrIndex + 1);

            this.typeBuilderIndex[ctor.DeclaringType] = nextAttrIndex + 1;

            // Obtain the custom attribute we're interested in

            return attributes[nextAttrIndex++] as Attribute;
         }
         catch (System.Reflection.CustomAttributeFormatException)
         {
            // This means the ctor didn't match the one specified in
            // the metadata.  In this case we just keep trying with
            // the next ctor...  Hacky, I know, but there's no easy
            // way to compare a CLR and Phoenix func signature.

            // The old type builder is now corrupt so we have to create
            // a new one

            this.typeBuilder = CreateTypeBuilder();
            this.typeBuilderIndex[ctor.DeclaringType] = 0;
         }
         catch (System.TypeLoadException ex)
         {
            Helpers.WriteError("Failed to load custom attribute {0}: {1}",
               ctor.DeclaringType.FullName,
               ex.Message);
         }
         catch (System.Exception ex)
         {
            Helpers.WriteError("{0} exception occured while loading custom attribute {1}: {2}",
               ex.GetType(),
               ctor.DeclaringType.FullName,
               ex.Message);
         }

         return null;
      }

      /// <summary>
      /// Creates a TypeBuilder used for instantiating attributes
      /// </summary>

      static System.Reflection.Emit.TypeBuilder
      CreateTypeBuilder
      (
      )
      {
         // Create a simple name for the assembly

         System.Reflection.AssemblyName myAssemblyName =
            new System.Reflection.AssemblyName("EmittedAssemblyForInstantiatingAttributes");

         // Create the dynamic assembly

         System.AppDomain currentDomain = System.Threading.Thread.GetDomain();

         System.Reflection.Emit.AssemblyBuilder myAssemblyBuilder =
            currentDomain.DefineDynamicAssembly(myAssemblyName,
               System.Reflection.Emit.AssemblyBuilderAccess.Run);

         // Create the dynamic module builder

         System.Reflection.Emit.ModuleBuilder myModuleBuilder =
            myAssemblyBuilder.DefineDynamicModule("EmittedModuleForInstantiatingAttributes");

         // Create a dynamic class to use for injecting our custom attributes

         return myModuleBuilder.DefineType("CustomClassForInstantiatingAttributes");
      }
   }
}
