//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Morph/Attributes/Handlers/MorphTypeAttributeHandlerBase.cs 6     5/24/06 9:53p Eaddy $
//
// Description:
//
//    Abstract base class for extension attributes that modify
//    a module's type table. Derived classes must implement
//    MorphImpl().
//
// Remarks:
//
//    This class is somewhat complex because we support creating
//    attributes from metadata and via new.  This requires us to
//    support a two-phase initialization.
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Morph/Attributes/Handlers/MorphTypeAttributeHandlerBase.cs $
 * 
 * 6     5/24/06 9:53p Eaddy
 * o Better diagnostics when the ctor param list for the inner
 * attribute doesn't match any ctor
 * 
 * 5     3/15/06 1:20p Eaddy
 * o Added ToString()
 * 
 * 4     3/03/06 3:38p Eaddy
 * o Name change
 * 
 * 3     3/03/06 3:38p Eaddy
 * o Name change
 * 
 * 2     2/24/06 8:24p Eaddy
 * o Added checkin comment
 */

using System;

using Dbg = System.Diagnostics.Debug;

using Phx.Morph.Editors;

namespace Phx.Morph.Attributes
{
   /// <summary>
   /// Abstract base class for extension attributes that modify
   /// a module's type table
   /// </summary>
   /// <remarks>
   /// Implements AttributeBase and IModuleMorphAttribute to
   /// associate this attribute with the type we want to morph.
   /// Derived classes must implement MorphImpl().
   /// <br></br>
   /// If this attribute was created using new, you can specify
   /// TypeToExtendString which will be resolved to the correct
   /// extends attribute during Init().
   /// </remarks>

   public abstract class MorphTypeAttributeHandlerBase : AttributeHandlerBase, IModuleMorphAttribute
   {
      ExtendsAttributeHandler typeToExtend = null;
      string typeToExtendString = null;

      /// <summary>
      /// Pointer to the extends attribute that specifies which class
      /// we are extending
      /// </summary>

      public ExtendsAttributeHandler TypeToExtend
      {
         get { return this.typeToExtend; }
      }

      /// <summary>
      /// Fully qualified name of the type we are going to extend
      /// </summary>
      /// <remarks>
      /// This write-only property should only be specified if this
      /// attribute instance was created using new (as opposed to
      /// being creating from metadata).
      /// </remarks>

      public string TypeToExtendString
      {
         set
         {
            Dbg.Assert(value != null);
            Dbg.Assert(value.Length != 0);
            typeToExtendString = value;
         }
      }

      /// <summary>
      /// Called after attribute creation to specify the
      /// 'thing' (ownerSym) that this attribute is attached
      /// to and to initialize the attribute
      /// </summary>
      /// <param name="ownerSym">
      /// The Phx symbol that this attribute is attached to
      /// or null if the attribute was created via the
      /// command-line</param>
      /// <param name="globalTypeTable">
      /// Type table used to resolve the type
      /// associated with the attribute
      /// </param>
      /// <returns>true on success</returns>

      public override bool
      Init
      (
         Phx.Syms.Sym ownerSym,
         Phx.Types.Table globalTypeTable
      )
      {
         // Execute the base class initialization

         if (!base.Init(ownerSym, globalTypeTable))
         {
            return false;
         }

         // Resolve the type string so we point to our parent
         // extends attribute

         Dbg.Assert(typeToExtend == null);

         if (typeToExtendString != null)
         {
            // Handle case when attribute was created via new
            // instead of as a real custom attribute

            // Since we are creating the attribute using new,
            // there is no corresponding ExtendsAttribute lexical
            // parent attribute so we need to synthesize one.

            typeToExtend = new ExtendsAttributeHandler(typeToExtendString);

            // ISSUE-HACK-t-marcea-2005/08/11-The ExtendsAttribute
            // must be attached to a particular extension class.
            // We can give it any valid class so we'll just use the
            // class we are extending.

            typeToExtend.Attach(typeToExtendString, typeToExtendString);

            // Specify null for the owner symbol so that 
            // the typeToExtendString will be used

            if (!typeToExtend.Init(null, globalTypeTable))
            {
               return false;
            }

            // We are done with this name now

            typeToExtendString = null;
         }
         else if (AttachedType.IsAggrType)
         {
            Dbg.Assert(AttachedType.AsAggrType.TypeSym != null);

            typeToExtend =
               ExtendsAttributeHandler.GetExtendsAttribute(
                  AttachedType.AsAggrType.TypeSym.NameString);
         }
         else if (AttachedSym != null)
         {
            if (AttachedSym.IsFuncSym)
            {
               typeToExtend = ExtendsAttributeHandler.GetExtendsAttribute(
                     AttachedSym.AsFuncSym.EnclosingAggrType.TypeSym.NameString);
            }
            else if (AttachedSym.IsFieldSym)
            {
               typeToExtend = ExtendsAttributeHandler.GetExtendsAttribute(
                     AttachedSym.AsFieldSym.EnclosingAggrType.TypeSym.NameString);
            }
            else if (AttachedSym.IsPropertySym)
            {
               typeToExtend = ExtendsAttributeHandler.GetExtendsAttribute(
                     AttachedSym.AsPropertySym.EnclosingAggrType.TypeSym.NameString);
            }
            else
            {
               Dbg.Assert(false, "Unknown symbol");
            }
         }
         else
         {
            Dbg.Assert(false, "Unknown symbol");
         }

         Dbg.Assert(IsInited());

         return IsInited();
      }

      /// <returns>
      /// true if we've attached to our parent extends attribute
      /// </returns>

      public override bool
      IsInited()
      {
         if (typeToExtend == null)
            return false;
         else
            return base.IsInited() && typeToExtend.IsInited();
      }

      /// <summary>
      /// Morphs the type
      /// </summary>
      /// <param name="moduleEditor">Module where the type lives</param>
      /// <returns>true on success</returns>

      public bool
      Morph
      (
         ModuleEditor moduleEditor
      )
      {
         Dbg.Assert(IsInited());

         TypeEditor typeEditor =
            moduleEditor.FindType(this.TypeToExtend.TypeToExtendString);

         if (typeEditor == null)
         {
            Helpers.WriteError("Failed to extend type '{0}'. Type '{0}' not found in the assembly '{1}'.",
               this.TypeToExtend.TypeToExtendString,
               moduleEditor.ToString());

            return false;
         }

         // Delegate the actual morphing to the subclass

         return this.Morph(moduleEditor, typeEditor);
      }

      /// <summary>
      /// Subclasses implement this to perform the actual morphing
      /// </summary>
      /// <param name="moduleEditor">Module where the type to be morphed lives</param>
      /// <param name="typeEditor">The Type to be morphed</param>
      /// <returns>true on success</returns>

      public abstract bool
      Morph
      (
         ModuleEditor moduleEditor,
         TypeEditor typeEditor
      );

      /// <returns>
      /// Type to extend
      /// </returns>
      public override string ToString()
      {
         if (!IsInited())
         {
            return "<NotInit>";
         }
         else
         {
            Dbg.Assert(typeToExtend != null);
            return typeToExtend.ToString();
         }
      }
   }
}
