//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Morph/Editors/BinaryWeaver.cs 16    5/24/06 9:53p Eaddy $
//
// Description:
//
//    Helper class for modifying and querying a method
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Morph/Editors/BinaryWeaver.cs $
 * 
 * 16    5/24/06 9:53p Eaddy
 * o Better diagnostics when the ctor param list for the inner
 * attribute doesn't match any ctor
 * 
 * 15    5/24/06 8:56a Eaddy
 * o Changed file names but forgot to change class
 * names
 * 
 * 14    5/24/06 1:53a Eaddy
 * o Breakpoint weaver now supports [Arg]
 * 
 * 13    5/24/06 1:35a Eaddy
 * o Switched names for MethodWeaver and MethodEditor
 * so now they make sense
 * o Fixed bug where AppendFuncArg didn't work when
 * weaving static methods
 * 
 * 12    5/12/06 1:03a Rr2318
 * Wiped diagnostic writeline on line 880
 * 
 * 11    5/11/06 10:49p Rr2318
 * Modified to support general blob parsing
 * 
 * 10    5/04/06 8:44p Eaddy
 * o AppendAttributeParam now works (mostly). I'm getting
 * a "resolutionScopeSym != NULL" assertion but it still works.
 * o Moved append methods where they belong
 * 
 * 9     5/03/06 8:14p Rr2318
 * 
 * 8     5/03/06 12:51p Rr2318
 * Added Support for AnnotationAttribute
 * 
 * 7     4/03/06 12:23a Eaddy
 * o Changed jp.WithinSignature to jp.Within and made it an
 * IJoinpointSignature.
 * This is needed because we need access to return type, arg types, etc.
 * just like we do for jp.Signature.
 * 
 * 6     3/15/06 1:17p Eaddy
 * o Added verbosity flag
 * 
 * 5     3/03/06 3:34p Eaddy
 * o Name change
 * 
 * 4     3/03/06 2:58p Asv2105
 * Refactored so all classes refer to base class of MethodWeaver rather
 * than MethodEditor, and create new instances through
 * MethodWeaverFactory.
 * 
 * 3     3/01/06 2:17a Eaddy
 * o Now uses Phx.Aop library for pointcuts, advice and joinpoint matching
 * 
 * 2     2/24/06 8:27p Eaddy
 * o Added checkin comment
 */

using System;
using System.Collections.Generic;

using Dbg = System.Diagnostics.Debug;

using Phx.Aop;
using Phx.Aop.Joinpoints;
using Phx.Aop.Attributes;

namespace Phx.Morph.Editors
{

   /// <summary>
   /// Helper class for modifying and querying a method
   /// </summary>

   public class BinaryWeaver : MethodEditor
   {

      /// <summary>
      /// Create a method editor using the type editor this method
      /// belongs to and the FuncSym of the method we want to edit
      /// </summary>

      public BinaryWeaver
      (
         TypeEditor typeEditor,
         Phx.Syms.FuncSym funcSym
      )
         : base(typeEditor, funcSym)
      {
      }

      /// <summary>
      /// Create a method editor using the FuncSym of the method
      /// we want to edit
      /// </summary>

      public BinaryWeaver
      (
         Phx.Syms.FuncSym funcSym
      )
         : base(funcSym)
      {
      }

      public override bool
      Weave
      (
         Phx.IR.Instr insertAfterInstr,
         MethodEditor methodToCall,
         Phx.Aop.AdviceType adviceType,
         Phx.Aop.Joinpoints.IJoinpoint jp
      )
      {
         // Verify args

         Dbg.Assert(insertAfterInstr != null);
         Dbg.Assert(methodToCall != null);
         if (insertAfterInstr == null || methodToCall == null)
            return false;

         // Sanity checks

         Dbg.Assert(methodToCall.FuncSym != null);
         Dbg.Assert(methodToCall.FuncSym.Type != null);
         Dbg.Assert(methodToCall.FuncSym.Type.IsFuncType);

         // ISSUE-HACK-t-marcea-2005/08/04-I expected the two FuncUnit's
         // to be the same but they aren't
         //Dbg.Assert(this.funcSym.FuncUnit == insertAfterInstr.FuncUnit);

         // Keep track of our original next instruction for rollback

         Phx.IR.Instr origNextInstr = insertAfterInstr.Next;

         MethodEditor importedMethod;

         if ((adviceType == Phx.Aop.AdviceType.around &&
           origNextInstr.IsCallInstr) ||
           methodToCall.FuncSym.Type.AsFuncType.IsInstanceMethod)
         {
            // ISSUE-TODO-t-marcea-2005/08/04-We only want to import
            // a method once for this particular pointcut signature.
            // However, we are creating redundant copies.  We should
            // do a lookup before the clone.

            // Step 1:  Create a clone of the advice method for each
            //          pointcut signature

            importedMethod = methodToCall.Clone(this.typeEditor);

            // Specify that the clone is a piece of advice, not to
            // be woven (otherwise we get into an infinite loop)

            importedMethod.IsAdvice = true;

            // Since we've imported the method, we should use the
            // imported methods function signature below

            methodToCall = importedMethod;

            // Step 2:  Replace the "proceed()" call with a call to
            // the original function call we are advising

            if (adviceType == Phx.Aop.AdviceType.around &&
             origNextInstr.IsCallInstr)
            {
               Phx.IR.CallInstr origCallInstr = origNextInstr.AsCallInstr;

               MethodEditor proceedMethod = MethodEditorFactory.CreateEditor(origCallInstr.FuncSym);

               // ISSUE-TODO-t-marcea-2005/08/04-Currently, we only allow
               // around advice on static methods because we can't handle
               // the this ptr.

               if (proceedMethod.FuncSym.Type.AsFuncType.IsInstanceMethod)
               {
                  Helpers.WriteError(
                     "Error weaving '{0}': Expected static call method '{1}'.",
                     Name,
                     proceedMethod.ToString());
                  return false;
               }

               // Patch the IR to replace calls to 'proceed' with calls to
               // the original joinpoint method

               importedMethod.SetProceed(proceedMethod);

               // Rewire the instruction list to 'skip' the original call
               // instruction since we will be replacing it with a call
               // to the cloned advice method

               insertAfterInstr.Next = origNextInstr.Next;
            }
         }
         else
         {
            // Create a MethodRef for the method we're going to call

            importedMethod =
             methodToCall.CloneRef(this.typeEditor.ModuleEditor);

            // Raise so we have access to the FuncArgList and
            // the ArgFuncArgList.  We have to do this explicitly
            // since RaiseIR() only raises function units for
            // the target module and the method we are calling
            // is probably from another module.

            methodToCall.RaiseToLir();
         }

         if (importedMethod == null)
         {
            // This can happen if we try to import a method that is private

            return false;
         }

         // Create a call instruction

         Phx.IR.Instr callInstr = Phx.IR.Instr.NewCall(this.funcSym.FuncUnit,
            Phx.Targets.Archs.MSIL.Opcode.call,
            importedMethod.funcSym);
         
         //Phx.IR.Instr newobjInstr = Phx.IR.Instr.NewBinary(this.funcSym.FuncUnit,
         //   Phx.Targets.Archs.MSIL.Opcode.newobj);
         //insertAfterInstr.InsertAfter(nop);
            //)
         //);
     

         // Add the call to the IR stream
         insertAfterInstr.InsertAfter(callInstr);
         
         if (methodToCall.FuncSym.FuncType == null)
         {
            // Method we're calling has no parameters; we're done

            return true;
         }

         try
         {
            // For iterating over the types

            Phx.Types.FuncArg funcArg =
             methodToCall.FuncSym.FuncType.ArgFuncArgList;

            // For iterating over the symbols

            Phx.Collections.FuncArgList.Iterator funcSymIter =
             new Phx.Collections.FuncArgList.Iterator(methodToCall.FuncSym.FuncArgList);

            if (methodToCall.FuncSym.FuncArgList != null)
            {
               // Skip over the return value slot
               funcSymIter.MoveNext();
            }

            // Simultaneously iterate over the method call parameter
            // types and symbols

            while (funcArg != null)
            {
               string currArgName = null;
               Phx.Syms.Sym funcArgTypeSym = null;

               if (funcArg.IsThisPtr)
               {
                  currArgName = "[ThisPtr]";
               }
               else if (methodToCall.FuncSym.FuncArgList != null &&
                     funcSymIter.MoveNext())
               {
                  Phx.Syms.FuncArg currFuncArg = funcSymIter.Curr;
                  Dbg.Assert(currFuncArg != null);

                  currArgName = currFuncArg.Name.NameString;

                  funcArgTypeSym = Helpers.GetUnderlyingTypeSym(funcArg.Type);
                  Dbg.Assert(funcArgTypeSym != null);
               }

               // Check each parameter to see if it has a custom AOP
               // attribute like [Target], [Signature], etc.

               // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
               // for unmanaged implementation

               if (funcArg.IsThisPtr ||
               ParamHasAttributeOrSpecialName(typeof(ThisAttribute),
                     funcArgTypeSym, currArgName))
               {
                  AppendThisArg(funcArg, callInstr, currArgName);
               }
               else if (ParamHasAttributeOrSpecialName(typeof(AnnotationAttribute),
                  funcArgTypeSym, currArgName))
               {
                  //if (kind == Phx.Morph.Aop.PointcutKind.call)
                  //{
                  //   throw new ArgumentException(string.Format(
                  //      "Arg attribute not supported for '{0}' pointcuts.",
                  //      kind.ToString()));
                  //}

                  string indexStr = currArgName.Substring(10);
                  int index = System.Int32.Parse(indexStr);

                  System.Collections.IEnumerator e = jp.Signature.Attributes;
                  int count = -1;

                  while (e.MoveNext())
                  {
                     if (++count == index)
                        break;
                  }

                  if (count != index)
                  {
                     // Remember to subtract 1 from the arg list count for the this pointer
                     throw new ArgumentException(string.Format(
                        "Invalid [Annotation({0})] parameter '{1}' in advice method. The joinpoint only has {2} attributes.",
                        index,
                        currArgName,
                        count+1));
                  }

                  AttributeWrapper attrWrapper = e.Current as AttributeWrapper;
                  if (attrWrapper == null)
                  {
                     throw new ArgumentException("AttributeWrapper is null.");
                  }

                  //have to cast as PhxJoinpointSignature to gain access to AttrObjects
                  AppendAttributeParam(funcArg, callInstr, currArgName, attrWrapper);
               }
               else if (ParamHasAttributeOrSpecialName(typeof(ArgAttribute),
                  funcArgTypeSym, currArgName))
               {
                  //if (kind == Phx.Morph.Aop.PointcutKind.call)
                  //{ 
                  //   throw new ArgumentException(string.Format(
                  //      "Arg attribute not supported for '{0}' pointcuts.",
                  //      kind.ToString()));
                  //}

                  string indexStr = currArgName.Substring(3);
                  int index = System.Int32.Parse(indexStr);
                  string paramName = null;

                  IEnumerable<System.Text.StringBuilder> argNames =
                     jp.Signature.ArgNames;

                  IEnumerator<System.Text.StringBuilder> enumArgNames = 
                     argNames.GetEnumerator();

                  int argCount = 0;
                  while(enumArgNames.MoveNext())
                  {
                     if (argCount == index)
                     {
                        paramName = enumArgNames.Current.ToString();
                        break;
                     }

                     ++argCount;
                  }

                  if (index < 0 || index > argCount)
                  {
                     throw new ArgumentException(string.Format(
                        "Invalid advice method parameter '[Arg({0})]'. Expected [Arg(0)]-[Arg({1})].",
                        index,
                        argCount));
                  }
                  
                  // Ex:
                  //    this.funcSym = void Print(string msg)
                  //    methodToCall = void Log([Arg(0)] object arg1)
                  //
                  //    funcArg = [Phx.Types.FuncArg] "[Arg(0)] object arg1"
                  //    currArgName = "arg1"
                  //    paramName = "msg"

                  Dbg.Assert(paramName != null);

                  AppendFuncParam(funcArg, callInstr, currArgName, paramName, index);
               }
               else if (ParamHasAttributeOrSpecialName(typeof(KindAttribute),
                  funcArgTypeSym, currArgName))
               {
                  AppendArg(funcArg, callInstr, currArgName, (uint)jp.JoinpointType);
               }
               else if (ParamHasAttributeOrSpecialName(typeof(SignatureAttribute),
                  funcArgTypeSym, currArgName))
               {
                  AppendArg(funcArg, callInstr, currArgName, jp.Signature.ToString());
               }
               else if (ParamHasAttributeOrSpecialName(typeof(SignatureAttribute.DeclaringTypeNameAttribute),
                  funcArgTypeSym, currArgName))
               {
                  AppendArg(funcArg, callInstr, currArgName, jp.Signature.DeclaringType.ToString());
               }
               else if (ParamHasAttributeOrSpecialName(typeof(SignatureAttribute.NameAttribute),
                  funcArgTypeSym, currArgName))
               {
                  AppendArg(funcArg, callInstr, currArgName, jp.Signature.MemberName.ToString());
               }
               else if (ParamHasAttributeOrSpecialName(typeof(SourceLocationAttribute),
                  funcArgTypeSym, currArgName))
               {
                  AppendArg(funcArg, callInstr, currArgName, jp.Within.SourceLocation.ToString());
               }
               else if (ParamHasAttributeOrSpecialName(typeof(SourceLocationAttribute.FilePathAttribute),
                  funcArgTypeSym, currArgName))
               {
                  AppendArg(funcArg, callInstr, currArgName, jp.Within.SourceLocation.FilePath.ToString());
               }
               else if (ParamHasAttributeOrSpecialName(typeof(SourceLocationAttribute.FileNameAttribute),
                  funcArgTypeSym, currArgName))
               {
                  AppendArg(funcArg, callInstr, currArgName, jp.Within.SourceLocation.FileName);
               }
               else if (ParamHasAttributeOrSpecialName(typeof(SourceLocationAttribute.LineAttribute),
                  funcArgTypeSym, currArgName))
               {
                  AppendArg(funcArg, callInstr, currArgName, jp.Within.SourceLocation.Line);
               }
               else if (ParamHasAttributeOrSpecialName(typeof(SourceLocationAttribute.WithinSignatureAttribute),
                  funcArgTypeSym, currArgName))
               {
                  AppendArg(funcArg, callInstr, currArgName, jp.Within.ToString());
               }
               else if (ParamHasAttributeOrSpecialName(typeof(SourceLocationAttribute.WithinTypeNameAttribute),
                  funcArgTypeSym, currArgName))
               {
                  AppendArg(funcArg, callInstr, currArgName, jp.Within.DeclaringType.ToString());
               }
               else if (ParamHasAttributeOrSpecialName(typeof(SourceLocationAttribute.WithinNameAttribute),
                  funcArgTypeSym, currArgName))
               {
                  AppendArg(funcArg, callInstr, currArgName, jp.Within.MemberName.ToString());
               }
               else if (ParamHasAttributeOrSpecialName(typeof(AdviceTypeAttribute),
                  funcArgTypeSym, currArgName))
               {
                  AppendArg(funcArg, callInstr, currArgName, (uint)adviceType);
               }
               else if (ParamHasAttributeOrSpecialName(typeof(TargetAttribute),
                  funcArgTypeSym, currArgName))
               {
                  //if (kind == Phx.Morph.Aop.PointcutKind.call)
                  //{
                  //   throw new ArgumentException(string.Format(
                  //      "Target attribute not supported for '{0}' pointcuts.",
                  //       kind.ToString()));
                  //}

                  AppendThisArg(funcArg, callInstr, currArgName);
               }
               else
               {
                  //throw new ArgumentException(string.Format(
                  //   "Unexpected parameter in advice method: '{0} {1}'. Did you forget to use a parameter attribute?",
                  //   Helpers.SafeGetName(funcArg.Type),
                  //   currFuncArg.Name));

                  // ISSUE-HACK-t-marcea-2005/08/04-Temporarily allowing
                  // this tomfoolary to succeed.

                  Helpers.WriteError(
                     "Warning weaving '{0}': Error: Unexpected parameter in advice method: '{1} {2}'. Did you forget to use a parameter attribute?",
                     ToString(),
                     Helpers.SafeGetName(funcArg.Type),
                     currArgName);
               }

               funcArg = funcArg.Next;
            }
         }
         catch (System.ArgumentException ex)
         {
            Helpers.WriteError("Error weaving '{0}': {1}",
               ToString(),
               ex.Message);
            callInstr.Remove();
            insertAfterInstr.Next = origNextInstr;
            return false;
         }

         // Assign regs.

         // ISSUE-TODO-AndyA-2004/11/29 This is really pointless, but required
         // for now.  We should just rerun stackification.

         Phx.IR.Opnd opnd = callInstr.SrcOpndList;
         if (opnd != null && callInstr.CallTargetOpnd != null)
            opnd = opnd.Next;

         for (; opnd != null; opnd = opnd.Next)
         {
            opnd.Reg = Phx.Targets.Archs.MSIL.Reg.SR0;
         }

         return true;
      }

      /// <summary>
      /// Checks a function argument to see if it has a special
      /// attribute or a special name
      /// </summary>
      /// <param name="attributeType">Attribute type to check for (CLR)</param>
      /// <param name="funcArgTypeSym">TypeSym for the function argument</param>
      /// <param name="funcArgName">FuncArg for the function argument</param>
      /// <returns>
      /// true if the function argument has the special attribute or name
      /// </returns>

      public bool
      ParamHasAttributeOrSpecialName
      (
         System.Type attributeType,
         Phx.Syms.Sym funcArgTypeSym,
         string funcArgName
      )
      {
         if (!(attributeType != null &&
           funcArgName != null && funcArgName.Length != 0))
         {
            Dbg.Assert(false);
            return false;
         }

         if (funcArgTypeSym == null)
         {
            return false;
         }

         // ISSUE-HACK-t-marcea-2005/07/23-Phx#? Parameter attributes
         // don't work yet so we have a kludge that tries to match the
         // parameter name

         if (AttributeHelper.HasAttribute(funcArgTypeSym, attributeType))
         {
            return true;
         }
         else
         {
            // See if the function argument has a special name
            // 1. Must match the name of the attribute
            // 2. Have to remove digits for Args (Arg0, Arg1, etc.)
            // 3. Case doesn't matter (we specify true below)

            char[] digits = new char[] { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' };
            return String.Compare(
                  funcArgName.TrimEnd(digits) + "Attribute",
                  attributeType.Name,
                  true) == 0;
         }
      }

      /// <summary>
      /// Loads an operand and appends it to the call instr
      /// </summary>

      void
      AppendArgHelper
      (
         Phx.IR.Instr callInstr,
         Phx.IR.Instr ldArgInstr
      )
      {
         callInstr.AppendSrc(ldArgInstr.DstOpnd);

         callInstr.InsertBefore(ldArgInstr);

         // Assign regs.

         // ISSUE-TODO-AndyA-2004/11/29 This is really pointless, but required
         // for now.  We should just rerun stackification.

         ldArgInstr.DstOpnd.Reg = Phx.Targets.Archs.MSIL.Reg.SR0;

         // Explicitly break the expr temp link here since we assigned
         // registers. Can go away when we fix the issue above.

         ldArgInstr.DstOpnd.BreakExprTmp();
      }

      /// <summary>
      /// Adds a string operand to the method call
      /// </summary>
      /// <param name="funcArg">Expected argument. Used for type checking.</param>
      /// <param name="callInstr">Method call</param>
      /// <param name="argNameStr">Name of the argument. Used for diagnostics.</param>
      /// <param name="val">String to pass to the method</param>

      public void
      AppendArg
      (
         Phx.Types.FuncArg funcArg,
         Phx.IR.Instr callInstr,
         string argNameStr,
         string val
      )
      {
         if (!Helpers.IsStringType(funcArg.Type))
         {
            throw new ArgumentException(string.Format(
               "Invalid type '{0}' for parameter '{1}': Expected type '{2}'.",
               Helpers.SafeGetName(funcArg.Type),
               argNameStr,
               "string"));
         }

         Phx.PEModuleUnit moduleUnit = this.typeEditor.ModuleEditor.ModuleUnit;

         Phx.Types.Type type = moduleUnit.TypeTable.GetObjPtrType(
            moduleUnit.TypeTable.SystemStringAggrType);

         Phx.Syms.ConstSym newSym = this.typeEditor.ModuleEditor.AddUserString(val);

         Phx.IR.Opnd opnd = Phx.IR.MemOpnd.NewAddr(this.funcSym.FuncUnit,
            type, newSym, null, 0,
            Phx.Align.NaturalAlign(type),
            this.funcSym.FuncUnit.AliasInfo.NonAliasedMemTag);

         Phx.IR.Instr ldStrInstr = Phx.IR.Instr.NewUnaryExpr(this.funcSym.FuncUnit,
            Phx.Targets.Archs.MSIL.Opcode.ldstr, type, opnd);

         AppendArgHelper(callInstr, ldStrInstr);
      }

      /// <summary>
      /// Adds a uint operand to the method call
      /// </summary>
      /// <param name="funcArg">
      /// Expected argument. Used for type checking.
      /// </param>
      /// <param name="callInstr">Method call</param>
      /// <param name="argNameStr">
      /// Name of the argument. Used for diagnostics.
      /// </param>
      /// <param name="val">UInt to pass to the method</param>

      public void
      AppendArg
      (
         Phx.Types.FuncArg funcArg,
         Phx.IR.Instr callInstr,
         string argNameStr,
         uint val
      )
      {
         Phx.Types.Type typeToUse = null;

         if (Helpers.IsPrimTypeEqual(funcArg.Type,
            this.typeEditor.ModuleEditor.ModuleUnit.TypeTable.UInt32Type))
         {
            typeToUse = this.typeEditor.ModuleEditor.ModuleUnit.TypeTable.UInt32Type;
         }
         else if (funcArg.Type != null && funcArg.Type.TypeSym != null)
         {
            // Can only check IsEnum for type defs not type refs

            Phx.Types.AggrType funcArgTypeAsDef =
             Helpers.FindAggrType(this.typeEditor.ModuleEditor.ModuleUnit.TypeTable,
                  funcArg.Type.TypeSym.Name,
                  true);

            if (funcArgTypeAsDef != null && funcArgTypeAsDef.IsEnum)
            {
               typeToUse = funcArgTypeAsDef.UnderlyingPrimType;
            }
         }

         if (typeToUse == null)
         {
            throw new ArgumentException(string.Format(
               "Invalid type '{0}' for parameter '{1}': Expected type '{2}'.",
               Helpers.SafeGetName(funcArg.Type),
               argNameStr,
               "uint"));
         }

         AppendNumArg(funcArg, callInstr, typeToUse, val);
      }

      /// <summary>
      /// Adds a number (immediate) operand to the method call
      /// </summary>
      /// <param name="funcArg">
      /// Expected argument. Used for type checking.
      /// </param>
      /// <param name="callInstr">Method call</param>
      /// <param name="type">Type of the operand</param>
      /// <param name="val">Number to pass to the method</param>

      public void
      AppendNumArg
      (
         Phx.Types.FuncArg funcArg,
         Phx.IR.Instr callInstr,
         Phx.Types.Type type,
         long val
      )
      {
         //Dbg.Assert(funcArg.Type.Equals(type));

         Phx.IR.Opnd opnd = Phx.IR.ImmOpnd.NewImm(this.funcSym.FuncUnit,
            type, val);

         Phx.IR.Instr ldInstr = Phx.IR.Instr.NewUnaryExpr(this.funcSym.FuncUnit,
            Phx.Targets.Archs.MSIL.Opcode.ldc, type,
            opnd);

         AppendArgHelper(callInstr, ldInstr);
      }

      public void
      AppendPrimArg
      (
         Phx.Types.FuncArg funcArg,
         Phx.IR.Instr callInstr,
         Phx.Types.Type type,
         object val
      )
      {
         //Dbg.Assert(funcArg.Type.Equals(type));
         Phx.IR.Opnd opnd = Phx.IR.ImmOpnd.NewImm(this.funcSym.FuncUnit,
            type, 1);
         if (type.IsInt) opnd.IntValue = System.Int32.Parse(val.ToString());
         else if (type.IsUnsignedInt) opnd.UIntValue = System.UInt32.Parse(val.ToString());
         else if (type.IsFloat) opnd.FloatValue32 = (float)System.Double.Parse(val.ToString());

         Phx.IR.Instr ldInstr = Phx.IR.Instr.NewUnaryExpr(this.funcSym.FuncUnit,
            Phx.Targets.Archs.MSIL.Opcode.ldc, type,
            opnd);

         AppendArgHelper(callInstr, ldInstr);
      }
      /// <summary>
      /// Adds the this ptr as an operand to the method call
      /// </summary>
      /// <param name="funcArg">
      /// Expected argument. Used for type checking.
      /// </param>
      /// <param name="callInstr">Method call</param>
      /// <param name="argNameStr">
      /// Name of the argument. Used for diagnostics.
      /// </param>

      public void
      AppendThisArg
      (
         Phx.Types.FuncArg funcArg,
         Phx.IR.Instr callInstr,
         string argNameStr
      )
      {
         Phx.PEModuleUnit moduleUnit = this.typeEditor.ModuleEditor.ModuleUnit;

         Phx.Types.Type objectType = moduleUnit.TypeTable.GetObjPtrType(
            moduleUnit.TypeTable.SystemObjectAggrType);

         // Retrieve the "this" ptr

         Phx.Syms.LocalVarSym thisPtrSym =
           Helpers.FindThisPtrLocalVar(this.funcSym.FuncUnit.SymTable);

         if (thisPtrSym == null)
         {
            throw new ArgumentException(string.Format(
               "Invalid [This] parameter '{0}' in advice method: Not an instance method.",
               argNameStr));
         }

         // The user can specify the actual type of the this ptr
         // (e.g. Node) or object.  If they specify the actual type
         // we have to import it and verify that the this ptr is
         // that type.

         if (!Helpers.IsObjectType(funcArg.Type))
         {
            Phx.Types.Type funcArgType =
             this.typeEditor.ModuleEditor.AddTypeRef(funcArg.Type);

            Dbg.Assert(thisPtrSym.Type.IsPtrType);

            if (!funcArgType.Equals(thisPtrSym.Type.AsPtrType.ReferentType))
            {
               throw new ArgumentException(string.Format(
                  "Invalid type '{0}' for parameter '{1}': Expected type '{2}' or '{3}'.",
                  Helpers.SafeGetName(funcArg.Type),
                  argNameStr,
                  Helpers.SafeGetName(thisPtrSym.Type),
                  "object"));
            }
         }

         Phx.IR.VarOpnd thisOpnd = Phx.IR.VarOpnd.New(this.funcSym.FuncUnit,
           thisPtrSym.Type, thisPtrSym);

         Phx.IR.Instr ldThisInstr = Phx.IR.Instr.NewUnaryExpr(this.funcSym.FuncUnit,
            Phx.Targets.Archs.MSIL.Opcode.ldarg, objectType, thisOpnd);

         AppendArgHelper(callInstr, ldThisInstr);
      }

      /// <summary>
      /// Adds a local function parameter value as an operand
      /// to the method call
      /// </summary>
      /// <remarks>
      /// ISSUE-TODO-t-marcea-2005/08/03-
      /// Since we aren't able to unstackify the function calls,
      /// the only parameters we have access to are in the symbol
      /// table of the current (this) function.  We don't have
      /// access to the parameters of a called function.  That means
      /// we can support passing joinpoint Args (thisJoinpoint.Args)
      /// only for execution joinpoints.
      /// </remarks>
      /// <param name="funcArg">
      /// Expected argument. Used for type checking.
      /// </param>
      /// <param name="callInstr">Method call</param>
      /// <param name="argNameStr">
      /// Name of the argument. Used for diagnostics.
      /// </param>
      /// <param name="index"></param>

      public void
      AppendFuncParam
      (
         Phx.Types.FuncArg funcArg,
         Phx.IR.Instr callInstr,
         string argNameStr,
         string paramNameStr,
         int index
      )
      {
         Phx.Syms.LocalVarSym argSym =
           Helpers.FindLocalVarSymByName(this.funcSym.FuncUnit.SymTable, paramNameStr);

         if (argSym == null)
         {
            throw new ArgumentException(string.Format(
               "Failed to obtain context parameter '{0}' for advice method parameter '[Arg({1})] {2} {3}'.",
               paramNameStr,
               index,
               Helpers.SafeGetName(funcArg.Type),
               argNameStr));
         }

         // The user can specify the actual type of the this ptr
         // (e.g. Node) or object.  If they specify the actual type
         // we have to import it and verify that the this ptr is
         // that type.

         if (!Helpers.IsObjectType(funcArg.Type))
         {
            Phx.Types.Type funcArgType =
             this.typeEditor.ModuleEditor.AddTypeRef(funcArg.Type);

            Dbg.Assert(argSym.Type.IsPtrType);

            if (!funcArgType.Equals(argSym.Type.AsPtrType.ReferentType))
            {
               throw new ArgumentException(string.Format(
                  "Invalid type '{0}' for advice method parameter '{1}': Expected type '{2}' or '{3}'.",
                  Helpers.SafeGetName(funcArg.Type),
                  argNameStr,
                  Helpers.SafeGetName(argSym.Type),
                  "object"));
            }
         }

         Phx.PEModuleUnit moduleUnit = this.typeEditor.ModuleEditor.ModuleUnit;

         Phx.Types.Type objectType = moduleUnit.TypeTable.GetObjPtrType(
            moduleUnit.TypeTable.SystemObjectAggrType);

         Phx.IR.VarOpnd argOpnd = Phx.IR.VarOpnd.New(this.funcSym.FuncUnit,
           argSym.Type, argSym);

         Phx.IR.Instr ldArgInstr = Phx.IR.Instr.NewUnaryExpr(this.funcSym.FuncUnit,
            Phx.Targets.Archs.MSIL.Opcode.ldarg, objectType, argOpnd);

         AppendArgHelper(callInstr, ldArgInstr);
      }

      public void
      AppendAttributeParam
      (
         Phx.Types.FuncArg funcArg,
         Phx.IR.Instr callInstr,
         string argNameStr,
         AttributeWrapper attrWrapper
      )
      {
         Dbg.Assert(attrWrapper != null);

         if (attrWrapper.Attr == null)
         {
            throw new ArgumentException(string.Format("Attribute is null for parameter '{1}'.",
               argNameStr));
         }
         else if (attrWrapper.Ctor == null)
         {
            throw new ArgumentException(string.Format("Attribute constructor is null for parameter '{0}'.",
               argNameStr));
         }

         Dbg.Assert(attrWrapper.Ctor.EnclosingAggrType != null);

         // Step 1:  Import the type ref

         Phx.Types.Type importedPhxType = this.typeEditor.moduleEditor.AddTypeRef(attrWrapper.Ctor.EnclosingAggrType);
         if (importedPhxType == null)
         {
            throw new ArgumentException(string.Format("Failed to add type reference for '{0}'.",
               argNameStr));
         }

         Phx.Syms.FuncSym fs = attrWrapper.Ctor;
         if (attrWrapper.IsStatement)
         {
            MethodEditor ctorToCall = MethodEditorFactory.CreateEditor(attrWrapper.Ctor).CloneRef(this.typeEditor.ModuleEditor);
            fs = ctorToCall.FuncSym;
         }

         // Step 3:  Create the newobj instruction
         Phx.IR.Instr newInstr = Phx.IR.CallInstr.NewExpr(this.funcSym.FuncUnit,
            Phx.Targets.Archs.MSIL.Opcode.newobj, importedPhxType.TypeSym.Type, fs);//attrWrapper.Ctor);//ctorToCall.funcSym);//ctorToCall.funcSym);//attrWrapper.Ctor);//////
         Dbg.Assert(newInstr != null);
         
         // Step 4:  Insert the newobj instruction before params added to stack
         // and before the advice method's call instruction

         AppendArgHelper(callInstr, newInstr);

         // Step 5:  Push the attributes's constructor arguments onto the stack

         System.Type type = ReflectionHelpers.ConvertPhxTypeToClrType(importedPhxType);
         int k = attrWrapper.Ctor.FuncType.CountArgs();
         int i = 0;

         this.typeEditor.moduleEditor.AddTypeRef(attrWrapper.Ctor.EnclosingAggrType);

         int argsExpected = attrWrapper.Ctor.FuncType.CountArgs();

         Phx.Types.FuncArg fa = attrWrapper.Ctor.FuncType.ArgFuncArgList;

         if (argsExpected > 0 && fa.IsThisPtr)
         {
            --argsExpected;
            fa = fa.Next;
         }

         Dbg.Assert(argsExpected >= 0);

         while (fa != null)
         {
            // The first arg to the constructor is the this ptr, ignore it
            Dbg.Assert(fa.Type != null && !fa.IsThisPtr);
            if (fa.Type == null || fa.IsThisPtr)
            {
               fa = fa.Next;
               continue;
            }

            System.Type fieldClrTypeExpected = ReflectionHelpers.ConvertPhxTypeToClrType(fa.Type);
            if (fieldClrTypeExpected == null)
            {
               throw new ArgumentException(string.Format(
                  "Unknown attribute field type '{0}.{1}' for argument '{2}'.",
                  attrWrapper.Attr.GetType(),
                  fa.Type,
                  argNameStr));
            }

            if (fieldClrTypeExpected == typeof(string))
            {
               string strWeh = GetNextTokenString( attrWrapper.useArgs, 
                                                   attrWrapper.args,
                                                   attrWrapper.blob,
                                                   ref attrWrapper.blobPos);
               AppendArg(fa, newInstr, "arg" + i, strWeh);
            }
            else if (fieldClrTypeExpected.IsPrimitive)
            {
               object objWeh = GetNextToken(       fa.Type,
                                                   attrWrapper.useArgs,
                                                   attrWrapper.args,
                                                   attrWrapper.blob,
                                                   ref attrWrapper.blobPos);
               AppendPrimArg(fa, newInstr, fa.Type, objWeh);
            }
            else if (fa.Type.IsEnumType || fieldClrTypeExpected.IsEnum)
            {
               int objWeh = GetNextTokenInt(       attrWrapper.useArgs,
                                                   attrWrapper.args,
                                                   attrWrapper.blob,
                                                   ref attrWrapper.blobPos);
               AppendPrimArg(fa, newInstr, 
                  this.typeEditor.ModuleEditor.ModuleUnit.TypeTable.Int32Type, objWeh);
            }
            else
            {
               throw new ArgumentException(string.Format(
                  "Unsupported attribute '{0}' for argument '{1}'.",
                  attrWrapper.Attr.GetType(),
                  argNameStr));
            }

            fa = fa.Next;
         }

         // Assign regs.

         // ISSUE-TODO-AndyA-2004/11/29 This is really pointless, but required
         // for now.  We should just rerun stackification.

         Phx.IR.Opnd opnd = newInstr.SrcOpndList;
         if (opnd != null && newInstr.CallTargetOpnd != null)
            opnd = opnd.Next;

         for (; opnd != null; opnd = opnd.Next)
         {
            opnd.Reg = Phx.Targets.Archs.MSIL.Reg.SR0;
         }
      }
      
      /*public void
      AppendAttributeParam2
      (
         Phx.Types.FuncArg funcArg,
         Phx.IR.Instr callInstr,
         string argNameStr,
         AttributeWrapper attrWrapper
      )
      {
         Dbg.Assert(attrWrapper != null);

         if (attrWrapper.Attr == null)
         {
            throw new ArgumentException(string.Format("Attribute is null for parameter '{1}'.",
               argNameStr));
         }
         else if (attrWrapper.Ctor == null)
         {
            throw new ArgumentException(string.Format("Attribute constructor is null for parameter '{0}'.",
               argNameStr));
         }

         Dbg.Assert(attrWrapper.Ctor.EnclosingAggrType != null);

         // Step 1:  Import the type ref

         Phx.Types.Type importedPhxType = this.typeEditor.moduleEditor.AddTypeRef(attrWrapper.Ctor.EnclosingAggrType);
         if (importedPhxType == null)
         {
            throw new ArgumentException(string.Format("Failed to add type reference for '{0}'.",
               argNameStr));
         }

         // Step 2:  Import the attribute's ctor

         MethodWeaver ctorToCall = MethodWeaverFactory.newWeaver(attrWrapper.Ctor).CloneRef(this.typeEditor.ModuleEditor);

         // Step 3:  Create the newobj instruction

         Phx.IR.Instr newInstr = Phx.IR.CallInstr.NewExpr(this.funcSym.FuncUnit,
            Phx.Targets.Archs.MSIL.Opcode.newobj, importedPhxType, ctorToCall.funcSym);
         Dbg.Assert(newInstr != null);

         // Step 4:  Insert the newobj instruction before the advice method's call instruction

         AppendArgHelper(callInstr, newInstr);

         // Step 5:  Push the attributes's constructor arguments onto the stack

         // Step 5a: Get all the attribute's fields

         System.Reflection.FieldInfo[] fi = attrWrapper.Attr.GetType().GetFields(   System.Reflection.BindingFlags.Instance |
                                                                                    System.Reflection.BindingFlags.NonPublic |
                                                                                    System.Reflection.BindingFlags.Public);
         if (fi == null)
         {
            throw new ArgumentException(string.Format(
               "Failed to retrieve the fields for the attribute type '{0}' for argument '{1}'.",
               attrWrapper.Attr.GetType(),
               argNameStr));
         }

         // Step 5b: Verify that there are at least enough fields as constructor
         //          arguments

         int argsExpected = attrWrapper.Ctor.FuncType.CountArgs();

         Phx.Types.FuncArg fa = attrWrapper.Ctor.FuncType.ArgFuncArgList;

         if (argsExpected > 0 && fa.IsThisPtr)
         {
            --argsExpected;
            fa = fa.Next;
         }
         
         Dbg.Assert(argsExpected >= 0);

         if (argsExpected > fi.Length)
         {
            throw new ArgumentException(string.Format(
               "Unsupported attribute '{0}' for argument '{1}'. Expected {2} field(s).",
               attrWrapper.Attr.GetType(),
               argNameStr,
               argsExpected));
         }

         // Step 5c: Push the args onto the stack
         //
         //          We walk the ctor args link list and the FieldInfo array at the
         //          same time and push field values onto the stack.

         int index = 0;
         
         while (fa != null)
         {
            // The first arg to the constructor is the this ptr, ignore it
            Dbg.Assert(fa.Type != null && !fa.IsThisPtr);
            if (fa.Type == null || fa.IsThisPtr) 
            { 
               fa = fa.Next; 
               continue; 
            }

            // Verify that the field is the same type as the expected ctor arg

            System.Reflection.FieldInfo fiCurr = fi[index++];

            System.Type fieldClrTypeExpected = ReflectionHelpers.ConvertPhxTypeToClrType(fa.Type);
            if (fieldClrTypeExpected == null)
            {
               throw new ArgumentException(string.Format(
                  "Unknown attribute field type '{0}.{1}' for argument '{2}'.",
                  attrWrapper.Attr.GetType(),
                  fa.Type,
                  argNameStr));
            }
            else if (fieldClrTypeExpected != fiCurr.FieldType)
            {
               throw new ArgumentException(string.Format(
                  "Unsupported attribute '{0}' for argument '{1}'. Expected field '{2}' to be of type '{3}'.",
                  attrWrapper.Attr.GetType(),
                  argNameStr,
                  fiCurr.Name,
                  fieldClrTypeExpected));
            }

            // Grab the field value for this attribute instance (e.g., "blah blah")

            object fieldValue = fiCurr.GetValue(attrWrapper.Attr);

            if (fieldValue is string)
            {
               AppendArg(fa, newInstr, fiCurr.Name, (string) fieldValue);
            }
            else if (fieldValue is uint)
            {
               AppendArg(fa, newInstr, fiCurr.Name, (uint) fieldValue);
            }
            else
            {
               throw new ArgumentException(string.Format(
                  "Unsupported attribute '{0}' for argument '{1}'. Field '{2}' must be of type '{3}'.",
                  attrWrapper.Attr.GetType(),
                  argNameStr,
                  fiCurr.Name,
                  "string, uint"));
            }

            fa = fa.Next;
         }

         // Assign regs.

         // ISSUE-TODO-AndyA-2004/11/29 This is really pointless, but required
         // for now.  We should just rerun stackification.

         Phx.IR.Opnd opnd = newInstr.SrcOpndList;
         if (opnd != null && newInstr.CallTargetOpnd != null)
            opnd = opnd.Next;

         for (; opnd != null; opnd = opnd.Next)
         {
            opnd.Reg = Phx.Targets.Archs.MSIL.Reg.SR0;
         }
      } */

      public static object
      GetNextToken
      (
         bool useArgs,
         object[] args,
         ref int blobPos
      )
      {
         Dbg.Assert(useArgs == true);
         blobPos++;
         return (blobPos >= args.Length) ? null : args[blobPos - 1];
      }

      public static string
      GetNextTokenString
      (
         bool useArgs,
         object[] args,
         byte[] blob,
         ref int blobPos
      )
      {
         if (useArgs == true)
         {
            blobPos++;
            return (blobPos > args.Length) ? null : args[blobPos - 1].ToString();
         }

         blobPos++;
         int TokenLength = blob[blobPos];
         string weh = 
            System.Text.ASCIIEncoding.ASCII.GetString(blob, blobPos + 1, TokenLength);
         blobPos += TokenLength;
         return weh;
      }

      public static uint
      GetNextTokenUInt
      (
         bool useArgs,
         object[] args,
         byte[] blob,
         ref int blobPos
      )
      {
         if (useArgs == true)
         {
            blobPos++;
            return (blobPos > args.Length) ? 0 : (uint)(args[blobPos - 1]);
         }

         blobPos++; System.UInt16 i1 = System.UInt16.Parse(blob[blobPos].ToString());
         blobPos++; System.UInt16 i2 = System.UInt16.Parse(blob[blobPos].ToString());
         blobPos++; System.UInt16 i3 = System.UInt16.Parse(blob[blobPos].ToString());
         blobPos++; System.UInt16 i4 = System.UInt16.Parse(blob[blobPos].ToString());
         uint tmp = (uint)((uint)((i4 * 256 + i3) * 256 + i2) * 256 + i1);
         return tmp;
      }

      public static int
      GetNextTokenInt
      (
         bool useArgs,
         object[] args,
         byte[] blob,
         ref int blobPos
      )
      {
         if (useArgs == true)
         {
            blobPos++;
            return (blobPos > args.Length) ? 0 : (int)(args[blobPos - 1]);
         }
         blobPos++;
         System.UInt16 i1 = System.UInt16.Parse(blob[blobPos].ToString());
         blobPos++;
         System.UInt16 i2 = System.UInt16.Parse(blob[blobPos].ToString());
         blobPos++;
         System.UInt16 i3 = System.UInt16.Parse(blob[blobPos].ToString());
         blobPos++;
         System.UInt16 i4 = System.UInt16.Parse(blob[blobPos].ToString());
         int tmp = (((i4 * 256 + i3) * 256 + i2) * 256 + i1);
         return tmp;
      }

      public static object
      GetNextToken
      (
         Phx.Types.Type t,
         bool useArgs,
         object[] args,
         byte[] blob,
         ref int blobPos
      )
      {
         if (useArgs == true)
         {
            return GetNextToken(useArgs, args, ref blobPos);
         }
         if (Helpers.IsStringType(t))
         {
            return GetNextTokenString(useArgs, args, blob, ref blobPos);
         }
         else if (t.IsUnsignedInt)
         {
            return GetNextTokenUInt(useArgs, args, blob, ref blobPos);
         }
         else if (t.IsInt)
         {
            return GetNextTokenInt(useArgs, args, blob, ref blobPos);
         }
         else
         {
            return null;
         }
      }
   }
}
