//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Morph/Editors/TypeEditor.cs 10    5/26/06 8:49p Eaddy $
//
// Description:
//
//    Helper class for modifying and querying a type
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Morph/Editors/TypeEditor.cs $
 * 
 * 10    5/26/06 8:49p Eaddy
 * o Fixed null ptr exception
 * 
 * 9     5/26/06 2:01a Eaddy
 * o Added extra checking but I'm still getting a null ref ex
 * when I run the self-morph sample
 * 
 * 8     5/24/06 9:50p Eaddy
 * o Better diagnostics when the ctor param list for the inner
 * attribute doesn't match any ctor
 * 
 * 7     5/24/06 8:56a Eaddy
 * o Changed file names but forgot to change class
 * names
 * 
 * 6     5/24/06 1:35a Eaddy
 * o Switched names for MethodWeaver and MethodEditor
 * so now they make sense
 * o Fixed bug where AppendFuncArg didn't work when
 * weaving static methods
 * 
 * 5     3/15/06 1:17p Eaddy
 * o Added verbosity flag
 * 
 * 4     3/03/06 2:58p Asv2105
 * Refactored so all classes refer to base class of MethodWeaver rather
 * than MethodEditor, and create new instances through
 * MethodWeaverFactory.
 * 
 * 3     3/03/06 10:42a Eaddy
 * o Improved warning message
 * 
 * 2     2/24/06 8:27p Eaddy
 * o Fixed null exception
 * o Added checkin comment
 */

using System;

using Dbg = System.Diagnostics.Debug;

namespace Phx.Morph.Editors
{
   public enum PropKind
   {
      Getter = 1,
      Setter = 2,
      GetterAndSetter = 3
   }

   /// <summary>
   /// Helper class for modifying and querying a type
   /// </summary>

   public class TypeEditor : System.Collections.IEnumerable
   {
      public ModuleEditor moduleEditor = null;
      Phx.Types.AggrType classToExtend = null;
      int verbosity = 0;

      /// <summary>
      /// Creates a type editor given a module editor and the
      /// aggregate type we want to edit
      /// </summary>
      /// <param name="moduleEditor">Module the type belongs to</param>
      /// <param name="classToExtend">Type to edit</param>

      public TypeEditor
      (
         ModuleEditor moduleEditor,
         Phx.Types.AggrType classToExtend
      )
      {
         this.moduleEditor = moduleEditor;
         this.classToExtend = classToExtend;
      }

      /// <summary>
      /// Creates a type editor given the aggregate type we
      /// want to edit
      /// </summary>
      /// <param name="classToExtend"></param>

      public TypeEditor
      (
         Phx.Types.AggrType classToExtend
      )
         : this(new ModuleEditor(classToExtend.TypeSym.Table.Unit.AsPEModuleUnit), classToExtend)
      { }

      /// <summary>
      /// The module editor associated with this type editor
      /// </summary>

      public ModuleEditor ModuleEditor
      {
         get { return this.moduleEditor; }
      }

      /// <summary>
      /// The aggregate type we are editing
      /// </summary>

      public Phx.Types.AggrType AggrType
      {
         get { return this.classToExtend; }
      }

      /// <summary>
      /// Allows you to enumerate over the methods in this type
      /// </summary>

      public System.Collections.IEnumerator
      GetEnumerator
      (
      )
      {
         return new TypeEditorEnumerator(this);
      }

      /// <summary>
      /// Verbosity of weaving messages
      /// </summary>

      public int Verbosity
      {
         get { return verbosity; }
         set { verbosity = value; }
      }

      /// <returns>
      /// Name of the module
      /// </returns>

      public override string
      ToString()
      {
         if (this.classToExtend == null || this.classToExtend.TypeSym == null)
            return "(null)";
         else
            return this.classToExtend.TypeSym.NameString;
      }

      /// <summary>
      /// Check to see if this class has the named method
      /// </summary>
      /// <param name="funcSym">Type of the method to find</param>
      /// <returns>
      /// Method editor for the method or null if the method
      /// is not found
      /// </returns>

      public MethodEditor
      FindMethod
      (
         Phx.Syms.FuncSym funcSym
      )
      {
         Dbg.Assert(funcSym != null);
         if (funcSym == null)
            return null;

         Phx.Syms.FuncSym methodSym = Helpers.FindMethod(this.classToExtend,
            funcSym);
         return methodSym == null ? null : MethodEditorFactory.CreateEditor(this, methodSym);
      }

      /// <summary>
      /// Check to see if this type has the named field
      /// </summary>
      /// <param name="name">Name of the field to find</param>
      /// <returns>The field symbol or null if the field is not found</returns>

      public Phx.Syms.FieldSym
      FindField
      (
         Phx.Name name
      )
      {
         return Helpers.FindField(this.classToExtend, name);
      }

      public Phx.Syms.PropertySym
      FindProperty
      (
         Phx.Name name
      )
      {
         return Helpers.FindProperty(this.classToExtend, name);
      }

      /// <summary>
      /// Adds a field to this class given its CLR FieldInfo
      /// </summary>
      /// <returns>true on success</returns>

      public bool
      AddField
      (
         System.Reflection.FieldInfo fieldInfo
      )
      {
         // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
         // for unmanaged implementation

         return AddField(fieldInfo.FieldType,
                           fieldInfo.Name,
                           fieldInfo.Attributes);
      }

      /// <summary>
      /// Adds a field to this class given the CLR field Type and
      /// name
      /// </summary>
      /// <remarks>
      /// The field will have public access
      /// </remarks>
      /// <param name="fieldType">Type of the field (CLR)</param>
      /// <param name="fieldNameStr">Name of the field</param>
      /// <returns>true on success</returns>

      public bool
      AddField
      (
         System.Type fieldType,
         string fieldNameStr
      )
      {
         // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
         // for unmanaged implementation

         return AddField(fieldType,
                           fieldNameStr,
                           System.Reflection.FieldAttributes.Public);
      }

      /// <summary>
      /// Adds a field to this class given the CLR field type,
      /// name, and access
      /// </summary>
      /// <param name="fieldType">Type of the field (CLR)</param>
      /// <param name="fieldNameStr">Name of the field</param>
      /// <param name="attributes">Access for the field (CLR)</param>
      /// <returns></returns>

      public bool
      AddField
      (
         System.Type fieldType,
         string fieldNameStr,
         System.Reflection.FieldAttributes attributes
      )
      {
         // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
         // for unmanaged implementation

         Dbg.Assert(fieldType != null);
         if (fieldType == null)
            return false;

         Phx.Types.Type phxFieldType =
            ReflectionHelpers.ConvertClrTypeToPhxType(
               this.moduleEditor.ModuleUnit.TypeTable,
               fieldType);

         if (phxFieldType == null)
         {
            return false;
         }

         // Convert access rights from CLR to Phx

         Phx.Syms.Access access =
            ReflectionHelpers.ConvertClrFieldAttrToPhxAccess(attributes);

         return AddField(phxFieldType, fieldNameStr, access) != null;
      }

      /// <summary>
      /// Adds a field to this class given the Phx field from another
      /// class
      /// </summary>
      /// <param name="fieldSym">Field to add</param>
      /// <returns></returns>

      public Phx.Syms.FieldSym
      AddField
      (
         Phx.Syms.FieldSym fieldSym
      )
      {
         Dbg.Assert(this.classToExtend != null);
         Dbg.Assert(this.classToExtend.TypeSym != null);
         Dbg.Assert(fieldSym != null);

         return AddField(fieldSym.Type, fieldSym.NameString, fieldSym.Access);
      }

      /// <summary>
      /// Adds a field to this class given the Phx field type, name,
      /// and access
      /// </summary>
      /// <param name="fieldType">Type of the field (Phx)</param>
      /// <param name="fieldNameStr">Name of the field</param>
      /// <param name="fieldAccess">Access for the field (Phx)</param>
      /// <returns></returns>

      public Phx.Syms.FieldSym
      AddField
      (
         Phx.Types.Type fieldType,
         string fieldNameStr,
         Phx.Syms.Access fieldAccess
      )
      {
         Dbg.Assert(this.classToExtend != null);
         Dbg.Assert(this.classToExtend.TypeSym != null);
         Dbg.Assert(fieldNameStr != null);
         Dbg.Assert(fieldType != null);

         if (fieldNameStr.Length == 0)
         {
            Helpers.WriteError("Failed to add field to type '{1}': " +
               "Field name is empty.",
               this.classToExtend.TypeSym.NameString);
            return null;
         }

         Phx.Name fieldName = Phx.Name.New(this.moduleEditor.Lifetime, fieldNameStr);
         Dbg.Assert(fieldName != null);
         if (fieldName == null)
            return null;

         // Step 1:  Make sure the field doesn't already exist

         Phx.Syms.FieldSym findFieldSym = FindField(fieldName);
         if (findFieldSym != null)
         {
            if (findFieldSym.Type.Equals(fieldType))
            {
               Helpers.WriteOutput("Failed to add field '{0}' to type '{1}': " +
                  "Field already exists.",
                  fieldNameStr, this.classToExtend.TypeSym.NameString);
               return findFieldSym;
            }
            else
            {
               Helpers.WriteError("Failed to add field '{0}' to type '{1}': " +
                  "Field with same name but different type already exists.",
                  fieldNameStr, this.classToExtend.TypeSym.NameString);
               return null;
            }
         }

         // Step 2:  Import the assembly and type if needed

         fieldType = this.moduleEditor.AddTypeRef(fieldType);
         Dbg.Assert(fieldType != null);

         Helpers.ConvertAggregateToPrimary(ref fieldType);

         Phx.Syms.FieldSym newField = null;

         // Step 3:  Create a new symbol for the new field

         newField = Phx.Syms.FieldSym.New(this.moduleEditor.SymTable,
               0, fieldName, fieldType);

         if (newField == null)
         {
            Helpers.WriteError("Failed to add field '{0}' to type '{1}'.",
               fieldNameStr, this.classToExtend.TypeSym.NameString);
            return null;
         }

         // ISSUE-HACK-t-marcea-2005/08/02-Phx should know if NeedsFixup
         // needs to occur. Currently we have to hammer this field.

         newField.NeedsFixup = true;

         // Set the access

         newField.Access = fieldAccess;

         // Step 4:  Add the field to the type

         // ISSUE-REVIEW-t-marcea-2005/08/02-Not sure why AppendFieldSym()
         // should be called instead of AddField() and
         // InsertInLexicalScope().  I assume it because we call
         // AppendFieldSym() on type defs and the other methods on type
         // refs (I assert this below).

         Dbg.Assert(this.classToExtend.IsDef);

         this.classToExtend.AddField(newField);
         this.classToExtend.TypeSym.InsertInLexicalScope(newField,
            newField.Name);

         Helpers.WriteOutput("Added field '{0}' to type '{1}'.",
            fieldNameStr, this.classToExtend.TypeSym.NameString);

         return newField;
      }

      /// <summary>
      /// Adds a method to this class
      /// </summary>
      /// <param name="methodSymToCopy">Method to add</param>
      /// <returns>A method editor for the new method</returns>

      public MethodEditor
      AddMethod
      (
         Phx.Syms.FuncSym methodSymToCopy
      )
      {
         return AddMethod(methodSymToCopy, false, methodSymToCopy.Access);
      }

      /// <summary>
      /// Adds a method to this class
      /// </summary>
      /// <param name="methodSymToCopy">Method to add</param>
      /// <param name="isOverride">
      /// True if this method should override a virtual method
      /// from its base class (equivalent to using the C# 'overrides'
      /// keyword).
      /// </param>
      /// <param name="methodAccess">
      /// Access rights for the new method
      /// </param>
      /// <returns>A method editor for the new method</returns>

      public MethodEditor
      AddMethod
      (
         Phx.Syms.FuncSym methodSymToCopy,
         bool isOverride,
         Phx.Syms.Access methodAccess
      )
      {
         Dbg.Assert(this.classToExtend != null);
         Dbg.Assert(this.classToExtend.TypeSym != null);
         Dbg.Assert(methodSymToCopy != null);

         // Make sure the method doesn't already exist

         // ISSUE-NYI-t-marcea-2005/08/02-We currently don't
         // support method overloading

         if (Helpers.FindMethod(this.classToExtend, methodSymToCopy) != null)
         {
            Helpers.WriteOutput("Failed to add method '{0}' to type '{1}': " +
               "Method with same name already exists.",
               methodSymToCopy.NameString,
               this.classToExtend.TypeSym.NameString);
            return null;
         }

         // "Smart clone" the method

         MethodEditor methodToCopy = MethodEditorFactory.CreateEditor(methodSymToCopy);
         if (methodToCopy == null)
         {
            return null;
         }

         MethodEditor methodWeaver = methodToCopy.Clone(this, isOverride, methodAccess);

         if (methodWeaver != null)
         {
            Helpers.WriteOutput("Added field '{0}' to type '{1}'.",
               methodSymToCopy.NameString, this.classToExtend.TypeSym.NameString);
         }

         return methodWeaver;
      }

      /// <summary>
      /// Adds a method to this class
      /// </summary>
      /// <param name="methodToCopy">Method to add</param>
      /// <param name="isOverride">
      /// True if this method should override a virtual method
      /// from its base class (equivalent to using the C# 'overrides'
      /// keyword).
      /// </param>
      /// <param name="methodAccess">
      /// Access rights for the new method
      /// </param>
      /// <returns>A method editor for the new method</returns>

      public MethodEditor
      AddMethod
      (
         MethodEditor methodToCopy,
         bool isOverride,
         Phx.Syms.Access methodAccess
      )
      {
         Dbg.Assert(this.classToExtend != null);
         Dbg.Assert(this.classToExtend.TypeSym != null);
         Dbg.Assert(methodToCopy != null);

         // Make sure the method doesn't already exist

         // ISSUE-NYI-t-marcea-2005/08/02-We currently don't
         // support method overloading

         if (Helpers.FindMethod(this.classToExtend, methodToCopy.FuncSym) != null)
         {
            Helpers.WriteOutput("Failed to add method '{0}' to type '{1}': " +
               "Method with same name already exists.",
               methodToCopy.NameString,
               this.classToExtend);
            return null;
         }

         // "Smart clone" the method

         MethodEditor methodWeaver = methodToCopy.Clone(this, isOverride, methodAccess);

         if (methodWeaver != null)
         {
            Helpers.WriteOutput("Added field '{0}' to type '{1}'.",
               methodToCopy.NameString, this.classToExtend.TypeSym.NameString);
         }

         return methodWeaver;
      }

      /// <summary>
      /// Adds this class as a TypeRef to the supplied module
      /// </summary>
      /// <param name="moduleDest">Module to add ourselves to</param>
      /// <returns>Type editor for the new TypeRef</returns>

      public TypeEditor
      CloneRef
      (
         ModuleEditor moduleDest
      )
      {
         Phx.Types.Type type = moduleDest.AddTypeRef(this.classToExtend);
         if (type == null || !type.IsAggrType)
            return null;

         return new TypeEditor(moduleDest, type.AsAggrType);
      }

      /// <summary>
      /// Makes the supplied class our base class, possibly moving
      /// our original base class up in the derivation hierarchy
      /// </summary>
      /// <param name="clrBaseClassType">Base class to add</param>

      public Phx.Syms.MsilTypeSym
      InsertBaseClass
      (
         System.Type clrBaseClassType
      )
      {
         Phx.Types.Type phxBaseClassType =
            ReflectionHelpers.ConvertClrTypeToPhxType(moduleEditor.ModuleUnit.TypeTable,
               clrBaseClassType);

         if (phxBaseClassType == null)
         {
            return null;
         }

         Phx.Types.AggrType phxBaseClassAggrType =
            phxBaseClassType.AsAggrType;
         Dbg.Assert(phxBaseClassAggrType != null);

         if (phxBaseClassAggrType == null)
         {
            return null;
         }

         return InsertBaseClass(phxBaseClassAggrType);
      }

      /// <summary>
      /// Makes the supplied class our base class, possibly moving
      /// our original base class up in the derivation hierarchy
      /// </summary>
      /// <param name="phxBaseClassAggrType">Base class to add</param>

      public Phx.Syms.MsilTypeSym
      InsertBaseClass
      (
         Phx.Types.AggrType phxBaseClassAggrType
      )
      {
         return AddInterfaceOrClass(phxBaseClassAggrType, false);
      }

      /// <summary>
      /// Adds the supplied interface to this class's collection of
      /// implemented interfaces
      /// </summary>
      /// <param name="clrBaseInterfaceType">Interface to implement</param>

      public Phx.Syms.MsilTypeSym
      AddInterface
      (
         System.Type clrBaseInterfaceType
      )
      {
         Phx.Types.Type phxBaseInterfaceType =
            ReflectionHelpers.ConvertClrTypeToPhxType(moduleEditor.ModuleUnit.TypeTable,
               clrBaseInterfaceType);

         if (phxBaseInterfaceType == null)
         {
            return null;
         }

         Phx.Types.AggrType phxBaseInterfaceAggrType =
            phxBaseInterfaceType.AsAggrType;
         Dbg.Assert(phxBaseInterfaceAggrType != null);

         if (phxBaseInterfaceAggrType == null)
         {
            return null;
         }

         return AddInterface(phxBaseInterfaceAggrType);
      }

      /// <summary>
      /// Adds the supplied interface to this class's collection of
      /// implemented interfaces
      /// </summary>
      /// <remarks>Not Yet Implemented</remarks>
      /// <param name="phxBaseInterfaceType">Interface to implement</param>

      public Phx.Syms.MsilTypeSym
      AddInterface
      (
         Phx.Types.AggrType phxBaseInterfaceType
      )
      {
         return AddInterfaceOrClass(phxBaseInterfaceType, true);
      }

      /// <summary>
      /// Add interface to implement or class to extend
      /// </summary>
      /// <param name="phxBaseAggrType">
      /// Interface to implement or class to extend</param>
      /// <param name="isInterface">Is this an interface or a class?</param>

      Phx.Syms.MsilTypeSym
      AddInterfaceOrClass
      (
         Phx.Types.AggrType phxBaseAggrType,
         bool isInterface
      )
      {
         if (phxBaseAggrType == null)
         {
            return null;
         }

         phxBaseAggrType =
            this.moduleEditor.AddTypeRef(phxBaseAggrType)
            as Phx.Types.AggrType;

         Phx.Types.BaseTypeLink baseTypeLink =
            Phx.Types.BaseTypeLink.New(phxBaseAggrType, isInterface);

         return AddBase(baseTypeLink);
      }

      /// <summary>
      /// Add interface to implement or class to extend
      /// </summary>
      /// <param name="baseTypeLink">
      /// Interface to implement or class to extend</param>

      public Phx.Syms.MsilTypeSym
      AddBase
      (
         Phx.Types.BaseTypeLink baseTypeLink
      )
      {
         if (baseTypeLink == null)
         {
            return null;
         }

         Dbg.Assert(baseTypeLink.BaseAggrType != null);
         Dbg.Assert(baseTypeLink.BaseAggrType.TypeSym != null);

         // We assume all classes already derive from System.Object

         if (Helpers.IsObjectType(baseTypeLink.BaseAggrType))
         {
            return null;
         }

         // Handle case when the extension subclasses the class its
         // extending (us)

         Dbg.Assert(this.AggrType != null);
         Dbg.Assert(this.AggrType.TypeSym != null);

         if (Helpers.IsMsilTypeSymEqual(this.AggrType.TypeSym.AsMsilTypeSym, 
            baseTypeLink.BaseAggrType.TypeSym.AsMsilTypeSym))
         {
            return this.AggrType.TypeSym.AsMsilTypeSym;
         }

         Phx.Types.AggrType phxBaseAggrType =
            this.moduleEditor.AddTypeRef(baseTypeLink.BaseAggrType)
            as Phx.Types.AggrType;

         Phx.Types.BaseTypeLink baseType = this.AggrType.BaseTypeLinkList;
         while (baseType != null)
         {
            Dbg.Assert(baseType.BaseAggrType != null);
            if (baseType.BaseAggrType.Equals(phxBaseAggrType))
            {
               Dbg.Assert(baseType.BaseAggrType.TypeSym != null);
               return baseType.BaseAggrType.TypeSym.AsMsilTypeSym;
            }

            baseType = baseType.Next;
         }

         Phx.Types.BaseTypeLink newBaseTypeLink =
            Phx.Types.BaseTypeLink.New(phxBaseAggrType, baseTypeLink.IsVirtual);

         Dbg.Assert(newBaseTypeLink != null);

         this.classToExtend.AppendBaseTypeLink(newBaseTypeLink);

         Dbg.Assert(newBaseTypeLink.BaseAggrType != null);
         Dbg.Assert(newBaseTypeLink.BaseAggrType.TypeSym != null);
         Dbg.Assert(newBaseTypeLink.BaseAggrType.TypeSym.IsMsilTypeSym);

         Helpers.WriteOutput("Added base '{0}' to type '{1}'.",
            newBaseTypeLink.BaseAggrType.TypeSym.AsMsilTypeSym.NameString, 
            this.classToExtend.TypeSym.NameString);

         return newBaseTypeLink.BaseAggrType.TypeSym.AsMsilTypeSym;
      }

      /// <summary>
      /// Adds an attribute to a member of the class
      /// </summary>
      /// <remarks>Not Yet Implemented</remarks>
      /// <param name="symToExtend">Symbol to add the attribute to</param>
      /// <param name="attrType">Type of the attribute</param>
      /// <param name="ctorArgs">Initialization data for the attribute</param>

      public Phx.Syms.AttrSym
      AddAttribute
      (
         Phx.Syms.Sym symToExtend,
         Phx.Types.Type attrType,
         object[] ctorArgs
      )
      {
         throw new System.Exception("The method or operation is not implemented.");
      }

      // <remarks>Not Yet Implemented</remarks>

      /// <summary>
      /// Adds an Event to this class
      /// </summary>
      /// <remarks>Not Yet Implemented</remarks>
      /// <param name="delegateType">Delegate type for the event</param>
      /// <param name="eventName">Name of the event</param>
      /// <param name="eventAccess">Event access modifier</param>

      public Phx.Syms.EventSym
      AddEvent
      (
         Phx.Types.Type delegateType,
         string eventName,
         Phx.Syms.Access eventAccess
      )
      {
          throw new System.Exception("The method or operation is not implemented.");
      }

      /// <summary>
      /// Adds a Property to this class
      /// </summary>
      /// <param name="phxPropSym">Property to add</param>
      /// <param name="propAccess">Access modifier</param>
      /// <returns></returns>

      public Phx.Syms.PropertySym
      AddProperty
      (
         Phx.Syms.PropertySym phxPropSym,
         Phx.Syms.Access propAccess)
      {
         if (phxPropSym == null)
            return null;

         Phx.Syms.PropertySym findPropSym = FindProperty(phxPropSym.Name);
         if (findPropSym != null && findPropSym.Type != null)
         {
            if (findPropSym.Type.Equals(phxPropSym.Type))
            {
               Helpers.WriteOutput("Failed to add property '{0}' to type '{1}': " +
                  "Property already exists.",
                  findPropSym.NameString,
                  this.classToExtend.TypeSym.NameString);
               return findPropSym;
            }
            else
            {
               Helpers.WriteError("Failed to add property '{0}' to type '{1}': " +
                  "Property with same name but different type already exists.",
                  findPropSym.NameString,
                  this.classToExtend.TypeSym.NameString);
               return null;
            }
         }

         Phx.Name propName = Phx.Name.New(this.moduleEditor.Lifetime,
            phxPropSym.Name);
         Dbg.Assert(propName != null);
         if (propName == null)
            return null;

         Phx.Syms.PropertySym newPropSym = null;

         // Step 3:  Create a symbol for the new property

         newPropSym = Phx.Syms.PropertySym.New(this.moduleEditor.SymTable,
               propName);

         if (newPropSym == null)
         {
            Helpers.WriteError("Failed to add property '{0}' to type '{1}'.",
               phxPropSym.NameString, this.classToExtend.TypeSym.NameString);
            return null;
         }

         // Step 4:  Add the property to the class

         Dbg.Assert(this.classToExtend.IsDef);

         this.classToExtend.AddPropertySym(newPropSym);

         this.classToExtend.TypeSym.InsertInLexicalScope(newPropSym,
            newPropSym.Name);

         Helpers.WriteOutput("Added property '{0}' to type '{1}'.",
            phxPropSym.NameString,
            this.classToExtend.TypeSym.NameString);

         // Step 4:  Create the Property getter and setter functions
         //
         //          This also associates the getter/setter with the
         //          PropertySym, and the PropertySym with the
         //          getter/setter.

         if (phxPropSym.FuncSymGetter != null)
         {
            MethodEditor newGetterMethod = AddMethod(phxPropSym.FuncSymGetter,
               false, propAccess);
            if (newGetterMethod == null)
               return null;
            else
            {
               Helpers.WriteOutput("Added property getter 'get_{0}' to type '{1}'.",
                  phxPropSym.NameString,
                  this.classToExtend.TypeSym.NameString);
            }
         }

         if (phxPropSym.FuncSymSetter != null)
         {
            MethodEditor newSetterMethod = AddMethod(phxPropSym.FuncSymSetter,
               false, propAccess);
            if (newSetterMethod == null)
               return null;
            else
            {
               Helpers.WriteOutput("Added property getter 'get_{0}' to type '{1}'.",
                  phxPropSym.NameString,
                  this.classToExtend.TypeSym.NameString);
            }
         }

         return newPropSym;
      }
   }

   public class TypeEditorEnumerator : System.Collections.IEnumerator
   {
      Phx.Syms.FuncSym first = null;
      Phx.Syms.FuncSym curr = null;
      TypeEditor typeEditor = null;

      public TypeEditorEnumerator(TypeEditor typeEditor)
      {
         this.typeEditor = typeEditor;

         if (typeEditor != null &&
            typeEditor.AggrType != null)
         {
            first = typeEditor.AggrType.FuncSymList;
         }
      }

      public object Current
      {
         get
         {
            if (curr != null)
            {
               Dbg.Assert(curr.IsFuncSym);
               return MethodEditorFactory.CreateEditor(typeEditor, curr);
            }
            else
            {
               return null;
            }
         }
      }

      public bool MoveNext()
      {
         if (curr == null)
            curr = first;
         else
            curr = curr.NextTypeFuncSym;

         return curr != null;
      }

      public void Reset()
      {
         curr = null;
      }
   }
}
