//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Morph/Helpers.cs 12    5/27/06 1:26a Eaddy $
//
// Description:
//
//    Helper functions
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Morph/Helpers.cs $
 * 
 * 12    5/27/06 1:26a Eaddy
 * o Default verbosity now shows errors and warnings
 * 
 * 11    5/24/06 10:18p Eaddy
 * o Verbosity of 0 now surpresses all messages
 * 
 * 10    5/24/06 9:56p Eaddy
 * o Now tries to resolve assemblies that can't be loaded
 * o Improved instantiation of custom attributes and finding
 * the correct ctor for a given blob
 * 
 * 9     5/24/06 1:35a Eaddy
 * o Switched names for MethodWeaver and MethodEditor
 * so now they make sense
 * o Fixed bug where AppendFuncArg didn't work when
 * weaving static methods
 * 
 * 8     5/12/06 1:24a Eaddy
 * o Fixed bug where we had a double ptr
 * 
 * 7     5/04/06 5:08p Eaddy
 * o Put back Rajesh's fix
 * 
 * 6     5/04/06 3:32p Eaddy
 * o Joinpoints now have source column so that we can weave
 * the same advice multiple times at the same joinpoint line 
 * 
 * 5     3/03/06 3:00p Eaddy
 * 
 * 4     3/01/06 2:17a Eaddy
 * o Now uses Phx.Aop library for pointcuts, advice and joinpoint matching
 * 
 * 3     2/27/06 5:24p Rr2318
 * Fixed line number issues
 * 
 * 2     2/24/06 8:27p Eaddy
 * o Added checkin comment
 */

using System;
using System.Text;

using System.Collections.Generic;

using Dbg = System.Diagnostics.Debug;

namespace Phx.Morph
{
   /// <summary>
   /// Helper functions
   /// </summary>

   public class Helpers
   {
      static int verbosity = 2; // Errors and warnings only
      static Dictionary<string, string> mapFQN2keyword;

      static Helpers()
      {
         mapFQN2keyword = new Dictionary<string, string>();
         mapFQN2keyword[typeof(string).FullName] = "string";
         mapFQN2keyword[typeof(object).FullName] = "object";
         mapFQN2keyword[typeof(short).FullName] = "short";
         mapFQN2keyword[typeof(ushort).FullName] = "ushort";
         mapFQN2keyword[typeof(int).FullName] = "int";
         mapFQN2keyword[typeof(uint).FullName] = "uint";
         mapFQN2keyword[typeof(long).FullName] = "long";
         mapFQN2keyword[typeof(ulong).FullName] = "ulong";
         mapFQN2keyword[typeof(char).FullName] = "char";
         mapFQN2keyword[typeof(bool).FullName] = "bool";
         mapFQN2keyword[typeof(float).FullName] = "float";
         mapFQN2keyword[typeof(double).FullName] = "double";
         mapFQN2keyword[typeof(byte).FullName] = "byte";
         mapFQN2keyword[typeof(sbyte).FullName] = "sbyte";
         mapFQN2keyword[typeof(decimal).FullName] = "decimal";
      }

      /// <summary>
      /// So no one accidentally creates one
      /// </summary>
      internal Helpers()
      { } 

      /// <summary>
      /// Verbosity of weaving messages
      /// </summary>

      public static int Verbosity
      {
         get { return verbosity; }
         set { verbosity = value; }
      }

      /// <summary>
      /// Returns the symbol table's name map, creating it if necessary
      /// </summary>
      /// <param name="symTable">Symbol table</param>
      /// <returns>Table's name map</returns>

      static public Phx.Syms.NameMap
      GetNameMap
      (
         Phx.Syms.Table symTable
      )
      {
         Dbg.Assert(symTable != null);
         if (symTable == null)
         {
            return null;
         }

         Phx.Syms.NameMap nameMap = symTable.NameMap;

         // If the module was not compiled with debug, the module sym
         // table may not have a name map.

         if (nameMap == null)
         {
            nameMap = Phx.Syms.NameMap.New(symTable, 64);
            symTable.AddMap(nameMap);
         }

         return nameMap;
      }

      /// <summary>
      /// Find a symbol in the symbol table by name
      /// </summary>
      /// <param name="symTable">Symbol table to search</param>
      /// <param name="name">Name of the symbol to find</param>
      /// <returns>The symbol or null if the symbol is not found</returns>

      static public Phx.Syms.SectSym
      FindSectSym
      (
         Phx.Syms.Table symTable,
         Phx.Name name
      )
      {
         Dbg.Assert(symTable != null);
         if (symTable == null)
            return null;

         return GetNameMap(symTable).Lookup(name, Phx.Syms.SymKind.Sect) as
            Phx.Syms.SectSym;
      }

      /// <summary>
      /// Find a symbol in the symbol table by name
      /// </summary>
      /// <param name="symTable">Symbol table to search</param>
      /// <param name="name">Name of the symbol to find</param>
      /// <returns>The symbol or null if the symbol is not found</returns>

      static public Phx.Syms.SectSym
      FindSectSym
      (
         Phx.Syms.Table symTable,
         string name
      )
      {
         return FindSectSym(symTable, Phx.Name.New(Phx.GlobalData.GlobalLifetime, name));
      }

      /// <summary>
      /// Find a MSIL Type symbol in the symbol table by name
      /// </summary>
      /// <param name="symTable">Symbol table to search</param>
      /// <param name="name">Name of the symbol to find</param>
      /// <returns>The symbol or null if the symbol is not found</returns>

      static public Phx.Syms.MsilTypeSym
      FindMsilTypeSym
      (
         Phx.Syms.Table symTable,
         Phx.Name name
      )
      {
         // Verify args

         Dbg.Assert(symTable != null);
         if (symTable == null)
            return null;

         if (name.IsNull)
         {
            // Not a user type (an internal Phx type)

            return null;
         }

         // ISSUE-BUG-t-marcea-2005/08/05-Lookup() has AV when name.IsNull

         return GetNameMap(symTable).Lookup(name, Phx.Syms.SymKind.MsilType) as
            Phx.Syms.MsilTypeSym;
      }

      /// <summary>
      /// Find a MSIL Type symbol in the symbol table by name
      /// </summary>
      /// <param name="symTable">Symbol table to search</param>
      /// <param name="name">Name of the symbol to find</param>
      /// <returns>The symbol or null if the symbol is not found</returns>

      static public Phx.Syms.MsilTypeSym
      FindMsilTypeSym
      (
         Phx.Syms.Table symTable,
         string name
      )
      {
         return FindMsilTypeSym(symTable, Phx.Name.New(Phx.GlobalData.GlobalLifetime, name));
      }

      /// <summary>
      /// Find a Local Var symbol in the symbol table by name
      /// </summary>
      /// <param name="symTable">Symbol table to search</param>
      /// <param name="name">Name of the symbol to find</param>
      /// <returns>The symbol or null if the symbol is not found</returns>

      static public Phx.Syms.LocalVarSym
      FindLocalVarSymByName
      (
         Phx.Syms.Table symTable,
         Phx.Name name
      )
      {
         // Verify args

         Dbg.Assert(symTable != null);
         if (symTable == null)
            return null;

         return GetNameMap(symTable).Lookup(name, Phx.Syms.SymKind.LocalVar) as
            Phx.Syms.LocalVarSym;
      }

      /// <summary>
      /// Find a MSIL Type symbol in the symbol table by name
      /// </summary>
      /// <param name="symTable">Symbol table to search</param>
      /// <param name="name">Name of the symbol to find</param>
      /// <returns>The symbol or null if the symbol is not found</returns>

      static public Phx.Syms.LocalVarSym
      FindLocalVarSymByName
      (
         Phx.Syms.Table symTable,
         string name
      )
      {
         return FindLocalVarSymByName(symTable, Phx.Name.New(Phx.GlobalData.GlobalLifetime, name));
      }

      /// <summary>
      /// Find a Local Var symbol in the symbol table by index
      /// </summary>
      /// <param name="symTable">Symbol table to search</param>
      /// <param name="argIndex">Index of the arg starting from 0</param>
      /// <returns>The symbol or null if the symbol is not found</returns>

      static public Phx.Syms.LocalVarSym
      FindLocalVarSymByIndex
      (
         Phx.Syms.Table symTable,
         int argIndex
      )
      {
         // Verify args

         Dbg.Assert(symTable != null && argIndex >= 0);
         if (symTable == null || argIndex < 0)
            return null;

         // Sanity checks

         Dbg.Assert(symTable.LocalIdMap != null);
         if (symTable.LocalIdMap == null)
            return null;

         // ISSUE-HACK-t-marcea-2005/08/02-We assume that
         // ExternId corresponds to the order in the argument
         // list.

         return symTable.LocalIdMap.Lookup((uint)argIndex) as
            Phx.Syms.LocalVarSym;
      }

      static public Phx.Syms.LocalVarSym
      FindThisPtrLocalVar
      (
         Phx.Syms.Table symTable
      )
      {
         // Verify args

         Dbg.Assert(symTable != null);
         if (symTable == null)
            return null;

         // Sanity checks

         Dbg.Assert(symTable != null);
         Dbg.Assert(symTable.LocalIdMap != null);
         Dbg.Assert(symTable.LocalIdMap.InternalMap != null);
         if (symTable == null || symTable.LocalIdMap == null ||
            symTable.LocalIdMap.InternalMap == null)
         {
            return null;
         }

         foreach (Phx.Syms.Sym sym in symTable.LocalIdMap.InternalMap)
         {
            Phx.Syms.LocalVarSym localVar = sym.AsLocalVarSym;

            if (localVar != null && localVar.IsThisParam)
            {
               return localVar;
            }
         }

         return null;
      }

      /// <summary>
      /// Find a user string (Const Symbol) in the symbol table by value
      /// </summary>
      /// <param name="symTable">Symbol table to search</param>
      /// <param name="strVal">String value to find</param>
      /// <returns>The symbol or null if the symbol is not found</returns>

      static public Phx.Syms.ConstSym
      FindUserString
      (
         Phx.Syms.Table symTable,
         string strVal
      )
      {
         // Verify args

         Dbg.Assert(symTable != null && strVal != null);
         if (symTable == null || strVal == null)
            return null;

         // Sanity checks

         Dbg.Assert(symTable != null);
         Dbg.Assert(symTable.LocalIdMap != null);
         Dbg.Assert(symTable.LocalIdMap.InternalMap != null);
         if (symTable == null || symTable.LocalIdMap == null ||
            symTable.LocalIdMap.InternalMap == null)
         {
            return null;
         }

         foreach (Phx.Syms.Sym sym in symTable.LocalIdMap.InternalMap)
         {
            if (sym.IsConstSym && IsConstSymEqual(sym.AsConstSym, strVal))
            {
               return sym.AsConstSym;
            }
         }

         return null;
      }

      static public Phx.Syms.ConstSym
      FindOrCreateUserString
      (
         Phx.Syms.Table symTable,
         Phx.PEModuleUnit moduleUnit,
         string strVal
      )
      {
         // Verify args

         Dbg.Assert(symTable != null && strVal != null);
         if (symTable == null || strVal == null)
            return null;

         Phx.Name nameString = Phx.Name.New(Phx.GlobalData.GlobalLifetime, strVal);
         Phx.Syms.ConstSym symString = GetNameMap(symTable).Lookup(nameString, Phx.Syms.SymKind.Const)
            as Phx.Syms.ConstSym;
         if (symString != null)
            return symString;

         symString = Phx.Syms.ConstSym.New(  symTable,
                                             0,
                                             nameString,
                                             moduleUnit.TypeTable.SystemStringAggrType,
                                             strVal);
         return symString;
      }

      /// <summary>
      /// Find a Const Symbol in the symbol table by value
      /// </summary>
      /// <param name="symTable">Symbol table to search</param>
      /// <param name="constSymVal">Const Symbol to compare to</param>
      /// <returns>The symbol or null if the symbol is not found</returns>

      static public Phx.Syms.ConstSym
      FindConstSym
      (
         Phx.Syms.Table symTable,
         Phx.Syms.ConstSym constSymVal
      )
      {
         // Verify args

         Dbg.Assert(symTable != null && constSymVal != null);
         if (symTable == null || constSymVal == null)
            return null;

         // Sanity checks

         Dbg.Assert(symTable != null);
         Dbg.Assert(symTable.LocalIdMap != null);
         Dbg.Assert(symTable.LocalIdMap.InternalMap != null);
         if (symTable == null || symTable.LocalIdMap == null ||
            symTable.LocalIdMap.InternalMap == null)
         {
            return null;
         }

         foreach (Phx.Syms.Sym sym in symTable.LocalIdMap.InternalMap)
         {
            if (sym.IsConstSym && IsConstSymEqual(sym.AsConstSym, constSymVal))
            {
               return sym.AsConstSym;
            }
         }

         return null;
      }

      /// <summary>
      /// Check to see if the type has the named field
      /// </summary>
      /// <remarks>
      /// ISSUE-REVIEW-t-marcea-2005/08/04-Why isn't there
      /// an AggrType.FindField(Name)?
      /// </remarks>
      /// <param name="aggrType">Aggregate Type to check</param>
      /// <param name="name">Name of the field to find</param>
      /// <returns>The field symbol or null if the field is not found</returns>

      static public Phx.Syms.FieldSym
      FindField
      (
         Phx.Types.AggrType aggrType,
         Phx.Name name
      )
      {
         Dbg.Assert(aggrType != null);
         if (aggrType == null || aggrType.TypeSym == null)
            return null;

         return aggrType.TypeSym.LookupInLexicalScope(name) as Phx.Syms.FieldSym;
      }

      /// <summary>
      /// Check to see if the type has the named field
      /// </summary>
      /// <remarks>
      /// ISSUE-REVIEW-t-marcea-2005/08/04-Why isn't there
      /// an AggrType.FindField(Name)?
      /// </remarks>
      /// <param name="aggrType">Aggregate Type to check</param>
      /// <param name="name">Name of the field to find</param>
      /// <returns>The field symbol or null if the field is not found</returns>

      static public Phx.Syms.PropertySym
      FindProperty
      (
         Phx.Types.AggrType aggrType,
         Phx.Name name
      )
      {
         Dbg.Assert(aggrType != null);
         if (aggrType == null || aggrType.TypeSym == null)
            return null;

         return aggrType.TypeSym.LookupInLexicalScope(name) as Phx.Syms.PropertySym;
      }

      /// <summary>
      /// Check to see if the type has the named method
      /// </summary>
      /// <remarks>
      /// ISSUE-TODO-t-marcea-2005/08/04-We need to handle overloaded
      /// functions.  Name is not good enough; we also need the type.
      /// </remarks>
      /// <param name="aggrType">Aggregate Type to check</param>
      /// <param name="funcSym">Symbol of the method to find</param>
      /// <returns>The function symbol or null if the function is not found</returns>

      static public Phx.Syms.FuncSym
      FindMethod
      (
         Phx.Types.AggrType aggrType,
         Phx.Syms.FuncSym funcSym
      )
      {
         Dbg.Assert(aggrType != null);
         if (aggrType == null || aggrType.FuncSymList == null)
            return null;

         Phx.Syms.FuncSym curr = aggrType.FuncSymList;
         while (curr != null)
         {
            if (EqualSignaturesIgnoringThisAndPtrs(curr, funcSym))
            {
               return curr;
            }

            curr = curr.NextTypeFuncSym;
         }

         return null;
      }

      static public bool
      EqualSignaturesIgnoringThisAndPtrs
      (
         Phx.Syms.FuncSym firstFuncSym,
         Phx.Syms.FuncSym secondFuncSym
      )
      {
         // ISSUE-REVIEW-MDrinic-2003/12/18 Signatures can be equal even if FuncTypes
         // differ since some languages allow generalization of parameters and
         // specialization of return types. Currently, there is no suport for such
         // cases.

         bool isEqual =
            Phx.Name.Equals(firstFuncSym.Name, secondFuncSym.Name);

         if (!isEqual)
         {
            return false;
         }

         Phx.Types.FuncType firstFuncType = firstFuncSym.FuncType;
         Phx.Types.FuncType secondFuncType = secondFuncSym.FuncType;

         if (firstFuncType.IsInstanceMethod && secondFuncType.IsInstanceMethod)
         {
            // ThisArg does not have to be equal.

            if (firstFuncType.CallingConventionKind !=
               secondFuncType.CallingConventionKind)
            {
               return false;
            }

            Phx.Types.FuncArg firstFuncArgList = firstFuncType.ArgFuncArgList;
            Phx.Types.FuncArg secondFuncArgList = secondFuncType.ArgFuncArgList;

            while ((firstFuncArgList != null) && (secondFuncArgList != null))
            {
               if (!(firstFuncArgList.IsThisPtr && secondFuncArgList.IsThisPtr))
               {
                  if (!Helpers.IsArgEqualIgnorePtrs(firstFuncArgList, secondFuncArgList))
                  {
                     return false;
                  }
               }

               firstFuncArgList = firstFuncArgList.Next;
               secondFuncArgList = secondFuncArgList.Next;
            }

            // ISSUE-HACK-t-marcea-2005/08/15-Phx.Types.FuncType.IsRetListEqual()
            // only equates two MSIL types if their MSIL handles are equal (and
            // not null)

            //if (!Phx.Types.FuncType.IsRetListEqual(firstFuncType, secondFuncType,
            //      false))

            if (!Helpers.IsArgEqualIgnorePtrs(firstFuncType.RetFuncArgList,
               secondFuncType.RetFuncArgList))
            {
               return false;
            }

            return true;
         }

         return firstFuncSym.FuncType.Equals(secondFuncSym.FuncType);
      }

      /// <summary>
      /// Looks for a type by TypeSym name in the module unit's type table
      /// </summary>
      /// <param name="typeTable">Type table to search</param>
      /// <param name="typeName">Name of the TypeSym to look for</param>
      /// <param name="mustBeDef">true if the TypeSym must be a definition, 
      /// false if it can be a definition or a reference</param>
      /// <returns>The Aggregate Type or null if the TypeSym is not found</returns>

      static public Phx.Types.AggrType
      FindAggrType
      (
         Phx.Types.Table typeTable,
         Phx.Name typeName,
         bool mustBeDef
      )
      {
         Dbg.Assert(typeTable != null);
         Dbg.Assert(typeTable.AllTypes != null);

         // ISSUE-HACK-t-marcea-2005/08/12-Check the builtin symbol table
         // first.  A backdoor to get access to this is to access the symbol
         // table of a known builtin.  This is a hack!

         Phx.Syms.TypeSym typeSym =
            typeTable.SystemObjectAggrType.TypeSym.Table.LookupByName(typeName)
            as Phx.Syms.TypeSym;

         // ISSUE-HACK-t-marcea-2005/08/17-For some reason, on ChuckM's
         // machine I was finding a typeref for IEnumerator in the builtin
         // symbol table.  So now, if I require a TypeDef, I ignore
         // the TypeRef.  We can remove this once the builtin symbol
         // table works correctly.

         if (typeSym != null && 
            typeSym.Type != null && 
            typeSym.Type.IsAggrType &&
            (!mustBeDef || typeSym.Type.AsAggrType.IsDef))
         {
            return typeSym.Type.AsAggrType;
         }

         // Now check the global type table

         for (Phx.Types.Type currType = typeTable.AllTypes;
               currType != null;
               currType = currType.Next)
         {
            if (!currType.IsAggrType)
            {
               continue;
            }

            if (currType.TypeSym == null)
            {
               // Instantiations of generic types do not have
               // a TypeSym

               Dbg.Assert(currType.AsAggrType.IsInstantiation ||
                  currType.IsNonSelfDescribingAggr);
               continue;
            }

            if (Phx.Name.Equals(currType.TypeSym.Name, typeName))
            {
               if (mustBeDef && !currType.AsAggrType.IsDef)
                  continue;

               return currType.AsAggrType;
            }
         }

         return null;
      }

      /// <summary>
      /// Looks for a type by TypeSym name in the module unit's type table
      /// </summary>
      /// <param name="typeTable">Type table to search</param>
      /// <param name="typeName">Name of the TypeSym to look for</param>
      /// <param name="mustBeDef">true if the TypeSym must be a definition, 
      /// false if it can be a definition or a reference</param>
      /// <returns>The Aggregate Type or null if the TypeSym is not found</returns>

      static public Phx.Types.AggrType
      FindAggrType
      (
         Phx.Types.Table typeTable,
         string typeName,
         bool mustBeDef
      )
      {
         return FindAggrType(typeTable,
            Phx.Name.New(Phx.GlobalData.GlobalLifetime, typeName),
            mustBeDef);
      }

      static public bool
      IsArgEqualIgnorePtrs
      (
         Phx.Types.FuncArg lhs,
         Phx.Types.FuncArg rhs
      )
      {
         // Let the "offical" IsEqual() method have a chance
         // (we throw away negative results since they are not reliable)

         if (Phx.Types.FuncArg.IsEqual(lhs, rhs, false))
         {
            return true;
         }
         else if (lhs == null || rhs == null)
         {
            return false;
         }
         else if (Object.ReferenceEquals(lhs, rhs))
         {
            return true;
         }
         else if (lhs.FuncArgKind != rhs.FuncArgKind)
         {
            return false;
         }

         // Vararg functions have no type field.

         if (lhs.IsAnEllipsis)
         {
            return true;
         }

         Phx.Types.Type lhsTypeToCmp = lhs.Type;
         while (lhsTypeToCmp.IsPtrType)
            lhsTypeToCmp = lhsTypeToCmp.AsPtrType.ReferentType;

         Phx.Types.Type rhsTypeToCmp = rhs.Type;
         while (rhsTypeToCmp.IsPtrType)
            rhsTypeToCmp = rhsTypeToCmp.AsPtrType.ReferentType;

         // ISSUE-HACK-t-marcea-2005/08/15-Phx#1414 IsEqualWorker() only equates
         // two MSIL types if their MSIL handles are equal.  However, if the
         // types are refered to in different modules, their handles are likely
         // to be different.

         Phx.Types.AggrType lhsAggrType = lhsTypeToCmp.AsAggrType;
         Phx.Types.AggrType rhsAggrType = rhsTypeToCmp.AsAggrType;

         if ((lhsAggrType != null && rhsAggrType != null) &&
            !(lhsAggrType.IsInstantiation && rhsAggrType.IsInstantiation) &&
            (lhsAggrType.TypeSym != null && rhsAggrType.TypeSym != null))
         {
            Phx.Syms.MsilTypeSym lhsMsilTypeSym = lhsAggrType.TypeSym.AsMsilTypeSym;
            Phx.Syms.MsilTypeSym rhsMsilTypeSym = rhsAggrType.TypeSym.AsMsilTypeSym;

            if (lhsMsilTypeSym != null &&
               rhsMsilTypeSym != null)
            {
               return IsMsilTypeSymEqual(lhsMsilTypeSym, rhsMsilTypeSym);
            }
         }

         if (!Phx.Types.Type.IsEqualWorkerRecurse(lhsTypeToCmp, rhsTypeToCmp,
               false))
         {
            return false;
         }

         if ((lhs.IsIn != rhs.IsIn)
            || (lhs.IsOut != rhs.IsOut))
         {
            return false;
         }

         return true;
      }

      static public bool
      IsMsilTypeSymEqual
      (
         Phx.Syms.MsilTypeSym lhs,
         Phx.Syms.MsilTypeSym rhs
      )
      {
         // Give "official" Equals() a chance to say they are equal
         // (we throw away negative results since they are not reliable)

         if (Phx.Syms.MsilTypeSym.Equals(lhs, rhs))
         {
            return true;
         }
         else if (lhs == null || rhs == null)
         {
            return false;
         }

         // Compare the names

         else if (lhs.NameString != rhs.NameString)
         {
            return false;
         }

         // Give "official" Equals() a chance to say they are equal
         // (we throw away negative results since they are not reliable)

         else if (Phx.Syms.Sym.Equals(lhs.ResolutionScopeSym, rhs.ResolutionScopeSym))
         {
            return true;
         }

         // Verify the type names come from the same assembly

         return MyGetAssemblyName(lhs) == MyGetAssemblyName(rhs);
      }

      static public string
      MyGetAssemblyName
      (
         Phx.Syms.MsilTypeSym msilTypeSym
      )
      {
         if (msilTypeSym == null || msilTypeSym.LexicalParentSym == null)
            return null;

         Phx.Syms.ImportModuleSym importSym = msilTypeSym.LexicalParentSym.AsImportModuleSym;
         if (importSym != null)
         {
            if (System.IO.Path.GetDirectoryName(importSym.NameString).Length == 0)
               return importSym.NameString;
            else
               return System.IO.Path.GetFileNameWithoutExtension(importSym.NameString);
         }

         Phx.Syms.AssemblySym assemblySym = msilTypeSym.LexicalParentSym.AsAssemblySym;
         if (assemblySym != null)
            return assemblySym.NameString;

         return null;
      }

      /// <summary>
      /// Compares two Const Syms
      /// </summary>
      /// <returns>True if the symbols are equivalent</returns>
      /// <remarks>Can only do value comparisons for string constants</remarks>

      static public bool
      IsConstSymEqual
      (
         Phx.Syms.ConstSym lhs,
         Phx.Syms.ConstSym rhs
      )
      {
         if (lhs == rhs)
         {
            return true;
         }
         else if (lhs == null || rhs == null)
         {
            return false;
         }
         else if (Object.ReferenceEquals(lhs, rhs))
         {
            return true;
         }
         else if (!lhs.Type.Equals(rhs.Type))
         {
            return false;
         }
         else if (IsStringType(lhs.Type))
         {
            return String.Equals(lhs.Str, rhs.Str);
         }
         else
         {
            Dbg.Fail("NYI: Unhandled ConstSym type");
         }

         return false;
      }

      /// <summary>
      /// Compares a Const Sym with a string
      /// </summary>
      /// <returns>True if the const sym value and string val are equivalent</returns>
      /// <remarks>Const Sym must contain a string</remarks>

      static public bool
      IsConstSymEqual
      (
         Phx.Syms.ConstSym lhs,
         string val
      )
      {
         if (lhs == null && val == null)
         {
            return true;
         }
         else if (lhs == null || val == null)
         {
            return false;
         }
         else if (!IsStringType(lhs.Type))
         {
            return false;
         }
         else
         {
            return String.Equals(lhs.Str, val);
         }
      }

      /// <summary>
      /// Checks for string type
      /// </summary>
      /// <remarks>
      /// ISSUE-REVIEW-t-marcea-2005/08/04-Phx#2295-Phoenix seems
      /// to have multiple types in the type table to represent
      /// primitive types.  If I use the aggregate type with the
      /// TypeSym = "System.Object", Phoenix will create an invalid
      /// assembly (you'll see System.Object in metadata).  I have
      /// to locate the primitive type version of System.Object
      /// (which I call "object" since you'll see "object" in the
      /// metadata intead of "System.Object").
      /// </remarks>
      /// <param name="type">Type to check</param>
      /// <returns>true if the type is "System.String" or "string"</returns>

      static public bool
      IsStringType
      (
         Phx.Types.Type type
      )
      {
         if (type == null)
         {
            return false;
         }

         // Dereference pointer types

         while (type.IsPtrType)
         {
            type = type.AsPtrType.ReferentType;
         }

         Phx.Types.AggrType typeAggr = type.AsAggrType;
         if (typeAggr == null)
         {
            return false;
         }

         Dbg.Assert(typeAggr.TypeTable != null);
         Dbg.Assert(typeAggr.TypeTable.SystemStringAggrType != null);

         Phx.Types.AggrType strType = typeAggr.TypeTable.SystemStringAggrType;
         Dbg.Assert(strType != null);

         // ISSUE-HACK-t-marcea-2005/08/11-If one type is IsRef and the
         // other is IsDef, Phoenix incorrectly says they are not equal
         // even though they could be a TypeDef and TypeRef for the same
         // class.

         if (typeAggr.IsRef || strType.IsRef)
         {
            return typeAggr.TypeSym.NameString == strType.TypeSym.NameString;
         }
         else
         {
            return typeAggr.Equals(strType);
         }
      }

      /// <summary>
      /// Checks for object type
      /// </summary>
      /// <remarks>
      /// ISSUE-REVIEW-t-marcea-2005/08/04-Phx#2295-Phoenix seems
      /// to have multiple types in the type table to represent
      /// primitive types.  If I use the aggregate type with the
      /// TypeSym = "System.Object", Phoenix will create an invalid
      /// assembly (you'll see System.Object in metadata).  I have
      /// to locate the primitive type version of System.Object
      /// (which I call "object" since you'll see "object" in the
      /// metadata intead of "System.Object").
      /// </remarks>
      /// <param name="type">Type to check</param>
      /// <returns>true if the type is "System.Object" or "object"</returns>

      static public bool
      IsObjectType
      (
         Phx.Types.Type type
      )
      {
         if (type == null)
         {
            return false;
         }

         // Dereference pointer types

         while (type.IsPtrType)
         {
            type = type.AsPtrType.ReferentType;
         }

         Phx.Types.AggrType typeAggr = type.AsAggrType;
         if (typeAggr == null)
         {
            return false;
         }

         Dbg.Assert(typeAggr.TypeTable != null);
         Dbg.Assert(typeAggr.TypeTable.SystemObjectAggrType != null);

         Phx.Types.AggrType objType = typeAggr.TypeTable.SystemObjectAggrType;
         Dbg.Assert(objType != null);

         // ISSUE-HACK-t-marcea-2005/08/11-If one type is IsRef and the
         // other is IsDef, Phoenix incorrectly says they are not equal
         // even though they could be a TypeDef and TypeRef for the same
         // class.

         if (typeAggr.IsRef || objType.IsRef)
         {
            return typeAggr.TypeSym.NameString == objType.TypeSym.NameString;
         }
         else
         {
            return typeAggr.Equals(objType);
         }
      }

      /// <summary>
      /// Compares a type with a primitive type
      /// </summary>
      /// <param name="type">Type in question</param>
      /// <param name="primType">Primitive type to compare to</param>
      /// <returns>true if the type, or the referent type in the case
      /// of pointer types, in question is equivalent to 
      /// the primitive type</returns>

      static public bool
      IsPrimTypeEqual
      (
         Phx.Types.Type type,
         Phx.Types.PrimType primType
      )
      {
         Dbg.Assert(primType != null);
         if (type == null || primType == null)
            return false;

         while (type.IsPtrType)
            type = type.AsPtrType.ReferentType;

         return type.Equals(primType);
      }

      /// <summary>
      /// Checks the type table for a primitive type that represents
      /// the aggregate type
      /// </summary>
      /// <remarks>
      /// Builtin types always have an aggregate and primitive type
      /// pair.
      /// </remarks>
      /// <param name="type">Aggregate type to look for</param>
      /// <returns>
      /// The primary type ('bool8') that underlies the aggregate
      /// type ('System.Boolean').
      /// </returns>

      static public void ConvertAggregateToPrimary
      (
         ref Phx.Types.Type type
      )
      {
         // ISSUE-HACK-t-marcea-2005/08/02-Phx#2295 When adding a
         // primitive field using the fully qualified name, e.g., 
         // System.Boolean, you must make sure Phoenix uses the underlying
         // primitive type (bool8).  Otherwise, Phoenix will output the
         // field as being of type "System.Boolean" instead of "bool"
         // which is not CLS compliant.  Subsequently, the C# compiler
         // will complain, for example:
         //
         // client.cs(10,35): error CS0570: 'HelloNS.Hello.marcsField'
         // is not supported by the language

         if (type != null && type.IsPrimaryAggr)
         {
            Phx.Types.PrimType primitiveType = type.AsAggrType.UnderlyingPrimType;
            if (primitiveType != null)
            {
               type = primitiveType;
            }
         }
      }

      /// <summary>
      /// Diagnostic routine that outputs the type table
      /// (including TypeSym names) to standard output
      /// </summary>
      /// <param name="table">Type table to dump</param>

      static public void
      DumpTypeTable
      (
         Phx.Types.Table table
      )
      {
         #if PHX_DEBUG_DUMPS

         for (Phx.Types.Type currType = table.AllTypes;
               currType != null;
               currType = currType.Next)
         {
            //currType.Dump();

            // ISSUE-REVIEW-t-marcea-2005/08/02-Phx.Types.Table.Dump()
            // should display the TypeSym name to make the output more
            // useful.

            if (currType.TypeSym != null)
            {
               Helpers.WriteOutput("Symbol:");
               //currType.TypeSym.Dump();
            }
         }

         #endif
      }

      /// <summary>
      /// Retrieves the underlying TypeSym, dereferencing pointer
      /// types and primitive types if necessary
      /// </summary>
      /// <returns></returns>

      public static Phx.Syms.TypeSym
      GetUnderlyingTypeSym
      (
         Phx.Types.Type type
      )
      {
         if (type == null)
         {
            return null;
         }

         // Dereference ptr types

         while (type.IsPtrType)
         {
            type = type.AsPtrType.ReferentType;
         }

         // Dereference primitive types

         if (type.IsPrimType)
         {
            // This converts 'bool8' to 'System.Boolean'

            Phx.Types.Type verAggrType = type.AsPrimType.VerAggrType;
            if (verAggrType == null)
            {
               // This is normal when the primitive type is 'void'
               return null;
            }

            type = verAggrType;
         }

         return type.TypeSym;
      }

      /// <summary>
      /// Returns the name of the underlying TypeSym or "(null)"
      /// </summary>
      /// <returns>name of the underlying TypeSym or "(null)"</returns>

      public static string
      SafeGetName
      (
         Phx.Types.Type type
      )
      {
         Phx.Syms.TypeSym typeSym = GetUnderlyingTypeSym(type);
         if (typeSym != null)
         {
            return typeSym.NameString;
         }
         else
         {
            return "void";
         }
      }

      public static void
      SafeGetName
      (
         Phx.Types.Type type,
         ref StringBuilder sb
      )
      {
         Phx.Syms.TypeSym typeSym = GetUnderlyingTypeSym(type);
         if (typeSym != null)
         {
            if (sb == null)
            {
               sb = new StringBuilder(typeSym.NameString);
            }
            else
            {
               sb.Length = 0;
               sb.Append(typeSym.NameString);
            }
         }
         else
         {
            if (sb == null)
            {
               sb = new StringBuilder("void");
            }
            else
            {
               sb.Length = 0;
               sb.Append("void");
            }
         }
      }

      static public string
      SafeGetName2
      (
         Phx.Types.Type type
      )
      {
         string typeName = SafeGetName(type);

         string newName;
         bool found = mapFQN2keyword.TryGetValue(typeName, out newName);
         if (found)
         {
            return newName;
         }
         else
            return typeName;
      }

      static public void
      SafeGetName2
      (
         Phx.Types.Type type,
         ref StringBuilder sb
      )
      {
         SafeGetName(type, ref sb);

         string newName;
         bool found = mapFQN2keyword.TryGetValue(sb.ToString(), out newName);
         if (found)
         {
            sb.Length = 0;
            sb.Append(newName);
         }
      }

      /// <summary>
      /// Tries to get the source file name for where the
      /// instruction is physically located
      /// </summary>
      /// <remarks>
      /// Optimization: If fileName is not null we assume the
      /// filename was retrieved on a previous call to
      /// SafeGetSourceFileName() so we don't get it again
      /// </remarks>
      /// <param name="instr">
      /// Instruction to get the source file name of
      /// </param>
      /// <param name="filePath">
      /// Will be set to the file path or "(null)"
      /// if there is no file path associated with the
      /// instruction
      /// </param>

      public static void
      SafeGetSourceFilePath
      (
         Phx.IR.Instr instr,
         ref string filePath
      )
      {
         if (filePath != null)
         {
            // We already retrieved it

            return;
         }
         else if (instr == null ||
            instr.FuncUnit == null ||
            instr.FuncUnit.DebugInfo == null ||
            instr.DebugTag == 0)
         {
            filePath = "(null)";
         }
         else
         {
            filePath = instr.FuncUnit.DebugInfo.GetFileName(instr.DebugTag);
         }
      }

      public static void
      SafeGetSourceFilePath
      (
         Phx.IR.Instr instr,
         ref StringBuilder filePath
      )
      {
         if (filePath == null)
            filePath = new StringBuilder();
         else
            filePath.Length = 0;

         if (instr == null ||
            instr.FuncUnit == null ||
            instr.FuncUnit.DebugInfo == null ||
            instr.DebugTag == 0)
         {
            filePath.Append("(null)");
         }
         else
         {
            filePath.Append(instr.FuncUnit.DebugInfo.GetFileName(instr.DebugTag));
         }
      }

      /// <summary>
      /// Tries to get the source file line number for where the
      /// instruction is physically located
      /// </summary>
      /// <param name="instr">
      /// Instruction to get the source line number of
      /// </param>
      /// <returns>
      /// Line number or 0 if there is no line number information
      /// </returns>

      public static uint
      SafeGetSourceLine
      (
         Phx.IR.Instr instr
      )
      {
         if (instr == null ||
            instr.FuncUnit == null ||
            instr.FuncUnit.DebugInfo == null ||
            instr.DebugTag == 0)
         {
            return 0;
         }
         else
         {
            // ISSUE-BUG-t-marcea-2006/05/04-GetLineNum() appears to return the
            // line number + 65536 (possibly System.Byte.MaxValue?)

            return instr.FuncUnit.DebugInfo.GetLineNum(instr.DebugTag) % 65536;
         }
      }

      /// <summary>
      /// Tries to get the source file column for where the
      /// instruction is physically located
      /// </summary>
      /// <param name="instr">
      /// Instruction to get the source column of
      /// </param>
      /// <returns>
      /// Column or 0 if there is no line number information
      /// </returns>

      public static uint
      SafeGetSourceColumn
      (
         Phx.IR.Instr instr
      )
      {
         if (instr == null ||
            instr.FuncUnit == null ||
            instr.FuncUnit.DebugInfo == null ||
            instr.DebugTag == 0)
         {
            return 0;
         }
         else
         {
            return instr.FuncUnit.DebugInfo.GetColumnNum(instr.DebugTag);
         }
      }

      public static Phx.Types.Type
      GetSymType
      (
         Phx.Syms.Sym sym
      )
      {
         if (sym == null)
            return null;

         Phx.Syms.PropertySym propSym = sym.AsPropertySym;

         if (propSym == null)
            return sym.Type;

         // Handle PropertySyms

         // ISSUE-REVIEW-t-marcea-2005/08/16-PropertySym.Type is always
         // null.  I think it should be the type of the property.

         if (propSym.FuncSymGetter != null)
         {
            return propSym.FuncSymGetter.FuncType.RetType;
         }
         else if (propSym.FuncSymSetter != null)
         {
            Phx.Types.FuncArg firstArg =
               propSym.FuncSymSetter.FuncType.ArgFuncArgList;

            if (propSym.FuncSymSetter.FuncType.IsInstanceMethod &&
               firstArg != null)
            {
               // Skip the this ptr for instance methods

               firstArg = firstArg.Next;
            }

            Dbg.Assert(firstArg != null);
            if (firstArg == null)
            {
               return null;
            }

            return firstArg.Type;
         }
         else
         {
            // propSym.Type is always null

            return propSym.Type;
         }
      }

      public static Phx.Syms.Access
      GetPropSymAccess
      (
         Phx.Syms.PropertySym propSym
      )
      {
         if (propSym == null)
            return Phx.Syms.Access._Illegal;

         // ISSUE-REVIEW-t-marcea-2005/08/16-PropertySym.Access is not
         // available.  I think it should be the access of the property.

         if (propSym.FuncSymGetter != null)
         {
            return propSym.FuncSymGetter.Access;
         }
         else if (propSym.FuncSymSetter != null)
         {
            return propSym.FuncSymSetter.Access;
         }
         else
         {
            return Phx.Syms.Access._Illegal;
         }
      }

      static public void WriteOutput(string format, params object[] args)
      {
         if (verbosity >= 3)
         {
            ConsoleColor fc = Console.ForegroundColor;
            Console.ForegroundColor = ConsoleColor.White;
            System.Console.WriteLine("Phx.Morph: " + format, args);
            Console.ForegroundColor = fc;
         }
      }

      static public void WriteWarning(string format, params object[] args)
      {
         if (verbosity >= 2)
         {
            ConsoleColor fc = Console.ForegroundColor;
            Console.ForegroundColor = ConsoleColor.Yellow;
            System.Console.WriteLine("Phx.Morph: Warning: " + format, args);
            Console.ForegroundColor = fc;
         }
      }

      static public void WriteError(string format, params object[] args)
      {
         if (verbosity >= 1)
         {
            ConsoleColor fc = Console.ForegroundColor;
            Console.ForegroundColor = ConsoleColor.Red;
            System.Console.WriteLine("Phx.Morph: " + format, args);
            Console.ForegroundColor = fc;
         }
      }
   }
}
