//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Morph/ReflectionHelpers.cs 17    6/02/06 2:37a Eaddy $
//
// Description:
//
//    .NET Reflection helper functions
//
// Remarks:
//
// ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
// for unmanaged implementation
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Morph/ReflectionHelpers.cs $
 * 
 * 17    6/02/06 2:37a Eaddy
 * o Shouldn't show warnings since we expect
 * them to happen
 * 
 * 16    5/25/06 8:24p Eaddy
 * o Fixed bug where module wasn't found if it wasn't in
 * the current directory
 * 
 * 15    5/24/06 9:56p Eaddy
 * o Now tries to resolve assemblies that can't be loaded
 * o Improved instantiation of custom attributes and finding
 * the correct ctor for a given blob
 * 
 * 14    5/12/06 1:25a Eaddy
 * o Fixed bug where we had a double ptr
 * 
 * 13    5/11/06 10:47p Rr2318
 * Modified to support blob parsing
 * 
 * 12    5/11/06 5:23p Eaddy
 * o Added additional probing steps during assembly
 * loading. For some reason this was failing when
 * loading assemblies from the current directory
 * whose names were obtaine via ResolutionScopeSym.
 * 
 * 11    5/11/06 5:04p Eaddy
 * o Fixed problem where assembly wasn't loaded if
 * it was an exe
 * 
 * 10    5/04/06 11:46p Eaddy
 * o Added comments
 * 
 * 9     5/04/06 11:33p Eaddy
 * o Fixed bug
 * 
 * 8     5/04/06 9:51p Eaddy
 * o Added ConvertConstructorInfoToPhxFuncSym
 * 
 * 7     5/04/06 8:49p Eaddy
 * o Fixed bug where we were unable to resolve a type
 * using Reflection even though the it was available via
 * Phx
 * 
 * 6     5/03/06 8:14p Rr2318
 * Deep search in ConvertPhxTypeToClrType
 * 
 * 5     5/03/06 5:08p Eaddy
 * o Added helpful comment
 * 
 * 4     4/02/06 4:34p Asv2105
 * Added ability for BreakpointWeaver to pass MethodInfo to the
 * coordinator
 * 
 * Adjusted ICoordinator, IEventListener interfaces to match reality
 * 
 * Added ability for Coordinator to call static methods when a breakpoint
 * fires and pass off to EventMgr for non-static methods
 * 
 * 3     3/15/06 1:18p Eaddy
 * o Fixed bug where we failed to load aspect assemblies that
 * were not in the current directory
 * 
 * 2     2/24/06 8:27p Eaddy
 * o Added checkin comment
 */

using System;
using System.IO;
using System.Collections.Generic;

using Dbg = System.Diagnostics.Debug;

namespace Phx.Morph
{
   /// <summary>
   /// .NET Reflection helper functions
   /// </summary>
   /// <remarks>
   /// ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
   /// for unmanaged implementation
   /// </remarks>

   public class ReflectionHelpers
   {
      /// <summary>
      /// Converts a Phx type (compile-time reflection type) to a CLR
      /// type (runtime reflection type)
      /// </summary>
      /// <remarks>
      /// For the conversion to succeed, the assembly that defines the
      /// type must be loaded by the CLR, either implicitly by the CLR
      /// or explicitly using System.Reflection.Assembly.Load().
      /// </remarks>
      /// <param name="phxType">The Phx type to convert</param>
      /// <returns>CLR type corresponding to the Phx type</returns>

      public static System.Type
      ConvertPhxTypeToClrType
      (
         Phx.Types.Type phxType
      )
      {
         // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
         // for unmanaged implementation
         // Dereference pointer types
         if (phxType == null)
         {
            return null;
         }

         while (phxType.IsPtrType)
         {
            phxType = phxType.AsPtrType.ReferentType;
         }

         if (ParsePhxPrimType(phxType) != null)
         {
            return ParsePhxPrimType(phxType);
         }

         if (phxType.TypeSym == null)
         {
            return null;
         }

         System.Type t = MyGetType(phxType.TypeSym.NameString);
         if(t != null) return t;

         if (phxType.TypeSym.ResolutionScopeSym != null &&
            phxType.TypeSym.ResolutionScopeSym.NameString != null)
         {
            System.Reflection.Assembly assembly = 
               ReflectionHelpers.LoadAssembly(phxType.TypeSym.ResolutionScopeSym.NameString);
            return MyGetType(assembly, phxType.TypeSym.NameString);
         }

         return null;
      }

      public static System.Type
      ParsePhxPrimType
      (
         Phx.Types.Type phxType
       )
      {
         // Some manhacks necessary for prim types -- phx stores them differently
         string tmp = phxType.ToString();
         switch (tmp)
         {
            case "i32": return System.Type.GetType("System.Int32");
            case "i64": return System.Type.GetType("System.Int64");
            case "u32": return System.Type.GetType("System.UInt32");
            case "u64": return System.Type.GetType("System.UInt64");
         }
         return null;
      }

      public static System.Reflection.MethodInfo
      ConvertPhxFuncSymToMethodInfo
      (
           Phx.Syms.FuncSym phxFunc
      )
      {
           if (phxFunc != null) {
               Type clrType = ConvertPhxTypeToClrType(phxFunc.LexicalParentSym.Type);
               if (clrType == null) {
                   return null;
               }
               List<Type> args = new List<Type>();
               Phx.Types.FuncArg item = phxFunc.FuncType.ArgFuncArgList;
               while (item != null) {
                   args.Add(ConvertPhxTypeToClrType(Helpers.GetUnderlyingTypeSym(item.Type).Type));
                   item = item.Next;
               }
               return clrType.GetMethod(phxFunc.Name, args.ToArray());
           } else {
               return null;
           }
       }

       /// <summary>
       /// Converts a CLR ConstructorInfo (runtime reflection type) to a
       /// Phx FuncSym (compile-time reflection type)
       /// </summary>
       /// <remarks>
       /// If the module that defines the ctor is not yet loaded, we will
       /// load it.
       /// </remarks>
       /// <param name="symTable">Symbol table to look in</param>
       /// <param name="ctorInfo">ConstructorInfo to convert</param>
       /// <returns>Phx FuncSym for the ConstructorInfo</returns>

      public static Phx.Syms.FuncSym
      ConvertConstructorInfoToPhxFuncSym
      (
         Phx.Syms.Table symTable,
         System.Reflection.ConstructorInfo ctorInfo
      )
      {
         if (ctorInfo == null || symTable == null)
            return null;

         Phx.Syms.MsilTypeSym classType = Helpers.FindMsilTypeSym(symTable, ctorInfo.DeclaringType.FullName);
         if (classType == null)
         {
            PEModuleUnit gunit = ReflectionHelpers.LoadModuleUnit(ctorInfo.Module.FullyQualifiedName);
            Dbg.Assert(gunit != null);
            if (gunit == null)
               return  null;
            //gunit.LoadPdb(); 
            //gunit.LoadGlobalSyms();
            classType = Helpers.FindMsilTypeSym(gunit.SymTable, ctorInfo.DeclaringType.FullName);
            if (classType == null)            
               return null;
         }

         symTable = classType.Table;
         
         // Sanity checks

         Dbg.Assert(symTable != null);
         Dbg.Assert(symTable.LocalIdMap != null);
         Dbg.Assert(symTable.LocalIdMap.InternalMap != null);
         if (symTable == null || symTable.LocalIdMap == null ||
            symTable.LocalIdMap.InternalMap == null)
         {
            return null;
         }
         System.Reflection.ParameterInfo[] pi = ctorInfo.GetParameters();
         foreach (Phx.Syms.Sym sym in symTable.LocalIdMap.InternalMap)
         {
            if (!sym.IsFuncSym)
            {
               continue;
            }

            Phx.Syms.FuncSym funcSym = sym.AsFuncSym;
            if (funcSym.EnclosingAggrType != classType.Type.AsAggrType)
            {
               continue;
            }

            if (EqualSignaturesIgnoringThisAndPtrs(funcSym, ctorInfo))
            {
               return funcSym;
            }
         }

         return null;
      }

      public static System.Reflection.ConstructorInfo
      FindMatchingConstructorForBlob
      (
         System.Type type,
         byte[] blob,
         bool showWarnings 
      )
      {
         foreach (System.Reflection.ConstructorInfo ci in type.GetConstructors())
         {
            if (ReflectionHelpers.DoesConstructorMatchBlob(ci, blob, showWarnings))
            {
               return ci;
            }
         }

         return null;
      }

      public static bool
      DoesConstructorMatchBlob
      (
         System.Reflection.ConstructorInfo ctorInfo,
         byte[] blob,
         bool showWarnings
      )
      {
         System.Reflection.ParameterInfo[] pi = ctorInfo.GetParameters();
         
         if (pi == null || pi.Length == 0)
         {
            return blob.Length <= 4;
         }

         int iter = 2;
         for (int i = 0; i < pi.Length && iter + 2 < blob.Length; ++i)
         {
            if (pi[i].ParameterType == typeof(string))
            {
               iter += (int) blob[iter] + 1;
            }
            else if (pi[i].ParameterType.IsPrimitive ||
                     pi[i].ParameterType.IsEnum)
            {
               // Hyperfine processing necessary, but this works for all
               // extensive purposes - Rajesh
               iter += 4;  
            }
            else if (pi[i].ParameterType.IsArray)
            {
               // Grab the array count
               int aryCount = (int)blob[iter];
               iter += 4;

               for(int j = 0; j < aryCount; ++j)
               {
                  CorSerializationType encodedType = (CorSerializationType) blob[iter];
                  iter += 1;

                  if (pi[i].ParameterType.GetElementType() == typeof(object))
                  {
                     if (encodedType == CorSerializationType.STRING)
                     {
                        // Skip over the string
                        iter += (int)blob[iter] + 1;
                     }
                     else if (encodedType == CorSerializationType.ENUM)
                     {
                        // Skip over type spec string for the enum
                        iter += (int)blob[iter] + 1;
                        
                        // Skip over enum value (hopefully not stored as a short)
                        iter += 4;
                     }
                     else if ((int) encodedType < (int) CorElementType.PTR)
                     {
                        // These are simple types
                        iter += 4;
                     }
                     else
                     {
                        if (showWarnings)
                        {
                           // Can't handle anything else
                           Helpers.WriteWarning("Ignoring unsupported attribute constructor parameter type: {0}.",
                              encodedType);
                        }
                        return false;
                     }
                  }
                  else
                  {
                     if (showWarnings)
                     {
                        // Can't handle anything else
                        Helpers.WriteWarning("Ignoring unsupported attribute constructor array parameter type: {0}.",
                           pi[i].ParameterType.GetElementType());
                     }
                     return false;
                  }
               }
            }
            else 
            {
               if (showWarnings)
               {
                  // Can't handle anything else
                  Helpers.WriteWarning("Ignoring unsupported attribute constructor parameter type: {0}.",
                     pi[i].ParameterType);
               }
               return false;
            }
         }

         return blob.Length == iter + 2; //2 for tail
      }

      /// <summary>
      /// Compares a Phx FuncSym with a CLR ConstructorInfo
      /// </summary>
      /// <remarks>
      /// This hasn't been thoroughly tested.
      /// </remarks>
      /// <param name="lhsFuncSym">FuncSym to compare</param>
      /// <param name="ctorInfo">ConstructorInfo to compare</param>
      /// <returns>true if they are equivalent, otherwise false</returns>

      static public bool
      EqualSignaturesIgnoringThisAndPtrs
      (
         Phx.Syms.FuncSym lhsFuncSym,
         System.Reflection.ConstructorInfo rhsCtorInfo
      )
      {
         if (lhsFuncSym.NameString != rhsCtorInfo.Name)
         {
            return false;
         }

         System.Reflection.ParameterInfo[] clrParams = rhsCtorInfo.GetParameters();

         Phx.Types.FuncType lhsFuncType = lhsFuncSym.FuncType;
         if (lhsFuncType == null)
         {
            return false;
         }

         int argsExpected = lhsFuncType.CountArgs();

         Phx.Types.FuncArg fa = lhsFuncType.ArgFuncArgList;
         if (fa == null)
         {
            return false;
         }

         if (argsExpected > 0 && fa.IsThisPtr)
         {
            --argsExpected;
            fa = fa.Next;
         }

         // Verify same number of params
         if (argsExpected != clrParams.Length)
         {
            return false;
         }

         if (argsExpected != clrParams.Length)
         {
            return false;
         }

         if ((lhsFuncType.IsInstanceMethod && rhsCtorInfo.IsStatic) ||
             (!lhsFuncType.IsInstanceMethod && !rhsCtorInfo.IsStatic))
         {
            return false;
         }

         // Ignore calling convention

         int clrParamIndex = 0;

         while ((fa != null))
         {
            Dbg.Assert(!fa.IsThisPtr);

            System.Reflection.ParameterInfo clrParam = clrParams[clrParamIndex];
            Dbg.Assert(clrParam != null);

            if (!ReflectionHelpers.IsArgEqualIgnorePtrs(fa, clrParam))
            {
               return false;
            }

            fa = fa.Next;
            ++clrParamIndex;
         }

         return true;
      }

      /// <summary>
      /// Compares a Phx FuncArg with a CLR ParameterInfo
      /// </summary>
      /// <remarks>
      /// This hasn't been thoroughly tested.
      /// </remarks>
      /// <param name="lhsFuncSym">FuncArg to compare</param>
      /// <param name="ctorInfo">ParameterInfo to compare</param>
      /// <returns>true if they are equivalent, otherwise false</returns>

      static public bool
      IsArgEqualIgnorePtrs
      (
         Phx.Types.FuncArg lhs,
         System.Reflection.ParameterInfo rhs
      )
      {
         if (lhs == null || rhs == null)
         {
            return false;
         }

         // Vararg functions have no type field.

         if (lhs.IsAnEllipsis)
         {
            return true;
         }

         Phx.Types.Type lhsTypeToCmp = lhs.Type;
         while (lhsTypeToCmp.IsPtrType)
            lhsTypeToCmp = lhsTypeToCmp.AsPtrType.ReferentType;

         // ISSUE-HACK-t-marcea-2005/08/15-Phx#1414 IsEqualWorker() only equates
         // two MSIL types if their MSIL handles are equal.  However, if the
         // types are refered to in different modules, their handles are likely
         // to be different.

         Phx.Types.AggrType lhsAggrType = lhsTypeToCmp.AsAggrType;
         Phx.Types.AggrType rhsAggrType = ReflectionHelpers.ConvertClrTypeToPhxType(lhs.Type.TypeTable, rhs.ParameterType).AsAggrType;

         if ((lhsAggrType != null && rhsAggrType != null) &&
            !(lhsAggrType.IsInstantiation && rhsAggrType.IsInstantiation) &&
            (lhsAggrType.TypeSym != null && rhsAggrType.TypeSym != null))
         {
            Phx.Syms.MsilTypeSym lhsMsilTypeSym = lhsAggrType.TypeSym.AsMsilTypeSym;
            Phx.Syms.MsilTypeSym rhsMsilTypeSym = rhsAggrType.TypeSym.AsMsilTypeSym;

            if (lhsMsilTypeSym != null &&
               rhsMsilTypeSym != null)
            {
               return Helpers.IsMsilTypeSymEqual(lhsMsilTypeSym, rhsMsilTypeSym);
            }
         }

         if (!Phx.Types.Type.IsEqualWorkerRecurse(lhsTypeToCmp, rhsAggrType,
               false))
         {
            return false;
         }

         if ((lhs.IsIn != rhs.IsIn)
            || (lhs.IsOut != rhs.IsOut))
         {
            return false;
         }

         return true;
      }

      /// <summary>
      /// Converts a CLR type (runtime reflection type) to a Phx type
      /// (compile-time reflection type)
      /// </summary>
      /// <remarks>
      /// For the conversion to succeed, the assembly that defines the
      /// type must be loaded by the Phx PE Reader (e.g.,
      /// Phx.PEModuleUnit.Open()).
      /// </remarks>
      /// <param name="phxTypeTable">Phx type table to lookup the type</param>
      /// <param name="clrType">CLR type to convert</param>
      /// <returns>Phx type corresponding to the CLR type</returns>

      public static Phx.Types.Type
      ConvertClrTypeToPhxType
      (
         Phx.Types.Table phxTypeTable,
         System.Type clrType
      )
      {
         // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
         // for unmanaged implementation

         Phx.Types.AggrType aggrType = Helpers.FindAggrType(phxTypeTable, clrType.ToString(), false);
         if (aggrType != null)
            return aggrType;

         PEModuleUnit gunit = ReflectionHelpers.LoadModuleUnit(clrType.Module.FullyQualifiedName);
         return Helpers.FindAggrType(gunit.TypeTable, clrType.ToString(), false);
      }

      /// <summary>
      /// Converts between .NET Reflection field attributes and
      /// Phx field access
      /// </summary>
      /// <param name="clrAttributes">CLR field attributes to convert</param>
      /// <returns>Phx field access portion of the field attributes</returns>

      public static Phx.Syms.Access
      ConvertClrFieldAttrToPhxAccess
      (
         System.Reflection.FieldAttributes clrAttributes
      )
      {
         // ISSUE-TODO-t-marcea-2005/07/27-We don't handle PrivateScope

         // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
         // for unmanaged implementation

         switch (clrAttributes)
         {
            case System.Reflection.FieldAttributes.Assembly:
               return Phx.Syms.Access.Assembly;
            case System.Reflection.FieldAttributes.FamANDAssem:
               return Phx.Syms.Access.FamilyAndAssembly;
            case System.Reflection.FieldAttributes.Family:
               return Phx.Syms.Access.Family;
            case System.Reflection.FieldAttributes.FamORAssem:
               return Phx.Syms.Access.FamilyOrAssembly;
            case System.Reflection.FieldAttributes.Private:
               return Phx.Syms.Access.Private;
            case System.Reflection.FieldAttributes.Public:
               return Phx.Syms.Access.Public;
            default:
               Dbg.Fail("Unknown field access");
               return Phx.Syms.Access.Public;
         }
      }

      /// <summary>
      /// Converts between Phx field access and .NET Reflection field 
      /// attributes
      /// </summary>
      /// <param name="access">Phx field access portion of the field attributes</param>
      /// <returns>CLR field attributes to convert</returns>

      public static System.Reflection.FieldAttributes
      ConvertPhxAccessToClrFieldAttr
      (
         Phx.Syms.Access access
      )
      {
         // ISSUE-TODO-t-marcea-2005/07/27-We don't handle PrivateScope

         // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
         // for unmanaged implementation

         switch (access)
         {
            case Phx.Syms.Access.Assembly:
               return System.Reflection.FieldAttributes.Assembly;
            case Phx.Syms.Access.FamilyAndAssembly:
               return System.Reflection.FieldAttributes.FamANDAssem;
            case Phx.Syms.Access.Family:
               return System.Reflection.FieldAttributes.Family;
            case Phx.Syms.Access.FamilyOrAssembly:
               return System.Reflection.FieldAttributes.FamORAssem;
            case Phx.Syms.Access.Private:
               return System.Reflection.FieldAttributes.Private;
            case Phx.Syms.Access.Public:
               return System.Reflection.FieldAttributes.Public;
            default:
               Dbg.Fail("Unknown field access");
               return System.Reflection.FieldAttributes.Public;
         }
      }

      /// <summary>
      /// Obtains the Type for a fully qualified type name by looking
      /// in all the loaded assemblies
      /// </summary>
      /// <remarks>
      /// Type.GetType(string) is supposed to do this exact thing
      /// but it doesn't work
      /// </remarks>
      /// <param name="typeName">Fully qualified name of the type
      /// (System.String)</param>
      /// <returns>Requested Type or null</returns>

      static public Type
      MyGetType
      (
         string typeName
      )
      {
         // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
         // for unmanaged implementation

         System.Type clrType = System.Type.GetType(typeName);
         if (clrType != null)
            return clrType;

         // Scour the other loaded assemblies for the type. This is
         // exactly what Type.GetType() is supposed to do (IMHO).

         System.AppDomain clrDomain = System.AppDomain.CurrentDomain;
         System.Reflection.Assembly[] clrAssemblies = clrDomain.GetAssemblies();
         foreach (System.Reflection.Assembly clrAssembly in clrAssemblies)
         {
            clrType = MyGetType(clrAssembly, typeName);
            if (clrType != null)
               return clrType;
         }

         // We did not find the CLR type because the module where the type
         // is defined has not been loaded via System.Reflection.  Try calling
         // ReflectionHelpers.LoadAssembly() first.
         return null;
      }

      /// <summary>
      /// Obtains the Type for a fully qualified type name by looking
      /// in all the loaded assemblies
      /// </summary>
      /// <remarks>
      /// Type.GetType(string) is supposed to do this exact thing
      /// but it doesn't work
      /// </remarks>
      /// <param name="clrAssembly">Assembly to search</param>
      /// <param name="typeName">Fully qualified name of the type
      /// (System.String)</param>
      /// <returns>Requested Type or null</returns>

      static public Type
      MyGetType
      (
         System.Reflection.Assembly clrAssembly,
         string typeName
      )
      {
         System.Reflection.Module[] clrModules = clrAssembly.GetModules();
         foreach (System.Reflection.Module clrModule in clrModules)
         {
            try
            {
               Type clrType = clrModule.GetType(typeName, false, false);
               if (clrType != null)
                  return clrType;
            }
            catch (TypeLoadException /*ex*/)
            {
               //Helpers.WriteError(ex.Message);
            }
         }

         // We did not find the CLR type because the module where the type
         // is defined has not been loaded via System.Reflection.  Try calling
         // ReflectionHelpers.LoadAssembly() first.
         return null;
      }

      /// <summary>
      /// Loads a module unit into Phx using Phx.PEModuleUnit.Open()
      /// </summary>
      /// <remarks>Also populates the module unit's type and symbol
      /// tables</remarks>
      /// <param name="assemblyString">Path of the assembly to load</param>
      /// <returns>
      /// Loaded Phx module unit or null if the assembly couldn't be loaded
      /// </returns>

      public static Phx.PEModuleUnit
      LoadModuleUnit
      (
         string assemblyString
      )
      {
         try
         {
            Phx.PEModuleUnit phxModuleUnit = Phx.PEModuleUnit.Open(assemblyString);
            Dbg.Assert(phxModuleUnit != null);
            if (phxModuleUnit == null)
               return null;

            // This populates the module's type and symbol tables

            phxModuleUnit.LoadEncodedIRUnitList();

            phxModuleUnit.Context.PopUnit();

            return phxModuleUnit;
         }
         catch (Exception ex)
         {
            Helpers.WriteError("Failed to open module reference '{0}': {1}",
               assemblyString,
               ex.Message);
            return null;
         }
      }

      /// <summary>
      /// Load an assembly into the CLR
      /// </summary>
      /// <param name="assemblyString">
      /// Can be an assembly short name (assembly name without
      /// an extension), relative path, or full path.
      /// </param>
      /// <returns>
      /// Loaded CLR assembly or null if the assembly couldn't be loaded
      /// </returns>

      public static System.Reflection.Assembly
      LoadAssembly
      (
         string assemblyString
      )
      {
         System.AppDomain currentDomain = System.Threading.Thread.GetDomain();
         currentDomain.AssemblyResolve += AssemblyResolveEventHandler;

         // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
         // for unmanaged implementation

         // Strip off the extension if present

         try
         {
            if (System.IO.Path.IsPathRooted(assemblyString))
            {
               return System.Reflection.Assembly.LoadFrom(assemblyString);
            }
            else
            {
               string assemblyStringWOExt = assemblyString;

               switch (Path.GetExtension(assemblyString))
               {
                  // ISSUE-TODO-t-marcea-2005/07/23-Load() and LoadFile()
                  // don't work if an extension is specified (verify?).
                  // However, the cases below probably don't cover all
                  // possibilities.

                  case ".dll":
                  case ".exe":
                  case ".netmodule":
                     assemblyStringWOExt = Path.Combine(
                        Path.GetDirectoryName(assemblyString),
                        Path.GetFileNameWithoutExtension(assemblyString));
                     break;
               }

               return System.Reflection.Assembly.Load(assemblyStringWOExt);
            }
         }
         catch (Exception ex)
         {
            Helpers.WriteWarning("Assembly load failed for '{0}': {1}",
               assemblyString,
               ex.Message);
            return null;
         }
      }

      public static System.Reflection.Assembly
      AssemblyResolveEventHandler
      (
         object sender,
         ResolveEventArgs e
      )
      {
         string assemblyName = e.Name;

         int comma = assemblyName.IndexOf(',');
         if (comma > 0)
         {
            assemblyName = assemblyName.Substring(0, comma);
         }

         string[] tryExts = { "", ".dll", ".exe", ".netmodule" };
         foreach (string ext in tryExts)
         {
            try
            {
               return System.Reflection.Assembly.LoadFrom(assemblyName + ext);
            }
            catch (Exception)
            { }
         }

         return null;
      }
   }

   // From corhdr.h
   enum CorElementType
   {
      END              = 0x0,
      VOID             = 0x1,
      BOOLEAN          = 0x2,
      CHAR             = 0x3,
      I1               = 0x4,
      U1               = 0x5,
      I2               = 0x6,
      U2               = 0x7,
      I4               = 0x8,
      U4               = 0x9,
      I8               = 0xa,
      U8               = 0xb,
      R4               = 0xc,
      R8               = 0xd,
      STRING           = 0xe,

      // every type above PTR will be simple type
      PTR              = 0xf,      // PTR <type>
      BYREF            = 0x10,     // BYREF <type>

      // Please use VALUETYPE. VALUECLASS is deprecated.
      VALUETYPE        = 0x11,     // VALUETYPE <class Token>
      CLASS            = 0x12,     // CLASS <class Token>
      VAR              = 0x13,     // a class type variable VAR <U1>
      ARRAY            = 0x14,     // MDARRAY <type> <rank> <bcount> <bound1> ... <lbcount> <lb1> ...
      GENERICINST      = 0x15,     // GENERICINST <generic type> <argCnt> <arg1> ... <argn>
      TYPEDBYREF       = 0x16,     // TYPEDREF  (it takes no args) a typed referece to some other type

      I                = 0x18,     // native integer size
      U                = 0x19,     // native unsigned integer size
      FNPTR            = 0x1B,     // FNPTR <complete sig for the function including calling convention>
      OBJECT           = 0x1C,     // Shortcut for System.Object
      SZARRAY          = 0x1D,     // Shortcut for single dimension zero lower bound array
      // SZARRAY <type>
      MVAR             = 0x1e,     // a method type variable MVAR <U1>

      // This is only for binding
      CMOD_REQD        = 0x1F,     // required C modifier : E_T_CMOD_REQD <mdTypeRef/mdTypeDef>
      CMOD_OPT         = 0x20,     // optional C modifier : E_T_CMOD_OPT <mdTypeRef/mdTypeDef>

      // This is for signatures generated internally (which will not be persisted in any way).
      INTERNAL         = 0x21,     // INTERNAL <typehandle>

      // Note that this is the max of base type excluding modifiers
      MAX              = 0x22,     // first invalid element type


      MODIFIER         = 0x40,
      SENTINEL         = 0x01 | MODIFIER, // sentinel for varargs
      PINNED           = 0x05 | MODIFIER,
      R4_HFA           = 0x06 | MODIFIER, // used only internally for R4 HFA types
      R8_HFA           = 0x07 | MODIFIER, // used only internally for R8 HFA types

   }

   //*****************************************************************************
   //
   // Serialization types for Custom attribute support
   //
   //*****************************************************************************

   // From corhdr.h
   enum CorSerializationType
   {
      UNDEFINED      = 0,
      BOOLEAN        = CorElementType.BOOLEAN,
      CHAR           = CorElementType.CHAR,
      I1             = CorElementType.I1,
      U1             = CorElementType.U1,
      I2             = CorElementType.I2,
      U2             = CorElementType.U2,
      I4             = CorElementType.I4,
      U4             = CorElementType.U4,
      I8             = CorElementType.I8,
      U8             = CorElementType.U8,
      R4             = CorElementType.R4,
      R8             = CorElementType.R8,
      STRING         = CorElementType.STRING,
      SZARRAY        = CorElementType.SZARRAY, // Shortcut for single dimension zero lower bound array
      TYPE           = 0x50,
      TAGGED_OBJECT  = 0x51,
      FIELD          = 0x53,
      PROPERTY       = 0x54,
      ENUM           = 0x55
   }
}
