//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/examples/IR-Longevity-Morph-Version/ir-longevity.cs 2     2/24/06 8:29p Eaddy $
//
// Description:
//
//    IR Longevity plug-in.
//
// Remarks:
//
//    This sample demonstrates the following concepts:
//
//    . Plug-Ins
//    . Phases
//    . Extension objects
//    . Event delegates
//
//    The longevity of IR objects is tracked during compilation. We take note
//    of the birth and death phase for each object. The count of birth-death
//    pairs is printed on the console when the process terminates.
//
//    We keep separate instruction and operand counts, but the distinct kinds
//    within those hierarchies are conglomerated.
//
// Usage:
//
//    csc -r:phx.dll -t:library ir-longevity.cs
//
//    cl ... -d2plugin:ir-longevity-plug-in.dll
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/examples/IR-Longevity-Morph-Version/ir-longevity.cs $
 * 
 * 2     2/24/06 8:29p Eaddy
 * o Now uses new namespace for attributes
 */

namespace IRLongevity
{

//-----------------------------------------------------------------------------
//
// Description:
//
//    InstrBirthExtensionObject is attached to IR instructions to indicate the
//    phase in which they were constructed.
//
//-----------------------------------------------------------------------------

#if !USE_MORPHED_PHX

public class InstrBirthExtensionObject : Phx.IR.InstrExtensionObject
{
   public Phx.Phases.Phase BirthPhase;

   // Boilerplate code

   public override uint Id
   {
      get { return InstrBirthExtensionObject.id; }
      set { }
   }

   static public void Register()
   {
      InstrBirthExtensionObject.id =
         Phx.ExtensionObject.Register("IRLongevity.InstrBirthExtensionObject");
   }

   static public InstrBirthExtensionObject
   Get
   (
      Phx.IR.Instr instr
   )
   {
      return instr.FindExtensionObject(InstrBirthExtensionObject.id)
         as InstrBirthExtensionObject;
   }

   private static uint id;
}

#endif

//-----------------------------------------------------------------------------
//
// Description:
//
//    OpndBirthExtensionObject is attached to IR operands to indicate the
//    phase in which they were constructed.
//
//-----------------------------------------------------------------------------

public class OpndBirthExtensionObject : Phx.IR.OpndExtensionObject
{
   public Phx.Phases.Phase BirthPhase;

   // Boilerplate code

   public override uint Id
   {
      get { return OpndBirthExtensionObject.id; }
      set { }
   }

   static public void Register()
   {
      OpndBirthExtensionObject.id =
         Phx.ExtensionObject.Register("IRLongevity.OpndBirthExtensionObject");
   }

   static public OpndBirthExtensionObject
   Get
   (
      Phx.IR.Opnd opnd
   )
   {
      return opnd.FindExtensionObject(OpndBirthExtensionObject.id)
         as OpndBirthExtensionObject;
   }

   private static uint id;
}

//-----------------------------------------------------------------------------
//
// Description:
//
//    SweepPhase walks the IR to "kill" all instructions and operands.
//
// Remarks:
//
//    Because of the way Phoenix does memory management, the IR existing at
//    the end of function phase processing is not deleted. This phase is
//    placed after encoding as the final resting place for the IR.
//
//    This code is packaged as a phase for demonstration purposes. To be more
//    accurate and flexible, we'd use the DeleteUnitEvent. Doing that would
//    require a minor change to the accounting/output code since we'd no
//    longer have a phase to hang our hat on.
//
//-----------------------------------------------------------------------------

public class SweepPhase : Phx.Phases.Phase
{
#if (PHX_DEBUG_SUPPORT)
   static public Phx.Ctrls.CompCtrl DebugCtrl;
#endif

   private PlugIn plugIn;

   //--------------------------------------------------------------------------
   //
   // Description:
   //
   //    Static constructor for SweepPhase.
   //
   // Arguments:
   //
   //    config - A pointer to a Phx::Phases::PhaseConfig that provides
   //       properties for retrieving the initial phase list, and the pre- and
   //       post-phase hook lists.
   //    plugIn - A pointer to an IRLongevity.PlugIn.
   //
   // Returns:
   //
   //    A pointer to the new phase.
   //
   //--------------------------------------------------------------------------

   public static SweepPhase
   New
   (
      Phx.Phases.PhaseConfig config,
      IRLongevity.PlugIn     plugIn
   )
   {
      SweepPhase phase = new SweepPhase();

      phase.plugIn = plugIn;

      phase.Init(config, "IR Longevity Sweep");

#if (PHX_DEBUG_SUPPORT)
      phase.PhaseCtrl = SweepPhase.DebugCtrl;
#endif

      return phase;
   }

   //--------------------------------------------------------------------------
   //
   // Description:
   //
   //    Execute the SweepPhase.
   //
   // Arguments:
   //
   //    unit - The Phx.Unit to work on.
   //
   // Remarks:
   //
   //    We walk the IR instructions and operands invoking the plug-in's
   //    DeleteOpndEventHandler. This simulates the IR being deleted at the
   //    end of compilation.
   //
   //    Having "killed" all the IR, we tell the plug-in to remove its event
   //    handlers. This is to protect against some other phase creating or
   //    deleting IR after us.
   //
   //--------------------------------------------------------------------------

   protected override void
   Execute
   (
      Phx.Unit unit
   )
   {
      Phx.FuncUnit funcUnit = unit as Phx.FuncUnit;
      Phx.IR.Instr instr;
      Phx.IR.Opnd  opnd;

      for (instr = funcUnit.FirstInstr; instr != null; instr = instr.Next)
      {
         for (opnd = instr.DstOpnd; opnd != null; opnd = opnd.Next)
         {
            this.plugIn.DeleteOpndEventHandler(opnd);
         }

         for (opnd = instr.SrcOpnd; opnd != null; opnd = opnd.Next)
         {
            this.plugIn.DeleteOpndEventHandler(opnd);
         }

         this.plugIn.DeleteInstrEventHandler(instr);
      }

      this.plugIn.RemoveIREventHandlers(funcUnit);
   }
}

//-----------------------------------------------------------------------------
//
// Description:
//
//    Phase visitor to collect all the phases in an array list.
//
// Remarks:
//
//    We ignore phase lists.
//
//-----------------------------------------------------------------------------

public class PhaseVisitor : Phx.Phases.PhaseVisitor
{
   public System.Collections.ArrayList arrayList;

   public override void
   Visit
   (
      Phx.Phases.Phase phase
   )
   {
      arrayList.Add(phase);
   }
};

//-----------------------------------------------------------------------------
//
// Description:
//
//    The IR longevity plug-in.
//
// Remarks:
//
//    The longevity of IR objects is tracked during compilation. We take note
//    of the birth and death phase for each object. The count of birth-death
//    pairs is printed on the console when the process terminates.
//
//    We keep separate instruction and operand counts, but the distinct kinds
//    within those hierarchies are conglomerated.
//
//-----------------------------------------------------------------------------

public class PlugIn : Phx.PlugIn
{
   private Phx.DependencyObject TermEventDependencyObject;
   private Phx.DependencyObject NewUnitDependencyObject;
   private Phx.DependencyObject NewInstrDependencyObject;
   private Phx.DependencyObject DeleteInstrDependencyObject;
   private Phx.DependencyObject NewOpndDependencyObject;
   private Phx.DependencyObject DeleteOpndDependencyObject;

   private System.Collections.ArrayList PhaseArrayList;
   private uint [,] InstrLongevityArray;
   private uint [,] OpndLongevityArray;

   //--------------------------------------------------------------------------
   //
   // Description:
   //
   //    Register objects used by the IR longevity plug-in.
   //
   //--------------------------------------------------------------------------

   public void
   RegisterObjects()
   {
      #if !USE_MORPHED_PHX

      InstrBirthExtensionObject.Register();

      #endif

      OpndBirthExtensionObject.Register();

      Phx.Init.EndInitEvent.Insert(this.EndInit);

      TermEventDependencyObject =
         Phx.DependencyObject.New(ref Phx.Term.TermEventDependencyList,
            "IRLongevity", "*");

      NewUnitDependencyObject =
         Phx.DependencyObject.New(ref Phx.Unit.NewUnitEventDependencyList,
            "IRLongevity", "*");

      NewInstrDependencyObject =
         Phx.DependencyObject.New(ref Phx.Unit.NewInstrEventDependencyList,
            "IRLongevity", "*");

      DeleteInstrDependencyObject =
         Phx.DependencyObject.New(ref Phx.Unit.DeleteInstrEventDependencyList,
            "IRLongevity", "*");

      NewOpndDependencyObject =
         Phx.DependencyObject.New(ref Phx.Unit.NewOpndEventDependencyList,
            "IRLongevity", "*");

      DeleteOpndDependencyObject =
         Phx.DependencyObject.New(ref Phx.Unit.DeleteOpndEventDependencyList,
            "IRLongevity", "*");

#if (PHX_DEBUG_SUPPORT)

      SweepPhase.DebugCtrl = Phx.Ctrls.CompCtrl.New("irlongevitysweep",
         "Debug the IR longevity sweep phase",
         "ir-longevity.cs");

#endif
   }

   //--------------------------------------------------------------------------
   //
   // Description:
   //
   //    Add handlers for dependent static events.
   //
   //--------------------------------------------------------------------------

   private void EndInit()
   {
      Phx.Term.TermEvent.Insert(this.TermEventHandler,
         this.TermEventDependencyObject);
      Phx.Unit.NewUnitEvent.Insert(this.NewUnitEventHandler,
         this.NewUnitDependencyObject);
   }

   //--------------------------------------------------------------------------
   //
   // Description:
   //
   //    Add the SweepPhase after the Encoding phase.
   //
   // Arguments:
   //
   //    config - A pointer to a Phx::Phases::PhaseConfig that will own the
   //       phases.
   //
   //--------------------------------------------------------------------------

   public void
   BuildPhases
   (
      Phx.Phases.PhaseConfig config
   )
   {
      Phx.Phases.Phase encodingPhase;
      SweepPhase       sweepPhase = SweepPhase.New(config, this);

      encodingPhase = config.PhaseList.FindByName("Encoding");
      if (encodingPhase != null)
      {
         encodingPhase.InsertAfter(sweepPhase);
      }
      else
      {
         Phx.Output.WriteLine("Encoding phase not found in phaselist:");
         Phx.Output.Write(config.ToString());
         Phx.Output.WriteLine("Appending.");

         // Didn't find the encoding phase; simply append to the phaselist.

         config.PhaseList.AppendPhase(sweepPhase);
      }

      // Assume the phase list is now complete and walk it to compute the
      // number of phases.

      PhaseVisitor visitor = new PhaseVisitor();
      visitor.arrayList = new System.Collections.ArrayList(30);
      visitor.WalkPhases(config.PhaseList);
      this.PhaseArrayList =
         System.Collections.ArrayList.ReadOnly(visitor.arrayList);

      int numPhases = this.PhaseArrayList.Count;
      this.InstrLongevityArray = new uint[numPhases,numPhases];
      this.OpndLongevityArray = new uint[numPhases,numPhases];
   }

   //--------------------------------------------------------------------------
   //
   // Description:
   //
   //    Handler for the NewUnitEvent.
   //
   // Arguments:
   //
   //    unit - The unit that was just constructed.
   //
   // Remarks:
   //
   //    We only care about FuncUnits. When we see one, we add our IR event
   //    handlers to the FuncUnit's event lists.
   //
   //--------------------------------------------------------------------------

   private void
   NewUnitEventHandler
   (
      Phx.Unit unit
   )
   {
      Phx.FuncUnit funcUnit = unit as Phx.FuncUnit;

      if (funcUnit == null)
      {
         return;
      }

      // Add our IR event handlers.

      funcUnit.NewInstrEvent.Insert(this.NewInstrEventHandler,
         this.NewInstrDependencyObject);
      funcUnit.DeleteInstrEvent.Insert(this.DeleteInstrEventHandler,
         this.DeleteInstrDependencyObject);

      funcUnit.NewOpndEvent.Insert(this.NewOpndEventHandler,
         this.NewOpndDependencyObject);
      funcUnit.DeleteOpndEvent.Insert(this.DeleteOpndEventHandler,
         this.DeleteOpndDependencyObject);
   }

   //--------------------------------------------------------------------------
   //
   // Description:
   //
   //    Remove our IR event handlers from a FuncUnit's event lists.
   //
   // Arguments:
   //
   //    unit - The unit from which to remove our events handlers.
   //
   //--------------------------------------------------------------------------

   public void
   RemoveIREventHandlers
   (
      Phx.FuncUnit funcUnit
   )
   {
      // Remove our IR event handlers.

      funcUnit.NewInstrEvent.Remove(this.NewInstrEventHandler);
      funcUnit.DeleteInstrEvent.Remove(this.DeleteInstrEventHandler);

      funcUnit.NewOpndEvent.Remove(this.NewOpndEventHandler);
      funcUnit.DeleteOpndEvent.Remove(this.DeleteOpndEventHandler);
   }

   //--------------------------------------------------------------------------
   //
   // Description:
   //
   //    Handler for the NewInstrEvent.
   //
   // Arguments:
   //
   //    instr - The instruction being constructed.
   //
   // Remarks:
   //
   //    We create an InstrBirthExtensionObject to annotate the
   //    instruction. It will contain the phase currently executing, which
   //    indicates when the instruction was born.
   //
   //--------------------------------------------------------------------------

   private void
   NewInstrEventHandler
   (
      Phx.IR.Instr instr
   )
   {
      #if USE_MORPHED_PHX

      instr.BirthPhase = instr.FuncUnit.Phase;

      #else

      InstrBirthExtensionObject extObj = new InstrBirthExtensionObject();
      extObj.BirthPhase = instr.FuncUnit.Phase;
      instr.AddExtensionObject(extObj);

      #endif
   }

   //--------------------------------------------------------------------------
   //
   // Description:
   //
   //    Handler for the DeleteInstrEvent.
   //
   // Arguments:
   //
   //    instr - The instruction being deleted.
   //
   // Remarks:
   //
   //    We retrieve the instruction's birth phase from the attached
   //    InstrBirthExtensionObject. The current phase is noted as the death
   //    phase of the instruction. The InstrBirthExtensionObject is destroyed
   //    in the process, having served its purpose.
   //
   //--------------------------------------------------------------------------

   public void
   DeleteInstrEventHandler
   (
      Phx.IR.Instr instr
   )
   {
      #if USE_MORPHED_PHX

      this.NoteLifetime(this.InstrLongevityArray, instr.BirthPhase,
         instr.FuncUnit.Phase);

      #else

      InstrBirthExtensionObject extObj = InstrBirthExtensionObject.Get(instr);
      this.NoteLifetime(this.InstrLongevityArray, extObj.BirthPhase,
         instr.FuncUnit.Phase);
      instr.RemoveExtensionObject(extObj);

      #endif
   }

   //--------------------------------------------------------------------------
   //
   // Description:
   //
   //    Handler for the NewOpndEvent.
   //
   // Arguments:
   //
   //    opnd - The operand being constructed.
   //
   // Remarks:
   //
   //    We create a OpndBirthExtensionObject to annotate the operand. It will
   //    contain the phase currently executing, which indicates when the
   //    operand was born.
   //
   //--------------------------------------------------------------------------

   private void
   NewOpndEventHandler
   (
      Phx.IR.Opnd opnd
   )
   {
      OpndBirthExtensionObject extObj = new OpndBirthExtensionObject();
      extObj.BirthPhase = opnd.FuncUnit.Phase;
      opnd.AddExtensionObject(extObj);
   }

   //--------------------------------------------------------------------------
   //
   // Description:
   //
   //    Handler for the DeleteOpndEvent.
   //
   // Arguments:
   //
   //    opnd - The operand being deleted.
   //
   // Remarks:
   //
   //    We retrieve the operand's birth phase from the attached
   //    OpndBirthExtensionObject. The current phase is noted as the death
   //    phase of the operand. The InstrBirthExtensionObject is destroyed in
   //    the process, having served its purpose.
   //
   //--------------------------------------------------------------------------

   public void
   DeleteOpndEventHandler
   (
      Phx.IR.Opnd opnd
   )
   {
      OpndBirthExtensionObject extObj = OpndBirthExtensionObject.Get(opnd);
      this.NoteLifetime(this.OpndLongevityArray, extObj.BirthPhase,
         opnd.FuncUnit.Phase);
      opnd.RemoveExtensionObject(extObj);
   }

   //--------------------------------------------------------------------------
   //
   // Description:
   //
   //    Make note of the lifetime of an object.
   //
   // Arguments:
   //
   //    longevityArray - The longevity array to update.
   //    birthPhase - The phase the object was born in.
   //    deathPhase - The phase the object died in.
   //
   // Remarks:
   //
   //    We are keeping a simple count of birth-death pairs.
   //
   //--------------------------------------------------------------------------

   private void
   NoteLifetime
   (
      uint [,]         longevityArray,
      Phx.Phases.Phase birthPhase,
      Phx.Phases.Phase deathPhase
   )
   {
      if (birthPhase == null)
      {
         // START/END bracketing IR don't have a birth phase.

         return;
      }

      int birthIndex = this.PhaseArrayList.IndexOf(birthPhase);
      int deathIndex = this.PhaseArrayList.IndexOf(deathPhase);

      longevityArray[birthIndex,deathIndex]++;
   }

   //--------------------------------------------------------------------------
   //
   // Description:
   //
   //    Termination event handler to display longevity information.
   //
   // Arguments:
   //
   //    mode - unused.
   //
   // Remarks:
   //
   //    We display the birth-death information collected during compilation.
   //
   //--------------------------------------------------------------------------

   private void
   TermEventHandler
   (
      Phx.Term.Mode mode
   )
   {
      this.DisplayLongevityArray("Instruction", this.InstrLongevityArray);
      System.Console.WriteLine();
      this.DisplayLongevityArray("Operand", this.OpndLongevityArray);
   }

   //--------------------------------------------------------------------------
   //
   // Description:
   //
   //    Display longevity information in tabular format.
   //
   // Arguments:
   //
   //    typeName - Name for the type of longevity info being displayed.
   //    longevityArray - Array of longevity counts.
   //
   // Remarks:
   //
   //    Death phases are displayed across the top with birth phases down the
   //    left. Immediately after the birth phase name is total count of
   //    objects created by that phase. At the intersection of a birth-death
   //    phase pair is the number of objects created in the birth phase and
   //    deleted in the death phase.
   //
   //    Sample dump follows (gathered from compilation of 8queens.c)
   //
   //    Instruction longevity:
   //                                   CIL Reader
   //                                  /   MIR Lower
   //                                 /   /   Ssa Info Destruction
   //                                /   /   /   Address Mode Builder
   //                               /   /   /   /   Flow Optimization
   //                              /   /   /   /   /   Encoding
   //                             /   /   /   /   /   /   IR Longevity Sweep
   //                            /   /   /   /   /   /   /
   //    CIL Reader     ( 188)   4  28   2  13  34     107
   //    Loop recogniti (   4)                   4
   //    Ssa Info Const (   8)           8
   //    Canon          (   3)                           3
   //    Lower          (  44)                   2   7  35
   //    Frame Generati (  10)                          10
   //    Block Layout   (  15)                  15
   //
   //--------------------------------------------------------------------------

   private void
   DisplayLongevityArray
   (
      string   typeName,
      uint [,] longevityArray
   )
   {
      Phx.Phases.Phase birthPhase;
      Phx.Phases.Phase deathPhase;
      int              numPhases = this.PhaseArrayList.Count;
      int              birthIndex;
      int              deathIndex;
      bool []          phaseHasDeathInfo = new bool[numPhases];
      int              numPhasesWithDeathInfo = 0;

      // Determine which phases have object deaths in them.

      for (deathIndex = 0; deathIndex < numPhases ; deathIndex++)
      {
         for (birthIndex = 0; birthIndex < numPhases ; birthIndex++)
         {
            if (longevityArray[birthIndex,deathIndex] != 0)
            {
               numPhasesWithDeathInfo++;
               phaseHasDeathInfo[deathIndex] = true;
               break;
            }
         }
      }

      System.Console.WriteLine("{0} longevity:", typeName);

      // Display death phases across the top with angle lines.

      int deathInfoIndex = 0;
      for (deathIndex = 0; deathIndex < numPhases ; deathIndex++)
      {
         if (!phaseHasDeathInfo[deathIndex])
         {
            continue;
         }

         deathPhase = this.PhaseArrayList[deathIndex] as Phx.Phases.Phase;

         System.Console.Write("{0,24}", " ");

         for (int i = 0; i < numPhasesWithDeathInfo - deathInfoIndex; i++)
         {
            System.Console.Write(" ");
         }

         for (int i = 0; i < deathInfoIndex; i++)
         {
            System.Console.Write("/   ");
         }

         System.Console.WriteLine(deathPhase.NameString);

         deathInfoIndex++;
      }

      System.Console.Write("{0,24}", " ");

      for (int i = 0; i < deathInfoIndex; i++)
      {
         System.Console.Write("/   ");
      }

      System.Console.WriteLine();

      // Display birth phases down the left side with death counts across.

      deathInfoIndex = 0;
      for (birthIndex = 0; birthIndex < numPhases ; birthIndex++)
      {
         uint birthCount = 0;

         // Prescan to count all births.

         for (deathIndex = birthIndex; deathIndex < numPhases ; deathIndex++)
         {
            birthCount += longevityArray[birthIndex,deathIndex];
         }

         // Ignore 0 birth phases.

         if (birthCount != 0)
         {
            birthPhase = this.PhaseArrayList[birthIndex] as Phx.Phases.Phase;

            string truncString = birthPhase.NameString;

            if (truncString.Length > 14)
            {
               truncString = truncString.Substring(0, 14);
            }

            System.Console.Write("{0,-14} ({1,4})",
               truncString, birthCount);

            // Space over for each active death phase that is longer in play
            // (because birthIndex is beyond it).

            for (int i = 0; i < deathInfoIndex; i++)
            {
               System.Console.Write("    ");
            }

            for (deathIndex = birthIndex; deathIndex < numPhases ; deathIndex++)
            {
               uint count = longevityArray[birthIndex,deathIndex];

               if (count == 0)
               {
                  // There is no longevity info for this pair of phases, we
                  // still need to space over if the death phase has any
                  // activity.

                  if (phaseHasDeathInfo[deathIndex])
                  {
                     System.Console.Write("    ");
                  }

                  continue;
               }

               System.Console.Write("{0,4}", count);
            }

            System.Console.WriteLine();
         }

         if (phaseHasDeathInfo[birthIndex])
         {
            deathInfoIndex++;
         }
      }
   }
}

} // namespace IRLongevity
