@REM Use 'set _echo=on' to echo all the commands
@if not "%_ECHO%"=="on" @if not "%_ECHO%"=="ON" @echo off

REM
REM Performs a unit test.
REM
REM Environment required:
REM
REM    TARGET - Name (w/o ext) of the test program to execute (ex: hello)
REM    PROJECT_DIR - Source directory for the test
REM       (ex: c:\phx\rdk\samples\Morpher\samples\hello)
REM    InputAssembly - Path to the original (non-morphed) assembly
REM       (ex: orig.dll)
REM
REM Options:
REM
REM     -quiet      Supresses output of non error messages
REM     -windiff    Show visual diff using windiff
REM     -vf         Show visual diff using vf
REM     -debug      Show visual diff
REM     -show       Builds and executes the sample and shows the output
REM                 but doesn't compare it with the baseline
REM     -demo       Pauses between significant steps and supresses debug msgs
REM     -baseline   Used to bootstrap the test process. Runs the
REM                 test and create baseline files based on the test output
REM                 and metadata output. You should manually verify the
REM                 baseline files.
REM     -noildasm   Does not compare metadata files
REM     -nobuild    Skips the build process and just runs the test
REM
REM Returns:
REM
REM	0 if test succeeds, 1 otherwise
REM
REM Ex:
REM
REM	dotest.bat -quiet
REM

setlocal

REM *****
REM ***** Verify that Phx environment has been setup
REM ***** 

if "%RDKROOT%"=="" echo Must setup Phx environment first && exit /B 1

REM *****
REM ***** Local vars
REM ***** 

set MORPH_DIR=%~dp0..
if '%TARGET_DIR%'=='' set TARGET_DIR=%MORPH_DIR%\bin

REM *****
REM ***** Parse command-line options
REM ***** 

call "%MORPH_DIR%\scripts\checkswitch.bat" echo %*
if "%errorlevel%"=="1" (
	set _ECHO=on
	echo on
)

call "%MORPH_DIR%\scripts\checkswitch.bat" quiet %*
set QUIET=%errorlevel%

call "%MORPH_DIR%\scripts\checkswitch.bat" demo %*
set DEMO=%errorlevel%

REM The INDENT variable looks empty but actually contains spaces
if "%DEMO%"=="1" (
	set INDENT=      
) else (
	set INDENT=
)

call "%MORPH_DIR%\scripts\checkswitch.bat" show %*
set SHOW=%errorlevel%

call "%MORPH_DIR%\scripts\checkswitch.bat" baseline %*
set CREATE_BASELINE=%errorlevel%

call "%MORPH_DIR%\scripts\checkswitch.bat" windiff %*
if "%errorlevel%"=="1" (
	set SHOW_VISUAL_DIFF=windiff
)

call "%MORPH_DIR%\scripts\checkswitch.bat" vf %*
if "%errorlevel%"=="1" (
	set SHOW_VISUAL_DIFF=vf
)

call "%MORPH_DIR%\scripts\checkswitch.bat" debug %*
if "%errorlevel%"=="1" (
	set SHOW_VISUAL_DIFF=windiff
)

call "%MORPH_DIR%\scripts\checkswitch.bat" noildasm %*
set NO_ILDASM=%errorlevel%

call "%MORPH_DIR%\scripts\checkswitch.bat" nobuild %*
set NO_BUILD=%errorlevel%

REM *****
REM ***** Verify required environment vars
REM ***** 

call "%MORPH_DIR%\scripts\requireenv.bat" TARGET %TARGET%
if not "%errorlevel%"=="0" exit /B 1

call "%MORPH_DIR%\scripts\requireenv.bat" PROJECT_DIR "%PROJECT_DIR%"
if not "%errorlevel%"=="0" exit /B 1

call "%MORPH_DIR%\scripts\requireenv.bat" InputAssembly %InputAssembly%
if not "%errorlevel%"=="0" exit /B 1

set DIFF_PROG=fc
REM if not "%DIFF%"=="" set DIFF_PROG=%DIFF%

REM *****
REM ***** Local vars
REM ***** 

call :removequotes '%TEST_EXE%'
set BLAH=%RESULT%
if "%BLAH%"=="''" set TEST_EXE="%TARGET%.exe"

set OUT=%TARGET%.out

if '%OutputAssembly%'=='' set OutputAssembly=%InputAssembly%

call :getfilenamewthoutext %InputAssembly%
set BEFORE_OUT=%RESULT%.orig.md
set AFTER_OUT=%RESULT%.new.md

REM *****
REM ***** Build test files
REM *****

if not "%NO_BUILD%"=="1" call "%PROJECT_DIR%\buildall.bat" %*
if not "%NO_BUILD%"=="1" if not "%errorlevel%"=="0" goto :resultfail

REM *****
REM ***** Verify dirs and exes
REM *****

if not exist "%TARGET_DIR%" (
	echo Test directory not found: "%TARGET_DIR%"
	goto :resultfail
)

cd "%TARGET_DIR%"
if not "%errorlevel%"=="0" goto :resultfail

REM *****
REM ***** Execute test
REM *****

if exist "%OUT%" del "%OUT%"

if "%DEMO%"=="1" (
	echo.
	echo ===== Test the client =====
	echo.
	set /p ANS=%INDENT%Press enter to continue...
	echo.
	echo %INDENT%Executing %TEST_EXE%:
	echo.
)

%TEST_EXE% 2>&1 >> "%OUT%"

if not "%errorlevel%" == "0" goto :resultfail

call "%MORPH_DIR%\scripts\StripTokens.bat" "%OUT%"

if "%CREATE_BASELINE%"=="1" (

	if "%QUIET%"=="0" echo.
	if "%QUIET%"=="0" echo Initializing baseline files from current output files
	if "%QUIET%"=="0" echo.

	call :CopyBaseline "%OUT%" "%PROJECT_DIR%\%OUT%.baseline"
	if "%NO_ILDASM%"=="0" (
		call :CopyBaseline "%BEFORE_OUT%" "%PROJECT_DIR%\%BEFORE_OUT%.baseline"
		call :CopyBaseline "%AFTER_OUT%" "%PROJECT_DIR%\%AFTER_OUT%.baseline"
	)

	REM Return an error so no one confuses this with a real test
	exit /B 1
)

REM *****
REM ***** Validate test output
REM *****

if "%SHOW%"=="1" (
	echo.
	echo Output:
	echo.
)

if "%DEMO%"=="0" (

	if "%SHOW%"=="0" (

		if "%QUIET%"=="0" echo %INDENT%Comparing test output:

		if exist "%PROJECT_DIR%\%OUT%.%BUILD%.baseline" (
			call :CompareAgainstBaseline "%OUT%" "%PROJECT_DIR%\%OUT%.%BUILD%.baseline"
		) else (
			call :CompareAgainstBaseline "%OUT%" "%PROJECT_DIR%\%OUT%.baseline"
		)

		if "%NO_ILDASM%"=="0" (

			if "%QUIET%"=="0" echo %INDENT%Comparing pre-morph metadata:
			call :CompareAgainstBaseline "%BEFORE_OUT%" "%PROJECT_DIR%\%BEFORE_OUT%.baseline" strip_orig

			if "%QUIET%"=="0" echo %INDENT%Comparing post-morph metadata:
			if exist "%PROJECT_DIR%\%AFTER_OUT%.%BUILD%.baseline" (
				call :CompareAgainstBaseline "%AFTER_OUT%" "%PROJECT_DIR%\%AFTER_OUT%.%BUILD%.baseline" strip_orig
			) else (
				call :CompareAgainstBaseline "%AFTER_OUT%" "%PROJECT_DIR%\%AFTER_OUT%.baseline" strip_orig
			)
		)

	) else (
		type "%OUT%"
	)
) else (
	type "%OUT%"
)

if "%FAILED%"=="1" goto :resultfail

exit /B 0

:CopyBaseline

if exist "%2" (
	echo Baseline already exists: %2
) else (

	if exist "%1" (
		echo Copying baseline: %2
		copy %1 %2 > nul
	) else (
	
		color 0c
		echo Error: Baseline not found: %1
		color
	)
)
goto :eof

:CompareAgainstBaseline

REM Verify that the baseline file exists
if not exist %2 (
	color 0c
	echo   Error: Expected test baseline file not found:
	echo   %2
	color
	exit /B 1
)

set FILE2="%~2"

REM Strip out tokens and Public Keys from the orig metadata
if "%3"=="strip_orig" (
	set FILE2="%~2.stripped"
	copy %2 "%~2.stripped" > nul
	call "%MORPH_DIR%\scripts\StripTokens.bat" "%~2.stripped"
	set errorlevel=0
)

REM Do the file comparison
%DIFF_PROG% %1 %FILE2% > %1.diff
if "%errorlevel%"=="0" (
	if "%QUIET%"=="0" echo %INDENT%  PASSED
	exit /B 0
) else (
	color 0c
	echo.
	type %1.diff

	if "%QUIET%"=="0" echo %INDENT%  FAILED
	color

	if not "%SHOW_VISUAL_DIFF%"=="" start %SHOW_VISUAL_DIFF% %1 %FILE2%

	set FAILED=1

	exit /B 1
)
goto :eof

:resultfail

if "%QUIET%"=="0" (
	color 0c
	echo.
	echo ========== Test Results: 1 Failed ==========
	color
)

exit /B 1

:getfilenamewthoutext
set RESULT=%~n1
goto :eof

:getext
set RESULT=%~x1
goto :eof

:removequotes
set RESULT=%~1
goto :eof