//---------------------------------------------------------------------
//  This file is part of the CLR Managed Debugger (mdbg) Sample.
// 
//  Copyright (C) Microsoft Corporation.  All rights reserved.
//---------------------------------------------------------------------
using System;
using System.IO;
using System.Collections;
using System.Reflection;
using System.Diagnostics;
using System.Text;

using Microsoft.Samples.Debugging.CorDebug;
using Microsoft.Samples.Debugging.CorMetadata;
using Microsoft.Samples.Debugging.CorDebug.NativeApi;


namespace Microsoft.Samples.Debugging.MdbgEngine
{
    /// <summary>
    /// MDbg Thread class.
    /// </summary>
    public sealed class MDbgThread : MarshalByRefObject, IComparable
    {
        internal MDbgThread(MDbgThreadCollection threadCollection,CorThread thread,int threadNumber)
        {
            m_corThread = thread;
            m_threadNumber = threadNumber;
            m_threadMgr = threadCollection;

            m_logicalStopClock = -1;
        }

        /// <summary>
        /// Gets or Sets if the Thread is Suspended.
        /// </summary>
        /// <value>NotImplementedException.</value>
        public bool Suspended
        {
            get
            {
                throw new NotImplementedException();
            }
            set
            {
                throw new NotImplementedException();
            }
        }

        /// <summary>
        /// Gets the Thread Number.
        /// </summary>
        /// <value>The Thread Number.</value>
        public int Number
        {
            get
            {
                return m_threadNumber;
            }
        }

        /// <summary>
        /// Gets the Thread Id.
        /// </summary>
        /// <value>Gets the Thread Id.</value>
        public int Id
        {
            get
            {
                return m_corThread.Id;
            }
        }

        /// <summary>
        /// Gets the CorThread.
        /// </summary>
        /// <value>The CorThread.</value>
        public CorThread CorThread
        {
            get
            {
                return m_corThread;
            }
        }

        /// <summary>
        /// Returns current exception on the thread or MDbgValue representing N/A if there is no exception on the thread.
        /// </summary>
        /// <value>The current Exception.</value>
        public MDbgValue CurrentException
        {
            get
            {
                CorValue cv;
                try
                {
                    cv = CorThread.CurrentException;
                }
                catch(System.Runtime.InteropServices.COMException e)
                {
                    if(e.ErrorCode == (int)HResult.E_FAIL) 
                        cv = null;
                    else
                        throw;
                }
                return new MDbgValue(m_threadMgr.m_process,cv);
            }
        }

        /// <summary>
        /// The Bottom Frame in the stack for this Thread.
        /// </summary>
        /// <value>The Bottom Frame.</value>
        public MDbgFrame BottomFrame
        {
            get
            {
                if(m_logicalStopClock!=m_threadMgr.m_logicalStopClock)
                    RefreshFrameInfo();

                return m_bottomFrame;
            }
        }

        /// <summary>
        /// Returns all the frames for the thread.
        /// </summary>
        /// <value>The frames.</value>
        public IEnumerable Frames
        {
            get
            {
                return new MDbgFrameEnumerable(BottomFrame);
            }
        }

        private class MDbgFrameEnumerable: IEnumerable
        {
            public MDbgFrameEnumerable(MDbgFrame bottomFrame)
            {
                m_bottomFrame = bottomFrame;
            }
            
            public IEnumerator GetEnumerator()
            {
                return new MDbgFrameEnumerator(m_bottomFrame);
            }
            MDbgFrame m_bottomFrame;
        }

        /// <summary>
        /// Gets or Sets the Current Logical Frame
        /// </summary>
        /// <value>The Current Frame.</value>
        public MDbgFrame CurrentFrame
        {
            get
            {
                if( m_logicalStopClock!=m_threadMgr.m_logicalStopClock )
                    RefreshFrameInfo();
                if( m_currentFrame==null )
                    throw new MDbgNoCurrentFrameException();

                Debug.Assert( !m_currentFrame.IsInfoOnly );
                return m_currentFrame;
            }
            set
            {
                if( m_logicalStopClock!=m_threadMgr.m_logicalStopClock )
                    throw new InvalidOperationException("CurrentFrame can be only set to a frame being read!");

                if( value==null ||
                    value.Thread != this )
                    throw new ArgumentException();
                if( value.IsInfoOnly )
                    throw new InvalidOperationException("virtual frames cannot be set as current frames");
                m_currentFrame = value;
            }
        }

        /// <summary>
        /// Gets if the Thread has a Current Logical Frame.
        /// </summary>
        /// <value>true if it has, else false.</value>
        public bool HaveCurrentFrame
        {
            get
            {
                if(m_logicalStopClock!=m_threadMgr.m_logicalStopClock)
                    RefreshFrameInfo();
                return m_currentFrame!=null;
            }
        }

        /// <summary>
        /// Gets the current Source Position.
        /// </summary>
        /// <value>The Source Ponition.</value>
        public MDbgSourcePosition CurrentSourcePosition
        {
            get
            {
                return CurrentFrame.SourcePosition;
            }
        }

        /// <summary>
        /// Moves the Current Frame up or down.
        /// </summary>
        /// <param name="down">Moves frame down if true, else up.</param>
        public void MoveCurrentFrame(bool down)
        {
            MDbgFrame f = CurrentFrame;
            Debug.Assert( !f.IsInfoOnly );

            do
            {
                if(down)
                    f = f.NextDown;
                else
                    f = f.NextUp;
            }
            while( f!=null && f.IsInfoOnly );

            if( f==null )
                throw new MDbgException("Operation hit "+(down?"bottom":"top") + " of the stack.");
            
            m_currentFrame = f;
        }

        private void RefreshFrameInfo()
        {
            // we should only refresh when needed
            Debug.Assert(m_logicalStopClock<m_threadMgr.m_logicalStopClock);

            if( m_threadMgr.m_process.m_engine.Options.InteropCallStack &&
                m_threadMgr.m_process.m_engine.m_nativeDebuggerProvider!=null )
            {
                m_threadMgr.m_process.m_engine.m_nativeDebuggerProvider.RefreshStack();
            }
    
            CorChain c = null;
            try {
                c = m_corThread.ActiveChain;
            }
            catch( System.Runtime.InteropServices.COMException ce )
            {
                // Sometimes we cannot get the callstack.  For example, the thread
                // may not be scheduled yet (CORDBG_E_THREAD_NOT_SCHEDULED),
                // or the CLR may be corrupt (CORDBG_E_BAD_THREAD_STATE).
                // In either case, we'll ignore the problem and return an empty callstack.
                Debug.Assert( ce.ErrorCode == (int) HResult.CORDBG_E_BAD_THREAD_STATE ||
                    ce.ErrorCode == (int) HResult.CORDBG_E_THREAD_NOT_SCHEDULED );
            }

            if( m_threadMgr.m_process.m_engine.Options.InteropCallStack &&
                m_threadMgr.m_process.m_engine.m_nativeDebuggerProvider!=null &&
                c!=null && !c.IsManaged)
            {
                 m_bottomFrame = m_currentFrame = 
				m_threadMgr.m_process.m_engine.m_nativeDebuggerProvider.GetFrameFromChain(this, c, true);
            }
            else
            {
	            while(c!=null && 
	                  (!c.IsManaged || (c.IsManaged && c.ActiveFrame==null)) 
	                  )
	                c = c.Caller;

	            if(c==null)
	            {
	                m_bottomFrame = m_currentFrame = null; 
	                return;
	            }
	            Debug.Assert(c!=null && c.IsManaged);
	            m_bottomFrame = new MDbgILFrame(this,c.ActiveFrame);
            }            

            // set m_currentFrame to first non-virtual frame
            MDbgFrame f = m_bottomFrame;
            while( f!=null
                   && f.IsInfoOnly )
                f = f.NextUp;
            m_currentFrame = f;
            
            m_logicalStopClock = m_threadMgr.m_logicalStopClock;
        }

        /// <summary>
        /// This will return an MDbgFrame for the corresponding CorFrame.
        /// Note (the frame needs to correspond to this thread!!!!)
        /// </summary>
        /// <param name="f">The CorFrame to look up.</param>
        /// <returns>The coresponding MDbgFrame.</returns>
        public MDbgFrame LookupFrame(CorFrame f)
        {
            Debug.Assert(f!=null);
            return new MDbgILFrame(this,f);
        }

        /// <summary> 
        /// This function is called from implementor of MDbgFrame to get next frame in a Chain
        /// </summary>
        /// <param name="chain"> The chain that we are retrieving the frame from.</param>
        /// <param name="moveUp"> Indicates if the frame to be returned should be caller or callee.</param>        
        /// <returns> MDbgFrame object </returns>
        public MDbgFrame GetFrameFromNextChain( CorChain chain, bool moveUp )
        {
            while(true)
            {
                chain = moveUp?chain.Caller:chain.Callee;
                
                if( chain==null )                       // bottom frame
                    return null;
                if(  chain.IsManaged )
                {
                    MDbgFrame f = MDbgILFrame.GetFrameFromChain(this,chain,moveUp);
                    if( f!=null )
                        return f;
                }
                else
                {
                    if( m_threadMgr.m_process.m_engine.Options.InteropCallStack &&
                        m_threadMgr.m_process.m_engine.m_nativeDebuggerProvider!=null )
                    {
                        // native frame
                        MDbgFrame f = m_threadMgr.m_process.m_engine.m_nativeDebuggerProvider.GetFrameFromChain(this,chain,moveUp);
                        if(f!=null)
                            return f;
                    }
                }
            }
        }

        int IComparable.CompareTo(object obj)
        {
            return this.Number - (obj as MDbgThread).Number;
        }
            
        //////////////////////////////////////////////////////////////////////////////////
        //
        // Local variables
        //
        //////////////////////////////////////////////////////////////////////////////////
        
        internal MDbgThreadCollection m_threadMgr;

        private CorThread m_corThread;
        private int m_threadNumber;

        private int m_logicalStopClock;
        private MDbgFrame m_bottomFrame,m_currentFrame;
    }

    

    class MDbgFrameEnumerator : IEnumerator
    {
        internal MDbgFrameEnumerator(MDbgFrame bottomFrame)
        {
            m_bottomFrame = bottomFrame;
        }

        void IEnumerator.Reset()
        {
            m_currentFrame = null;
        }
        
        bool IEnumerator.MoveNext()
        {
            if(m_currentFrame==null)
            {
                if(m_bottomFrame==null)
                    return false;
                
                m_currentFrame=m_bottomFrame;
            }
            else
            {
                MDbgFrame f = m_currentFrame.NextUp;
                if(f==null)
                    return false;
                m_currentFrame = f;
            }
            return true;
        }

        Object IEnumerator.Current 
        {
            get 
            {
                if(m_currentFrame==null)
                    throw new InvalidOperationException();
                return m_currentFrame;
            }
        }

        MDbgFrame m_bottomFrame;
        MDbgFrame m_currentFrame=null;
    }

    /// <summary>
    /// MDbg Thread Collection class.
    /// </summary>
    public sealed class MDbgThreadCollection : MarshalByRefObject,IEnumerable
    {
        internal MDbgThreadCollection(MDbgProcess process)
        {
            m_process = process;
            m_freeThreadNumber=0;
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            MDbgThread[] ret = new MDbgThread[m_items.Count];
            m_items.Values.CopyTo(ret,0);
            Array.Sort(ret);
            return ret.GetEnumerator();
        }

        /// <summary>
        /// How many threads are in the collection.
        /// </summary>
        /// <value>How many threads.</value>
        public int Count
        {
            get
            {
                return m_items.Count;
            }
        }

        /// <summary>
        /// Allows for indexing of the collection by thread number.
        /// </summary>
        /// <param name="threadNumber">Which thread number to access.</param>
        /// <returns>The MDbgThread at that number.</returns>
        public MDbgThread this[int threadNumber]
        {
            get
            {
                return GetThreadFromThreadNumber(threadNumber);
            }
        }

        /// <summary>
        /// Lookup a MDbgThread using a CorThread.
        /// </summary>
        /// <param name="thread">The CorThread to use.</param>
        /// <returns>The rusulting MDbgThread with the same ID.</returns>
        public MDbgThread Lookup(CorThread thread)
        {
            return GetThreadFromThreadId(thread.Id);
        }
        
        /// <summary>
        /// Gets on sets the Active Thread.
        /// </summary>
        /// <value>The Active Thread.</value>
        public MDbgThread Active
        {
            get
            {
                if(m_active==null)
                    throw new MDbgNoActiveInstanceException("No active thread");
                return m_active;
            }
            set
            {
                Debug.Assert(value!=null && 
                             m_items.Contains(value.CorThread.Id));
                if(value==null || !m_items.Contains(value.CorThread.Id))
                    throw new ArgumentException();
                m_active=value;
            }
        }

        /// <summary>
        /// Gets if the collection has an Active Thread.
        /// </summary>
        /// <value>true if it has an Active Thread, else false.</value>
        public bool HaveActive
        {
            get
            {
                return m_active!=null;
            }
        }

        /// <summary>
        /// This function needs to be called when we do "setip" so that callstack will get refreshed.
        /// </summary>
        public void RefreshStack()
        {
            m_logicalStopClock++;
        }

        internal void Register(CorThread t)
        {
            m_items.Add(t.Id,new MDbgThread(this,t,m_freeThreadNumber++));
        }

        internal void UnRegister(CorThread t)
        {
            m_items.Remove(t.Id);
            if( m_active!=null &&
                m_active.CorThread.Id == t.Id )
                m_active = null;
        }

        internal void Clear()
        {
            m_items.Clear();
            m_active=null;
            m_freeThreadNumber=0;
        }


        internal void SetActiveThread(CorThread thread)
        {
            if(thread==null)
                m_active=null;
            else
                m_active=GetThreadFromThreadId(thread.Id);
            m_logicalStopClock++;
        }

        /// <summary>
        /// Gets the MDbgThread with the given ThreadId.
        /// </summary>
        /// <param name="threadId">The ThreadId to look up.</param>
        /// <returns>The MDbgThread.</returns>
        public MDbgThread GetThreadFromThreadId(int threadId)
        {
            MDbgThread te = (MDbgThread)m_items[threadId];
            return te;
        }

        private MDbgThread GetThreadFromThreadNumber(int threadNumber)
        {
            foreach(MDbgThread te in m_items.Values)
                if(te.Number==threadNumber)
                    return te;
            return null;
        }
        
        private Hashtable m_items = new Hashtable();
        private int m_freeThreadNumber;
        private MDbgThread m_active;

        internal int m_logicalStopClock=0; // the MDbgThread needs to access it
        internal MDbgProcess m_process;
    }

    /// <summary>
    /// MDbg Frame class
    /// MDbgFrame is defined as abstract so that we can have
    /// different kind of frames.  There could be e.g. ManagedFrames
    /// (IL code) and NativeFrames (native) corresponding to native
    /// code.
    /// </summary>
    public abstract class MDbgFrame
    {
        /// <summary>
        /// Returns cor-wrapper for frame or throws an exception if N/A.
        /// </summary>
        /// <value>The cor-wrapper for the frame.</value>
        public abstract CorFrame CorFrame 
        {
            get;
        }

        /// <summary>
        /// Returns thread owning this frame.
        /// </summary>
        /// <value>The Thread.</value>
        public abstract MDbgThread Thread
        {
            get;
        }

        /// <summary>
        /// Return true if frame has a managed IL code behind it.
        /// </summary>
        /// <value>true if frame is managed, else false.</value>
        public abstract bool IsManaged
        {
            get;
        }

        /// <summary>
        /// Returns true if the frame is informational (i.e. informative/internal).
        /// </summary>
        /// <value>true if frame is informational, else false.</value>
        public abstract bool IsInfoOnly
        {
            get;
        }

        /// <summary>
        /// Returns source position for the frame or null if not available.
        /// </summary>
        /// <value>The Source Position.</value>
        public abstract MDbgSourcePosition SourcePosition
        {
            get;
        }

        /// <summary>
        /// Returns function whose code is executed in this frame.
        /// </summary>
        /// <value>The Function.</value>
        public abstract MDbgFunction Function
        {
            get;
        }

        /// <summary>
        /// Returns next frame up the stack or null if topmost frame.
        /// </summary>
        /// <value>The next frame up.</value>
        public abstract MDbgFrame NextUp
        {
            get;
        }

        /// <summary>
        /// Returns next frame down the stack or null if bottom frame.
        /// </summary>
        /// <value>The next frame up.</value>
        public abstract MDbgFrame NextDown
        {
            get;
        }

        /// <summary>
        /// Returns a string that represents current frame
        /// Currently supported formats:
        /// null or empty string: returns short frame format (just frame name)
        /// "v"                 : returns long frame format (including module &amp; arguments)
        /// </summary>
        /// <param name="format">Which format to use.</param>
        /// <returns>The formatted string that represtents the current frame.</returns>
        public abstract string ToString(string format);

        /// <summary>
        /// Equality testing.  Allows for things like "if(thing1 == thing2)" to work properly.
        /// </summary>
        /// <param name="operand">First Operand.</param>
        /// <param name="operand2">Second Operand.</param>
        /// <returns>true if equal, else false.</returns>
        public static bool operator ==( MDbgFrame operand,MDbgFrame operand2)
        {
            if(Object.ReferenceEquals(operand,operand2))
                return true;
            return operand.Equals(operand2);
        }

        /// <summary>
        /// Inequality testing.  Allows for things like "if(thing1 != thing2)" to work properly.
        /// </summary>
        /// <param name="operand">First Operand.</param>
        /// <param name="operand2">Second Operand.</param>
        /// <returns>true if not equal, else false.</returns>
        public static bool operator !=( MDbgFrame operand,MDbgFrame operand2)
        {
            return !(operand==operand2);
        }

        /// <summary>
        /// Determines whether the specified object is equal to the other object.
        /// </summary>
        /// <param name="value">The object to compare with the current frame.</param>
		public abstract override bool Equals(Object value);

        /// <summary>
        /// Returns the hash code for this instance.
        /// </summary>
        /// <returns>A 32-bit signed integer hash code.</returns>
		public abstract override int GetHashCode();
    }
    /// <summary>
    /// The MDbgILFrame class is a frame with Intermediate Language code behind it.
    /// </summary>
    public sealed class MDbgILFrame : MDbgFrame
    {

		// @TODO - comment this
        internal MDbgILFrame(MDbgThread thread,CorFrame frame)
        {
            Debug.Assert(thread!=null && frame!=null);
            m_thread = thread;
            m_frame = frame;
        }

        /// <summary>
        /// Determines if the wrapped object is equal to another.
        /// </summary>
        /// <param name="value">The object to compare to.</param>
        /// <returns>true if equal, else false.</returns>
        public override bool Equals(Object value) 
        {
            if(!(value is MDbgILFrame))
                return false;
            return ((value as MDbgILFrame).m_frame== this.m_frame);
        }

        /// <summary>
        /// Required to implement MarshalByRefObject.
        /// </summary>
        /// <returns>Hash Code.</returns>
        public override int GetHashCode() 
        {
            return m_frame.GetHashCode();
        }

        /// <summary>
        /// Returns cor-wrapper for frame or throws an exception if N/A.
        /// </summary>
        /// <value>The cor-wrapper for the frame.</value>
        public override CorFrame CorFrame
        {
            get
            {
                return m_frame;
            }
        }

        /// <summary>
        /// Returns thread owning this frame.
        /// </summary>
        /// <value>The Thread.</value>
        public override MDbgThread Thread
        {
            get
            {
                return m_thread;
            }
        }

        /// <summary>
        /// Return true if frame has a managed IL code behind it.
        /// </summary>
        /// <value>true if frame is managed, else false.</value>
        public override bool IsManaged
        {
            get
            {
                return ( m_frame.FrameType == CorFrameType.ILFrame );
            }
        }

        /// <summary>
        /// Returns true if the frame is informational (i.e. informative/internal).
        /// </summary>
        /// <value>true if frame is informational, else false.</value>
        public override bool IsInfoOnly
        {
            get
            {
                return ( m_frame.FrameType == CorFrameType.InternalFrame );
            }
        }

        /// <summary>
        /// Returns source position for the frame or null if not available.
        /// </summary>
        /// <value>The Source Position.</value>
        public override MDbgSourcePosition SourcePosition
        {
            get
            {
                MDbgFunction f = null;

                // we can return source position only for ILFrames.
                if( IsManaged )
                    f = Function;
                
                return (f==null)?null:f.GetSourcePositionFromFrame(m_frame);
            }
        }

        /// <summary>
        /// Returns function whose code is executed in this frame.
        /// </summary>
        /// <value>The Function.</value>
        public override MDbgFunction Function
        {
            get
            {
                return m_thread.m_threadMgr.m_process.Modules.LookupFunction(m_frame.Function);
            }
        }

        /// <summary>
        /// Returns next frame up the stack or null if topmost frame.
        /// </summary>
        /// <value>The next frame up.</value>
        public override MDbgFrame NextUp
        { 
            get
            {
                CorFrame f = m_frame.Caller;
                if( f!=null )
                    return new MDbgILFrame(m_thread,f);
                else
                    return m_thread.GetFrameFromNextChain(m_frame.Chain,true);
            } 
        }

        /// <summary>
        /// Returns next frame down the stack or null if bottom frame.
        /// </summary>
        /// <value>The next frame up.</value>
        public override MDbgFrame NextDown
        { 
            get
            { 
                CorFrame f = m_frame.Callee;
                if( f!=null )
                    return new MDbgILFrame(m_thread,f);
                else 
                    return m_thread.GetFrameFromNextChain(m_frame.Chain,false);
            }
        }

        /// <summary>
        /// Returns a string that represents current frame using default formatting.
        /// </summary>
        /// <returns>The default formatted string that represtents the current frame.</returns>
        public override string ToString()
        {
            return ToString(null);
        }

        /// <summary>
        /// Get a CorType that function for this frame is declared in.
        /// </summary>
        // 
        // If the frame is Class<int>::Func<string>,
        // get a CorType representing 'Class<int>'. 
        // The CorClass would only give us 'Class<T>', and the
        // Function only gives Func<U>
        // Returns 'null' for global methods.
        
        //
        public CorType FunctionType
        {
            get
            {
                MDbgFunction f = this.Function;
                CorClass c = f.CorFunction.Class;

                Debug.Assert( TokenUtils.TypeFromToken(c.Token) == CorTokenType.mdtTypeDef );
                
                // Check if we're the "global" class, in which case this is a global method, so return null.

                //  global type should return NULL token,
                // currently they are returning either 1 or 0 RID.
                if ( TokenUtils.RidFromToken(c.Token) == 1
                     || TokenUtils.RidFromToken(c.Token) == 0 )
                    return null;


                // ICorDebug API lets us always pass ET_Class
                CorElementType et = CorElementType.ELEMENT_TYPE_CLASS;

                // Get type parameters.
                IEnumerable typars = m_frame.TypeParameters;
                //IEnumerator tyenum = typars.GetEnumerator();

                MDbgModule module = this.Function.Module;
                int cNumTyParamsOnClass = module.Importer.CountGenericParams(c.Token);
                

                CorType [] args = new CorType[cNumTyParamsOnClass];
                int i = 0;
                foreach(CorType arg in typars)
                {
                    if (i == cNumTyParamsOnClass)
                    {
                        break;
                    }
                    
                    args[i] = arg;
                    i++;
                }

                CorType t = c.GetParameterizedType(et, args);
                return t;
            }
        }

        /// <summary> Return an enumerator that enumerates through just the generic args on this frame's method </summary>
        /// <remarks>Enumerating generic args on the frame gives a single collection including both
        /// class and function args. This is a helper to skip past the class and use just the funciton args.</remarks>
        public IEnumerable FunctionTypeParameters
        {
            get
            {
                MDbgModule module = this.Function.Module;
                CorClass c = this.Function.CorFunction.Class;
                int cNumTyParamsOnClass = module.Importer.CountGenericParams(c.Token);
                return m_frame.GetTypeParamEnumWithSkip(cNumTyParamsOnClass);
            }
        }



        /// <summary>
        /// Returns a string that represents current frame
        /// Currently supported formats:
        /// null or empty string: returns short frame format (just frame name)
        /// "v"                 : returns long frame format (including module &amp; arguments)
        /// </summary>
        /// <param name="format">Which format to use.</param>
        /// <returns>The formatted string that represtents the current frame.</returns>
        public override string ToString(string format)
        {
            string fn;

            switch (m_frame.FrameType)
            {
            case CorFrameType.ILFrame:
            case CorFrameType.NativeFrame:
                MDbgSourcePosition sl = SourcePosition;
                string sp;
                
                if(sl!=null)
                {
                    string filePath = sl.Path;
                    if(!m_thread.m_threadMgr.m_process.m_engine.Options.ShowFullPaths)
                        filePath = Path.GetFileName(sl.Path);
                    sp = " ("+filePath+":"+sl.Line.ToString(System.Globalization.CultureInfo.CurrentUICulture)+")";
                }
                else
                    sp=" (source line information unavailable)";

                StringBuilder sbFuncName = new StringBuilder();                    

                MDbgModule module = this.Function.Module;
                MDbgProcess proc = m_thread.m_threadMgr.m_process;


                // Get class name w/ generic args.
                CorType tClass = this.FunctionType;
                if (tClass != null)
                    InternalUtil.PrintCorType(sbFuncName, proc, tClass);

                sbFuncName.Append('.');


                // Get method name w/ generic args.
                MethodInfo mi = this.Function.MethodInfo;
                sbFuncName.Append(mi.Name);
                InternalUtil.AddGenericArgs(sbFuncName, proc, this.FunctionTypeParameters);


                string stFuncName = sbFuncName.ToString();                    
                
                if( format=="v" )
                {
                    CorModule m = module.CorModule;
                    // verbose frame output
                    // in verbose output we'll print module name + arguments to the functions
                    StringBuilder sb = new StringBuilder();
                    bool bFirst = true;
                    foreach(MDbgValue v in this.Function.GetArguments(this))
                    {
                        if( sb.Length!=0 )
                            sb.Append(", ");
                        // skip this references
                        if( ! (bFirst && v.Name == "this") )
                            sb.Append(v.Name).Append("=").Append(v.GetStringValue(0));
                        bFirst = false;
                    }
                    
                    if( m.IsDynamic || m.IsInMemory )
                    {
                        fn = m.Name;
                    }
                    else
                    {
                        fn = System.IO.Path.GetFileName(m.Name);
                    }

                    MDbgAppDomain ad = this.m_thread.m_threadMgr.m_process.AppDomains.Lookup(m.Assembly.AppDomain);
                    fn += "#" + ad.Number
                        + "!" + stFuncName + "(" + sb.ToString() + ") " + sp;
                }
                else
                {
                    fn = stFuncName + sp;
                }
                break;
            case CorFrameType.InternalFrame:
                switch (m_frame.InternalFrameType)
                {
                case CorDebugInternalFrameType.STUBFRAME_NONE:
                    fn = "None";
                    break;
                case CorDebugInternalFrameType.STUBFRAME_M2U:
                    fn = "M-->U";
                    break;
                case CorDebugInternalFrameType.STUBFRAME_U2M:
                    fn = "U-->M";
                    break;
                case CorDebugInternalFrameType.STUBFRAME_APPDOMAIN_TRANSITION:
                    fn = "AD Switch";
                    break;
                case CorDebugInternalFrameType.STUBFRAME_LIGHTWEIGHT_FUNCTION:
                    fn = "LightWeight";
                    break;
                case CorDebugInternalFrameType.STUBFRAME_FUNC_EVAL:
                    fn = "FuncEval";
                    break;
                default:
                    fn = "UNKNOWN";
                    break;
                }
                fn = "[Internal Frame, '" + fn + "']";
                break;
            default:
                fn = "UNKNOWN Frame Type";
                break;
            }
            return fn;
        }


        // Type parameter information.
        

        internal MDbgThread m_thread;
        private CorFrame m_frame;

        // This static method is used by MDbgThread.GetFrameFromNextChain
        internal static MDbgFrame GetFrameFromChain(MDbgThread thread,CorChain chain,bool topmost)
        {
            Debug.Assert(thread!=null);
            Debug.Assert(chain!=null && chain.IsManaged);
            CorFrame f = null;
            if(topmost)
            {
                f = chain.ActiveFrame;
            }
            else
            {
                // finding the bottom frame from the chain, there is no handy method for this,
                // therefore we have to traverse from top frame until we see bottom.
                f = chain.ActiveFrame;
                if( f!=null )
                {                                       // iterate and find the bottom frame
                    CorFrame t = f.Caller;
                    while( t!=null )
                    {
                        f = t;
                        t = f.Caller;
                    }
                    Debug.Assert( f!=null ); // we started with f!=null, we should keep that
                }
            }
            return (f==null)?null:new MDbgILFrame(thread,f);
        }
    }

    // Class for some internal utility functions.
    
    /// <summary>
    /// InternalUtil - a place for utility functions.
    /// </summary>
    public class InternalUtil  
    {
        // Helper to append generic args from tyenum in pretty format.
        // This will add a string like '<int, Foo<string>>'
        internal static void AddGenericArgs(StringBuilder sb, MDbgProcess proc, IEnumerable tyenum)
        {
            int i = 0;
            foreach(CorType t1 in tyenum)
            {
                sb.Append((i == 0) ? '<' : ',');
                InternalUtil.PrintCorType(sb, proc, t1);
                i++;
            }
            if (i > 0)
            {
                sb.Append('>');
            }
        }
        
    
        // Return class as a string.
        /// <summary>
        /// Creates a string representation of CorType.
        /// </summary>
        /// <param name="proc">A debugged process.</param>
        /// <param name="ct">A CorType object representing a type in the debugged process.</param>
        /// <returns>String representaion of the passed in type.</returns>
        public static string PrintCorType(MDbgProcess proc, CorType ct)
        {
            StringBuilder sb = new StringBuilder();
            PrintCorType(sb, proc, ct);
            return sb.ToString();
        }
        
        // Print CorType to the given string builder.
        // Will print generic info. 
        
        internal static void PrintCorType(StringBuilder sb, MDbgProcess proc, CorType ct)
        {
            switch (ct.Type) 
            {
            case CorElementType.ELEMENT_TYPE_CLASS:
            case CorElementType.ELEMENT_TYPE_VALUETYPE:
                // We need to get the name from the metadata. We can get a cached metadata importer
                // from a MDbgModule, or we could get a new one from the CorModule directly.
                // Is this hash lookup to get a MDbgModule cheaper than just re-querying for the importer?
                CorClass cc = ct.Class;
                MDbgModule m = proc.Modules.Lookup(cc.Module);
                Type tn = m.Importer.GetType(cc.Token);
        
                sb.Append(tn.FullName);
                AddGenericArgs(sb, proc, ct.TypeParameters);
            return;
        
            // Primitives
            case CorElementType.ELEMENT_TYPE_BOOLEAN:       
                sb.Append("System.Boolean"); return;
            case CorElementType.ELEMENT_TYPE_CHAR:          
                sb.Append("System.Char"); return;
            case CorElementType.ELEMENT_TYPE_I1:            
                sb.Append("System.SByte"); return;
            case CorElementType.ELEMENT_TYPE_U1:            
                sb.Append("System.Byte"); return;
            case CorElementType.ELEMENT_TYPE_I2:            
                sb.Append("System.Int16"); return;
            case CorElementType.ELEMENT_TYPE_U2:            
                sb.Append("System.UInt16"); return;
            case CorElementType.ELEMENT_TYPE_I4:            
                sb.Append("System.Int32"); return;
            case CorElementType.ELEMENT_TYPE_U4:            
                sb.Append("System.Uint32"); return;
            case CorElementType.ELEMENT_TYPE_I8:            
                sb.Append("System.Int64"); return;
            case CorElementType.ELEMENT_TYPE_U8:            
                sb.Append("System.UInt64"); return;
            case CorElementType.ELEMENT_TYPE_I:            
                sb.Append("System.IntPtr"); return;
            case CorElementType.ELEMENT_TYPE_U:            
                sb.Append("System.UIntPtr"); return;
            case CorElementType.ELEMENT_TYPE_R4:            
                sb.Append("System.Single"); return;
            case CorElementType.ELEMENT_TYPE_R8:            
                sb.Append("System.Double"); return;
        
            // Well known class-types.
            case CorElementType.ELEMENT_TYPE_OBJECT:        
                sb.Append("System.Object"); return;
            case CorElementType.ELEMENT_TYPE_STRING:        
                sb.Append("System.String"); return;
        
        
            // Special compound types. Based off first type-param
            case CorElementType.ELEMENT_TYPE_SZARRAY:
            case CorElementType.ELEMENT_TYPE_ARRAY:
            case CorElementType.ELEMENT_TYPE_BYREF:
            case CorElementType.ELEMENT_TYPE_PTR:
                CorType t = ct.FirstTypeParameter;
                PrintCorType(sb, proc, t);
                switch(ct.Type)
                {
                case CorElementType.ELEMENT_TYPE_SZARRAY: 
                    sb.Append("[]"); 
                    return;
                case CorElementType.ELEMENT_TYPE_ARRAY:   
                    int rank = ct.Rank;
                    sb.Append('[');
                    for(int i = 0; i < rank - 1; i++)
                    {
                        
                        sb.Append(',');
                    }
                    sb.Append(']'); 
                    return;
                case CorElementType.ELEMENT_TYPE_BYREF:   
                    sb.Append("&"); 
                    return;
                case CorElementType.ELEMENT_TYPE_PTR:     
                    sb.Append("*"); 
                    return;                    
                }
                Debug.Assert(false); // shouldn't have gotten here.             
                return;
            
            case CorElementType.ELEMENT_TYPE_FNPTR:
                sb.Append("*(...)");
                return;
            case CorElementType.ELEMENT_TYPE_TYPEDBYREF:
                sb.Append("typedbyref");
                return;
            default:
                sb.Append("<unknown>");
                return;
            }
        } // end PrintClass
    } // end class InternalUtil
}
