//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/wicca/DynamicWeaver.cs 12    5/24/06 9:42p Eaddy $
//
// Description:
//
// Remarks:
//
//  Load-time Weaving Algorithm
//  ---------------------------
//  -> When an executable is loaded, we should do the following:
//  	1. Extract aspects
//  	2. Apply aspects to all currently loaded executables (including
//  	   the one we just loaded)
//  
//  However, this is an O(n^2) algorithm.  Consider 3 executables: a, b, and c.
//  a will be parsed 3 times, b twice, and c once.  We can reduce this by only
//  extracting and applying the aspects when all modules are loaded, thus a,
//  b, and c would each be loaded once.  However, dynamic loading means that
//  we cannot know when the last module is loaded.  Furthermore, to be 100%
//  correct, we may need to weave a module before any code in that module has
//  executed.  Since, in theory, code in a module may be executed due to static
//  initialization, we'd have to perform weaving even before the Main() method
//  is executed.
//  
//  We can reduce the overhead of this algorithm using the following heuristics:
//  
//  1. Assume system modules have no aspects
//  2. Do not weave system modules
//  3. Do not begin weaving until the program executable has been loaded
//  4. Only apply new aspects to previously loaded modules
//  
//  We can also reduce overhead by doing the following:
//  
//  1. Only extract aspects once for each module the first time it is loaded
//  2. Never parse a module twice during a single weaving pass.  This requires
//     consolidating aspect extraction and weaving.
//  
//  Dynamic Weaving Algorithm
//  -------------------------
//  Dynamic weaving allows you to add/remove aspect behavior at any time from
//  one or more join points.
//  
//  -> Aspect
//  	-> Advice
//  		-> Matched Join Points
//  
//  - Each aspect and advice has an associated "exclusion pointcut
//    expression".  This allows fine-grained control over advising.  An entire
//    aspect or advice can be disabled by specifying "false".  Individual
//    join points can be disabled at the aspect or advice level.
//  - A common exclusion pointcut expression can also be specified for all
//    aspects.  All weaving can be completely disabled by specifying "false".
//    Weaving of specific classes can be completely disabled by specifying
//    "within(Hello)".
//  - To allow the user to disable individual join points, each advice will
//    keep track of its matched join points.
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/wicca/DynamicWeaver.cs $
 * 
 * 12    5/24/06 9:42p Eaddy
 * o Now Verbosity is a static field
 * 
 * 11    5/11/06 11:06p Eaddy
 * o Need to still weave loaded modules when user specifies
 * an aspect using the !weave command to ensure that the
 * aspect eventually gets loaded
 * 
 * 10    5/11/06 10:14a Eaddy
 * o Added shouldWeave helper function
 * 
 * 9     5/10/06 9:02p Eaddy
 * o Stores the MdbgModule in DynamicModule
 * so we don't have to keep looking it up
 * o ShouldWeave is set based on the module's EnC
 * flag
 * o Fixed problem where an aspect loaded using !weave
 * was not later woven
 * 
 * 8     5/10/06 8:03p Eaddy
 * o To reduce the amount of churn, does not try to weave system files
 * 
 * 7     5/05/06 6:18p Eaddy
 * o Did some renaming to improve clarity
 * o Can now list current advice and enable/disable (although
 * enable/disable doesn't actually work yet)
 * 
 * 6     5/05/06 12:27p Eaddy
 * o Improved diagnostic msgs
 * 
 * 5     5/05/06 11:49a Eaddy
 * o Multiple weaving works
 * 
 * 4     5/05/06 11:26a Eaddy
 * o Now weaves multiple times
 * 
 * 3     5/05/06 8:19a Eaddy
 * o Defaults to only weaving once until I can fix this
 * 
 * 2     5/03/06 8:14p Eaddy
 * o Now passes IMetaDataImport to CreateDeltas
 * o Now verifies that the module allows EnC before doing anything
 * 
 * 1     5/03/06 12:33a Eaddy
 * o Moved weaving and patching into DynamicWeaver
 */

using System.IO;
using System.Collections.Generic;

using Dbg = System.Diagnostics.Debug;

using Microsoft.Samples.Debugging.MdbgEngine;
using Microsoft.Samples.Debugging.CorMetadata;

using Phx.Morph;

namespace Wicca
{
    class DynamicModule
    {
        public DynamicModule(MDbgModule dbgModule, string originalPath, bool isLoaded, bool shouldWeave)
        {
            this.dbgModule = dbgModule;
            this.originalPath = originalPath;
            this.currentPath = originalPath;
            this.isLoaded = isLoaded;
            this.shouldWeave = shouldWeave;
        }

        public override string ToString()
        {
            return originalPath + " (Loaded:" + isLoaded + ", ShouldWeave: " + shouldWeave + ")";
        }

        public MDbgModule dbgModule;
        public string originalPath;
        public string currentPath;
        public bool isLoaded;
        public bool shouldWeave;
    }

    public class DynamicWeaver
    {
        public Morpher morpher = new Morpher();

        AssemblyHelperLib.EnCDeltaHelper deltaHelper =
           new AssemblyHelperLib.EnCDeltaHelper();

        List<DynamicModule> modules = new List<DynamicModule>();

        bool startWeaving = false;

        WeaverType m_weaverType; // This must be set via the Property

        Phx.Aop.BreakpointWeaverCallback setBreakpointCallback;

        static string systemDir = Path.GetDirectoryName(System.Environment.SystemDirectory).ToLower();

        static bool onlyWeaveOnce = false;

        public DynamicWeaver(   Phx.Aop.BreakpointWeaverCallback setBreakpointCallback,
                                WeaverType defaultWeaverType)
        {
            this.setBreakpointCallback = setBreakpointCallback;

            // Weaver type MUST be initialized via the Property setter so that
            // Morpher is updated correctly
            WeaverType = defaultWeaverType;
        }

        public int Verbosity
        {
            set { Morpher.Verbosity = value; }
        }

        public WeaverType WeaverType
        {
            get { return m_weaverType; }
            set
            {
                m_weaverType = value;

                if (m_weaverType == WeaverType.breakpoint)
                {
                    Morpher.BreakpointWeaverCallback = setBreakpointCallback;
                }
                else
                {
                    Morpher.BreakpointWeaverCallback = null;
                }
            }
        }

        public void addModule(string modulePath, bool isLoaded, bool isMainModule)
        {
            MDbgModule dbgModule = WiccaPlugin.Debugger.Processes.Active.Modules.Lookup(modulePath);

            bool shouldWeave = this.shouldWeave(dbgModule);

            foreach (DynamicModule module in modules)
            {
                if (Path.GetFileName(module.originalPath) == Path.GetFileName(modulePath))
                {
                    // The user already specified the aspect using the !weave command.
                    // Its likely that the aspect was woven in causing a subsequent
                    // ModuleLoad event which we will ignore.  The good news is that
                    // we will avoid weaving the same advice twice.  The bad news is
                    // that manually loaded aspects can't be themselves woven.
                    Dbg.Assert(!module.isLoaded && isLoaded);

                    if (module.dbgModule == null)
                    {
                        Dbg.Assert(dbgModule != null);

                        // Initialize and weave the module now that its loaded
                        module.dbgModule = dbgModule;
                        module.isLoaded = true;
                        module.shouldWeave = shouldWeave;
                        
                        if (shouldWeave && startWeaving)
                            weave(module);
                    }

                    return;
                }
            }

            bool isSystemModule = DynamicWeaver.isSystemModule(modulePath);

            bool weaveAnyway = false;

            if (dbgModule == null)
            {
                WiccaPlugin.WriteOutput("Module '{0}' has not been loaded yet, postponing weaving.", 
                    System.IO.Path.GetFileName(modulePath));
                shouldWeave = false;

                // The user specified an aspect using !weave. Even though the
                // aspect isn't loaded yet, and therefore we can't weave the
                // aspect itself, we still want to weave the other loaded
                // modules.  This will likely cause the aspect to be woven into
                // some of the modules which will in turn cause the aspect to
                // be loaded and then woven.
                weaveAnyway = true;
            }
            else if (dbgModule.CorModule.JITCompilerFlags != Microsoft.Samples.Debugging.CorDebug.CorDebugJITCompilerFlags.CORDEBUG_JIT_ENABLE_ENC)
            {
                WiccaPlugin.WriteOutput("Module '{0}' cannot be modified at runtime, skipping weaving.",
                    System.IO.Path.GetFileName(modulePath));
                shouldWeave = false;
            }

            modules.Add(new DynamicModule(dbgModule, modulePath, isLoaded, shouldWeave));

            // Heuristic: Assume system DLLs do not contain aspects
            if (!isSystemModule)
            {
               morpher.AddAspectCandidate(modulePath);
            }

            // Heuristic: Don't start weaving until the main module has been
            // loaded.

            if (isMainModule)
            {
               Dbg.Assert(startWeaving == false);
               startWeaving = true;
            }

            if ((shouldWeave || weaveAnyway) && startWeaving)
                weaveAll();
        }

        public bool shouldWeave(MDbgModule module)
        {
            return m_weaverType == WeaverType.breakpoint ||
                (module != null && module.CorModule.JITCompilerFlags == Microsoft.Samples.Debugging.CorDebug.CorDebugJITCompilerFlags.CORDEBUG_JIT_ENABLE_ENC);
        }

        public static bool isSystemModule(string path)
        {
            return path.ToLower().StartsWith(systemDir);
        }

        public void weaveAll()
        {
            foreach (DynamicModule module in modules)
            {
                weave(module);
            }
        }

        void weave(DynamicModule module)
        {
            if (module.isLoaded && module.shouldWeave)
            {
                bool success = weave(module.dbgModule, module.originalPath, module.currentPath);

                if (success && morpher.OutputPath != null)
                {
                    module.currentPath = morpher.OutputPath;
                }

                if (!success || onlyWeaveOnce)
                {
                    // If we can't weave the module, stop trying
                    module.shouldWeave = false;
                }

                morpher.OutputPath = null;
            }
        }

        /// <returns>
        /// true if there were no weaving errors, otherwise false
        /// </returns>
        public bool weave(MDbgModule oldModule, string originalModulePath, string currentModulePath)
        {
            Dbg.Assert(oldModule != null);
            Dbg.Assert(oldModule.CorModule.JITCompilerFlags == Microsoft.Samples.Debugging.CorDebug.CorDebugJITCompilerFlags.CORDEBUG_JIT_ENABLE_ENC);

            WiccaPlugin.WriteOutput("Weaving '{0}'.", originalModulePath);

            // ISSUE-REVIEW-t-marce-2006-05-10-Phx.Morph cannot use
            // ICorDebugModule and IMetaDataImport because Phx does not
            // support in-memory PE rewriting.  We fake it by always
            // morphing the last module we morphed.  To prevent reweaving
            // the same join point twice, each advice keeps a list of
            // jps it was applied to.
            morpher.InputModule = currentModulePath; //originalModulePath;

            Morpher.WeaveResult weaveResult = morpher.Weave();

            if (weaveResult == Morpher.WeaveResult.FAILED)
            {
               return false; // So we don't try to weave this module again
            }

            if (morpher.OutputPath == null ||
                weaveResult == Morpher.WeaveResult.NO_CHANGE)
            {
                return true;
            }

            Dbg.Assert(weaveResult == Morpher.WeaveResult.SUCCESS);

            return createAndApplyDeltas(oldModule, originalModulePath, morpher.OutputPath);
        }

        /// <returns>
        /// true if there were no weaving errors, otherwise false
        /// </returns>
        public bool createAndApplyDeltas(   string oldModulePath, 
                                            string newModulePath)
        {
            MDbgModule oldModule = WiccaPlugin.Debugger.Processes.Active.Modules.Lookup(oldModulePath);
            if (oldModule == null)
            {
                WiccaPlugin.WriteError("Module '{0}' has not been loaded yet.", 
                    System.IO.Path.GetFileName(oldModulePath));
                Dbg.Assert(false);
                return false;
            }
            else if (oldModule.CorModule.JITCompilerFlags != Microsoft.Samples.Debugging.CorDebug.CorDebugJITCompilerFlags.CORDEBUG_JIT_ENABLE_ENC)
            {
                WiccaPlugin.WriteOutput("Module '{0}' cannot be modified at runtime. Edit-and-Continue not enabled.",
                   System.IO.Path.GetFileName(oldModulePath));
                return false;
            }

            return createAndApplyDeltas(oldModule, oldModulePath, newModulePath);
        }

        /// <returns>
        /// true if there were no weaving errors, otherwise false
        /// </returns>
        public bool createAndApplyDeltas(   MDbgModule oldModule,
                                            string oldModulePath,
                                            string newModulePath)
        {
            string metadataDeltaPath;
            string iLDeltaPath;
            string pdbDeltaPath;

            Morpher.WeaveResult createDeltasResult = createDeltas(  oldModule,
                                                                    oldModulePath,
                                                                    newModulePath,
                                                                    out metadataDeltaPath,
                                                                    out iLDeltaPath,
                                                                    out pdbDeltaPath);
            
            if (createDeltasResult == Morpher.WeaveResult.FAILED)
            {
                return false;
            }
            else if (createDeltasResult == Morpher.WeaveResult.NO_CHANGE)
            {
                return true;
            }
            else
            {
                Dbg.Assert(createDeltasResult == Morpher.WeaveResult.SUCCESS);

                return applyDeltas(   oldModule,
                                      metadataDeltaPath,
                                      iLDeltaPath,
                                      pdbDeltaPath);
            }
        }

        Morpher.WeaveResult createDeltas(   MDbgModule oldModule, 
                                            string oldModulePath,
                                            string newModulePath,
                                            out string metadataDeltaPath,
                                            out string iLDeltaPath,
                                            out string pdbDeltaPath)
        {
            Dbg.Assert(oldModule != null);
            Dbg.Assert(oldModule.CorModule.JITCompilerFlags == Microsoft.Samples.Debugging.CorDebug.CorDebugJITCompilerFlags.CORDEBUG_JIT_ENABLE_ENC);
            Dbg.Assert(newModulePath != null);
            Dbg.Assert(newModulePath != string.Empty);
            Dbg.Assert(oldModulePath != null);
            Dbg.Assert(oldModulePath != string.Empty);
            Dbg.Assert(deltaHelper != null);

            // Initialize out params
            metadataDeltaPath = null;
            iLDeltaPath = null;
            pdbDeltaPath = null;

            WiccaPlugin.WriteOutput("Comparing with new module '{0}'.", newModulePath);

            string debugMsg;

            try
            {
                System.IntPtr piCorDebugModuleOrig = System.Runtime.InteropServices.Marshal.GetIUnknownForObject(
                    oldModule.CorModule.CorModuleInterface);

                Dbg.Assert(oldModule.MatchesModuleName(oldModulePath));

                deltaHelper.CreateDeltas(   (long) piCorDebugModuleOrig,
                                            oldModulePath,
                                            newModulePath,
                                            Path.GetDirectoryName(newModulePath),
                                            out metadataDeltaPath,
                                            out iLDeltaPath,
                                            out debugMsg);

                pdbDeltaPath = Path.Combine(Path.GetDirectoryName(newModulePath),
                    Path.GetFileNameWithoutExtension(newModulePath) + ".pdb");
            }
            catch (System.Runtime.InteropServices.COMException ex)
            {
                WiccaPlugin.WriteError(ex.Message);
                return Morpher.WeaveResult.FAILED;
            }

            if (debugMsg != null && debugMsg != string.Empty)
            {
                WiccaPlugin.WriteOutput(debugMsg);
                return Morpher.WeaveResult.SUCCESS;
            }
            else
            {
                WiccaPlugin.WriteOutput("No changes were detected.");
                return Morpher.WeaveResult.NO_CHANGE;
            }
        }

        bool applyDeltas(   MDbgModule oldModule,
                            string metadataDeltaPath,
                            string iLDeltaPath,
                            string pdbDeltaPath)
        {
            if (oldModule == null ||
                metadataDeltaPath == null || 
                metadataDeltaPath == string.Empty ||
                iLDeltaPath == null ||
                iLDeltaPath == string.Empty)
            {
                Dbg.Assert(false);
                return false;
            }

            Dbg.Assert(oldModule != null);
            Dbg.Assert(oldModule.CorModule.JITCompilerFlags == Microsoft.Samples.Debugging.CorDebug.CorDebugJITCompilerFlags.CORDEBUG_JIT_ENABLE_ENC);

            WiccaPlugin.WriteOutput("Applying delta files '{0}' and '{1}'.",
                metadataDeltaPath, iLDeltaPath);

            try
            {
                oldModule.ApplyEdit(    metadataDeltaPath,
                                        iLDeltaPath,
                                        pdbDeltaPath,
                                        null); // Edit source file

                WiccaPlugin.WriteOutput("Update complete.");
                return true;
            }
            catch (System.Runtime.InteropServices.COMException ex)
            {
                WiccaPlugin.WriteError("Update failed. " + ex.Message);
                return false;
            }
            catch (System.ArgumentException ex)
            {
                WiccaPlugin.WriteError("Update failed. " + ex.Message);
                return false;
            }
        }
    }
}
